#############################################################################
# menubuttons.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# Menu VOB buttons infos.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

## For each VTS, the following variables are set:
## ::buttons::butsinfo($type,$vts,errorcode)	;# 0=OK
## ::buttons::butsinfo($type,$vts,errormsg)
## ::buttons::butsinfo($type,$vts,numvobids)
## For each VOB ID
## ::buttons::butsinfo($type,$vts,$vobid,numcellids)
## For each Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,navpack)	;# first NAV pack LBA for current VOB/Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numbuts)	;# number of buttons (in first group)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups)

## Defined only if several sets are detected:
## ::buttons::butsinfo($type,multiplesets)					;# Bool: menus/BOVs contain multiple-sets of buttons/BOVs
## ::buttons::butsinfo($type,$vts,multiplesets)				;# Bool: domain contain multiple-sets of buttons/BOVs
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numsets)	;# number of sets of buttons in this V/CID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$set,lba)	;# LBA of first nav pack of this set
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$set,hl_status) ;# Highlight status of first nav pack of this set

## ::buttons::butsinfo($type,$vts,$vobid,$cellid,edited)	;# This cell has been edited
##															;# (Defined only after edition!)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,hlstarttime) ;# start and end time
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,hlendtime)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,selendtime)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceselectbutn) ;# forced buttons
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceactionbutn)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numselectablebuts)	;# number of numerically selectable buttons
## For each group
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,grouptype)	;# 1=normal, 2=letterbox, 4=pan&scan (ored together)
## For each button
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,autoaction) ;# 0 or 1
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,colorscheme) ;# 0 to 3
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,coords)	;# Button position: {x_start y_start - x_end y_end}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,udlr)		;# Adjacent buttons: {up down left right}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,command)	;# Button command: {XX XX XX XX XX XX XX XX}
## and
## ::pgcs($vts,$pgc,cellpostable)		;# Cell Position Table, in binary
##
## For each VTS in Title domain ONLY:
## ::buttons::butsinfo($type,$vts,scanned) ;# 1=the domain has already been scanned for BOVs, 0=not scanned, 2=scan error
## ::pgcs($vts,$pgc,numbuttons)	;# number of BOVs in all cells of the PGC
## For each VTS in Menu domain:
## ::menus($vts,$lu,$pgc,numbuttons)	;# number of buttons in all cells of the PGC

namespace eval ::menubuttons {

proc get_menu_buttons_infos {vts type} {
	unset -nocomplain ::buttons::curbutsinfo

	set ::buttons::curbutsinfo(ispal) [lindex [get_screen_dims $type $vts] 2]

	if {$type == "menu"} {
		if {[info exists ::buttons::butsinfo(menu,$vts,errorcode)]} {
			set ::buttons::curbutsinfo(errorcode) $::buttons::butsinfo(menu,$vts,errorcode)
			set ::buttons::curbutsinfo(errormsg) $::buttons::butsinfo(menu,$vts,errormsg)
			if {$::buttons::curbutsinfo(errorcode) == 0} {
				set ::buttons::curbutsinfo(numvobids) $::buttons::butsinfo(menu,$vts,numvobids)
				for {set vobid 1} {$vobid <= $::buttons::curbutsinfo(numvobids)} {incr vobid} {
					set ::buttons::curbutsinfo($vobid,numcellids) $::buttons::butsinfo(menu,$vts,$vobid,numcellids)
					for {set cellid 1} {$cellid <= $::buttons::curbutsinfo($vobid,numcellids)} {incr cellid} {
						set ::buttons::curbutsinfo($vobid,$cellid,navpack) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,navpack)
						set ::buttons::curbutsinfo($vobid,$cellid,numbuts) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)

						if {[info exists ::buttons::butsinfo(menu,$vts,$vobid,$cellid,numsets)]} {
							set numsets $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numsets)
						} else {
							set numsets 0
						}
						set ::buttons::curbutsinfo($vobid,$cellid,numsets) $numsets
						for {set set 2} {$set <= $numsets} {incr set} {
							set ::buttons::curbutsinfo($vobid,$cellid,$set,lba)       $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$set,lba)
							set ::buttons::curbutsinfo($vobid,$cellid,$set,hl_status) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$set,hl_status)
						}

						if {$::buttons::curbutsinfo($vobid,$cellid,numbuts) > 0} {
							set ::buttons::curbutsinfo($vobid,$cellid,numgroups) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numgroups)
							if {[info exists ::buttons::butsinfo(menu,$vts,$vobid,$cellid,numsets)]} {
								set ::buttons::curbutsinfo($vobid,$cellid,numsets) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numsets)
							} else {
								set ::buttons::curbutsinfo($vobid,$cellid,numsets) 1
							}
							set ::buttons::curbutsinfo($vobid,$cellid,hlstarttime) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,hlstarttime)
							set ::buttons::curbutsinfo($vobid,$cellid,hlendtime)   $::buttons::butsinfo(menu,$vts,$vobid,$cellid,hlendtime)
							set ::buttons::curbutsinfo($vobid,$cellid,selendtime)  $::buttons::butsinfo(menu,$vts,$vobid,$cellid,selendtime)
							set ::buttons::curbutsinfo($vobid,$cellid,numselectablebuts) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numselectablebuts)
							set ::buttons::curbutsinfo($vobid,$cellid,forceselectbutn) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,forceselectbutn)
							set ::buttons::curbutsinfo($vobid,$cellid,forceactionbutn) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,forceactionbutn)
							for {set group 1} {$group <= $::buttons::curbutsinfo($vobid,$cellid,numgroups)} {incr group} {
								set ::buttons::curbutsinfo($vobid,$cellid,$group,grouptype) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,grouptype)
								for {set butnum 1} {$butnum <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {incr butnum} {
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,autoaction) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,autoaction)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,colorscheme) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,colorscheme)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,coords) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,coords)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,udlr) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,udlr)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command) $::buttons::butsinfo(menu,$vts,$vobid,$cellid,$group,$butnum,command)
								}
							}
						}
					}
				}
			}
		} else {
			set ::buttons::curbutsinfo(errorcode) -2
			set ::buttons::curbutsinfo(errormsg) "Unknown error."
		}
	} else {
		if {[info exists ::buttons::butsinfo(pgc,$vts,scanned)] && $::buttons::butsinfo(pgc,$vts,scanned) == 1} {
			set ::buttons::curbutsinfo(errorcode) $::buttons::butsinfo(pgc,$vts,errorcode)
			set ::buttons::curbutsinfo(errormsg) $::buttons::butsinfo(pgc,$vts,errormsg)
			if {$::buttons::curbutsinfo(errorcode) == 0} {
				set ::buttons::curbutsinfo(numvobids) $::buttons::butsinfo(pgc,$vts,numvobids)
				for {set vobid 1} {$vobid <= $::buttons::curbutsinfo(numvobids)} {incr vobid} {
					set ::buttons::curbutsinfo($vobid,numcellids) $::buttons::butsinfo(pgc,$vts,$vobid,numcellids)
					for {set cellid 1} {$cellid <= $::buttons::curbutsinfo($vobid,numcellids)} {incr cellid} {
						set ::buttons::curbutsinfo($vobid,$cellid,navpack) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,navpack)
						set ::buttons::curbutsinfo($vobid,$cellid,numbuts) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts)

						if {[info exists ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numsets)]} {
							set numsets $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numsets)
						} else {
							set numsets 0
						}
						set ::buttons::curbutsinfo($vobid,$cellid,numsets) $numsets
						for {set set 2} {$set <= $numsets} {incr set} {
							set ::buttons::curbutsinfo($vobid,$cellid,$set,lba)       $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$set,lba)
							set ::buttons::curbutsinfo($vobid,$cellid,$set,hl_status) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$set,hl_status)
						}

						if {$::buttons::curbutsinfo($vobid,$cellid,numbuts) > 0} {
							set ::buttons::curbutsinfo($vobid,$cellid,numgroups) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numgroups)
							if {[info exists ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numsets)]} {
								set ::buttons::curbutsinfo($vobid,$cellid,numsets) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numsets)
							} else {
								set ::buttons::curbutsinfo($vobid,$cellid,numsets) 1
							}
							set ::buttons::curbutsinfo($vobid,$cellid,hlstarttime) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,hlstarttime)
							set ::buttons::curbutsinfo($vobid,$cellid,hlendtime)   $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,hlendtime)
							set ::buttons::curbutsinfo($vobid,$cellid,selendtime)  $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,selendtime)
							set ::buttons::curbutsinfo($vobid,$cellid,numselectablebuts) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numselectablebuts)
							set ::buttons::curbutsinfo($vobid,$cellid,forceselectbutn) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,forceselectbutn)
							set ::buttons::curbutsinfo($vobid,$cellid,forceactionbutn) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,forceactionbutn)
							for {set group 1} {$group <= $::buttons::curbutsinfo($vobid,$cellid,numgroups)} {incr group} {
								set ::buttons::curbutsinfo($vobid,$cellid,$group,grouptype) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,grouptype)
								for {set butnum 1} {$butnum <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {incr butnum} {
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,autoaction) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,autoaction)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,colorscheme) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,colorscheme)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,coords) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,coords)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,udlr) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,udlr)
									set ::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command) $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,$group,$butnum,command)
								}
							}
						}
					}
				}
			}
		} else {
			set ::buttons::curbutsinfo(errorcode) -1
			set ::buttons::curbutsinfo(errormsg) "This domain has not been parsed for BOVs."
		}
	}
	return $::buttons::curbutsinfo(errorcode)
}


#############################################################################
## Procedure:  dump_pgc_menu_buttons
## args: if false, suppress the [hex codes] from the dump
proc dump_pgc_menu_buttons {vts numcells cellpostable indentstring args} {
	get_menu_buttons_infos $vts menu

	if {$numcells == 0} {return ""}

	set txt "${indentstring}********** menu buttons commands:\n"
	for {set cell 0} {$cell < $numcells} {incr cell} {
		set idx [expr {$cell * 4}]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if {[info exists ::buttons::curbutsinfo($vobid,$cellid,numbuts)]} {
			set numbuts $::buttons::curbutsinfo($vobid,$cellid,numbuts)
			if {$numbuts == 0} {continue}
			set txt "${txt}${indentstring}VOB ID $vobid,  Cell ID $cellid  (First NAV pack LBA=$::buttons::curbutsinfo($vobid,$cellid,navpack))\n"

			set numgroups $::buttons::curbutsinfo($vobid,$cellid,numgroups)
			set txt "${txt}${indentstring}$numgroups Group"
			if {$numgroups > 1} {
				set txt "${txt}s"
			}
			set txt "$txt  Select button=$::buttons::curbutsinfo($vobid,$cellid,forceselectbutn),"
			set txt "$txt  Action button=$::buttons::curbutsinfo($vobid,$cellid,forceactionbutn)\n"
			for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
				for {set group 1} {$group <= $numgroups} {incr group} {
					set cmdtxt "${indentstring}  [format "%2d" $butnum]: "
					set hex $::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command)
					if {$args == false} {
						set cmdtxt "${cmdtxt}[::disasm::disasm $hex] "
					} else {
						set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
					}
					if {$::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,autoaction)} {
						set cmdtxt "$cmdtxt*** Autoaction "
					} else {
						set cmdtxt "$cmdtxt"
					}
					set cmdtext($group) $cmdtxt
				}
				set samecmds true
				if {$numgroups > 1} {
					if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
					if {$numgroups > 2} {
						if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
					}
				}
				if {! $samecmds} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::curbutsinfo($vobid,$cellid,$group,grouptype)])\n"
					}
				} else {
					set txt "${txt}$cmdtext(1)\n"
				}
			}
		}
	}
	return $txt
}

#############################################################################
## Procedure:  dump_pgc_BOVs
## args: if false, suppress the [hex codes] from the dump
proc dump_pgc_BOVs {vts numcells cellpostable indentstring args} {
	get_menu_buttons_infos $vts pgc

	if {$numcells == 0} {return ""}

	set txt ""
	for {set cell 0} {$cell < $numcells} {incr cell} {
		set idx [expr {$cell * 4}]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if {[info exists ::buttons::curbutsinfo($vobid,$cellid,numbuts)]} {
			set numbuts $::buttons::curbutsinfo($vobid,$cellid,numbuts)
			if {$numbuts == 0} {continue}
			set txt "${txt}${indentstring}VOB ID $vobid,  Cell ID $cellid  (First NAV pack LBA=$::buttons::curbutsinfo($vobid,$cellid,navpack))\n"

			set numgroups $::buttons::curbutsinfo($vobid,$cellid,numgroups)
			set txt "${txt}${indentstring}$numgroups Group"
			if {$numgroups > 1} {
				set txt "${txt}s"
			}
			set txt "$txt  Select button=$::buttons::curbutsinfo($vobid,$cellid,forceselectbutn),"
			set txt "$txt  Action button=$::buttons::curbutsinfo($vobid,$cellid,forceactionbutn)\n"
			for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
				for {set group 1} {$group <= $numgroups} {incr group} {
					set cmdtxt "${indentstring}  [format "%2d" $butnum]: "
					set hex $::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command)
					if {$args == false} {
						set cmdtxt "${cmdtxt}[::disasm::disasm $hex] "
					} else {
						set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
					}
					if {$::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,autoaction)} {
						set cmdtxt "$cmdtxt*** Autoaction "
					} else {
						set cmdtxt "$cmdtxt"
					}
					set cmdtext($group) $cmdtxt
				}
				set samecmds true
				if {$numgroups > 1} {
					if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
					if {$numgroups > 2} {
						if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
					}
				}
				if {! $samecmds} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::curbutsinfo($vobid,$cellid,$group,grouptype)])\n"
					}
				} else {
					set txt "${txt}$cmdtext(1)\n"
				}
			}
		}
	}
	if {$txt != ""} {
		set txt "${indentstring}********** BOV commands:\n$txt"
	}
	return $txt
}

#############################################################################
## Procedure:  dump_all_button_infos
proc dump_all_button_infos {header} {
	set out "$header"
	set out "$out\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set out "$out\nProvider ID: \"[::utils::get_provider_ID]\""
	set out "$out\nNumber of VTS: $::pgcs(numvts)\n"

	set progress_max [expr $::pgcs(numvts)+1]

	# VIDEO_TS.IFO
	show_msg "Dumping VMGM buttons..." 1 $progress_max
	set tmp ""
	if {$::buttons::butsinfo(menu,0,errorcode) == 0} {
		set tmp [dump_vts_button_infos 0 menu $::config(dump_hex_codes)]
	} else {
		set tmp $::buttons::butsinfo(menu,0,errormsg)\n\n
	}
	if {$tmp!=""} {
		set out "$out\nButtons in VMGM (VIDEO_TS.VOB)"
		set out "$out\n-------------------------------"
		set out $out\n$tmp
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		show_msg "Dumping VTS $vts buttons..." [expr $vts+1] $progress_max
		set tmp ""
		if {$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
			set tmp [dump_vts_button_infos $vts menu $::config(dump_hex_codes)]
		} else {
			set tmp $::buttons::butsinfo(menu,$vts,errormsg)\n\n
		}
		if {$tmp!=""} {
			set out "$out\nButtons in VTSM $vts (VTS_[format %02i $vts]_0.VOB)"
			set out "$out\n---------------------------------"
			set out $out\n$tmp
		}

		if {[info exists ::buttons::butsinfo(pgc,$vts,errorcode)] && \
				$::buttons::butsinfo(pgc,$vts,errorcode) == 0} {
			set doit false
			for {set vid 1} {$vid<=$::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
				for {set cid 1} {$cid<=$::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
					if {$::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)>0} {
						set doit true
						break
					}
				}
			}
			if {$doit} {
				set out "${out}** BOVs in VTST $vts (VTS_[format %02i $vts]_*.VOB)"
				set out "$out\n---------------------------------"
				set out "$out\n[dump_vts_button_infos $vts pgc $::config(dump_hex_codes)]"
			}
		}

	}
	show_msg ""
	return [string trimright $out "\n"]\n
}

#############################################################################
## Procedure:  dump_vts_button_infos
## args: if false, suppress the [hex codes] from the dump
proc dump_vts_button_infos {vts type args} {
	get_menu_buttons_infos $vts $type

	if {$type=="menu"} {
		if {$vts == 0} {set head "VMGM"} else {set head "VTSM $vts"}
	} else {
		set head "VTST $vts"
	}
	set txt ""
	for {set vobid 1} {$vobid <= $::buttons::curbutsinfo(numvobids)} {incr vobid} {
		if {$::buttons::curbutsinfo($vobid,numcellids) == 0} {
			set txt "$txt  $head,  VOB ID $vobid:  THIS VOB ID IS MISSING IN THE VOB FILE!\n\n"
		}
		for {set cellid 1} {$cellid <= $::buttons::curbutsinfo($vobid,numcellids)} {incr cellid} {
			if {$::buttons::curbutsinfo($vobid,$cellid,navpack) == -1} {
				set txt "$txt  $head,  VOB ID $vobid,  Cell ID $cellid:  THIS CELL ID IS MISSING IN THE VOB FILE!\n\n"
				break
			}
			if {$::buttons::curbutsinfo($vobid,$cellid,numbuts) > 0} {
				set txt "$txt  $head,  VOB ID $vobid,  Cell ID $cellid  (First NAV pack LBA=$::buttons::curbutsinfo($vobid,$cellid,navpack))\n"
				set txt "$txt  $::buttons::curbutsinfo($vobid,$cellid,numgroups) Group"
				if {$::buttons::curbutsinfo($vobid,$cellid,numgroups) > 1} {
					set txt "${txt}s"
				}
				set txt "$txt  Select but=$::buttons::curbutsinfo($vobid,$cellid,forceselectbutn),"
				set txt "$txt  Action but=$::buttons::curbutsinfo($vobid,$cellid,forceactionbutn)\n"
				set numgroups $::buttons::curbutsinfo($vobid,$cellid,numgroups)
				for {set butnum 1} {$butnum <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {incr butnum} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set cmdtxt "    button [format "%2d" $butnum]: "
						set hex $::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command)
						if {$args == false} {
							set cmdtxt "${cmdtxt}[::disasm::disasm $hex] "
						} else {
							set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
						}
						if {$::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,autoaction)} {
							set cmdtxt "$cmdtxt*** Autoaction "
						} else {
							set cmdtxt "$cmdtxt"
						}
						set cmdtext($group) $cmdtxt
					}
					set samecmds true
					if {$numgroups > 1} {
						if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
						if {$numgroups > 2} {
							if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
						}
					}
					if {! $samecmds} {
						for {set group 1} {$group <= $numgroups} {incr group} {
							set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::curbutsinfo($vobid,$cellid,$group,grouptype)])\n"
						}
					} else {
						set txt "${txt}$cmdtext(1)\n"
					}
				}
				set txt "$txt\n"
			}
		}
	}
	if {$txt!=""} {set txt "$txt\n"}
	return $txt
}

proc show_menu_buttons_info {} {
	if {! [info exists ::dvddir]} {bell; return}
	::main::showtext .buttoncmds "Menu buttons commands" [dump_all_button_infos "Menu buttons commands"]
}

proc grouptype_descr {type} {
	set descr ""
	if {[expr {$type & 1}] == 1} {set descr "${descr}+normal"}
	if {[expr {$type & 2}] == 2} {set descr "${descr}+letterbox"}
	if {[expr {$type & 4}] == 4} {set descr "${descr}+pan&scan"}
	set descr [string trim $descr "+"]
	if {$descr == ""} {set descr "undefined"}
	return $descr
}

###################################################
## Display Buttons GUI
###################################################

proc display_current_menu_buttons {mode} {
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {$::scratchbook_active} {
		bell
		return 0
	}

	set indexlist [split [lindex $::editor_cmds_from 1] ","]
	set ::buttons::groupnum 1

	if {[lindex $::editor_cmds_from 0] == "menu"} {
		return [display_menu_buttons [lindex $indexlist 0] [lindex $indexlist 1] \
				[lindex $indexlist 2] 1024 $mode 0]
	} else {
		return [display_BOVs [lindex $indexlist 0] [lindex $indexlist 1] 1024 $mode 0]
	}
}

# main entry
proc display_menu_buttons {vts lu pgc sprm8 mode show_cell {force_group -1}} {
	set rc [get_menu_buttons_infos $vts menu]
	if {$rc != 0} {
		set txt "Error retrieving menu buttons information in VOBs:"
		set txt "$txt\n$::buttons::curbutsinfo(errormsg)"
		tk_messageBox -title "Menu Button Editor" -icon error \
			-message $txt -parent .
		return 0
	}
	if {$::menus($vts,$lu,$pgc,numcells) == 0} {
		if {$mode == "info"} {
			show_msg "Menu buttons: This menu has no video cells"
			bell
		}
		return 0
	}
	if {$::menus($vts,$lu,$pgc,cellpostable) == ""} {
		show_msg "Menu buttons: This menu has no video cell position table"
		bell
		return 0
	}

	return [display_domain_buttons menu $vts $lu $pgc $sprm8 $mode $show_cell $force_group]
}

proc display_BOVs {vts pgc sprm8 mode show_cell {force_group -1}} {
	if {$vts == 0} {
		bell
		return
	}
	set rc [get_menu_buttons_infos $vts pgc]
	if {$rc != 0} {
		set txt "Error:\n"
		set txt "$txt\nError retrieving BOVs information in VOBs:"
		set txt "$txt\n$::buttons::curbutsinfo(errormsg)"
		tk_messageBox -title "Menu Button Editor" -icon error \
			-message $txt -parent .
		return 0
	}
	if {$::pgcs($vts,$pgc,numcells) == 0} {
		if {$mode == "info"} {
			show_msg "Menu buttons: This PGC has no video cells"
			bell
		}
		return 0
	}
	if {$::pgcs($vts,$pgc,cellpostable) == ""} {
		show_msg "Menu buttons: This PGC has no video cell position table"
		bell
		return 0
	}
	return [display_domain_buttons pgc $vts -1 $pgc $sprm8 $mode $show_cell $force_group]
}

proc display_domain_buttons {type vts lu pgc sprm8 mode show_cell {force_group -1}} {
	set ::menubuttons::menubuttons_done 1
	set utiltitle "Menu buttons"
	set w .menubuttons

	if {! [info exists ::buttons::groupnum]} {
		if {$force_group == -1} {
			set ::buttons::groupnum 1
		} else {
			set ::buttons::groupnum $force_group
		}
	}
	if {$::buttons::curbutsinfo(errorcode) != 0} {
		if {$::buttons::curbutsinfo(errorcode) != -1} {
			show_msg "Menu buttons: Error when parsing VOB file: $::buttons::curbutsinfo(errormsg)"
		} else {
			show_msg "Menu buttons: $::buttons::curbutsinfo(errormsg)"
		}
		bell
		return 0
	}
	if {$type == "menu"} {
		set programmap $::menus($vts,$lu,$pgc,programmap)
		for {set p 1} {$p <= $::menus($vts,$lu,$pgc,numprograms)} {incr p} {
			set cell_program([get_int programmap [expr {$p-1}] 1]) $p
		}
		set old_pg 1
		for {set c 1} {$c <= $::menus($vts,$lu,$pgc,numcells)} {incr c} {
			if {[info exists cell_program($c)]} {
				set old_pg $cell_program($c)
			} else {
				set cell_program($c) $old_pg
			}
		}
		set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
		set numcells $::menus($vts,$lu,$pgc,numcells)
	} else {
		set programmap $::pgcs($vts,$pgc,programmap)
		for {set p 1} {$p <= $::pgcs($vts,$pgc,numprograms)} {incr p} {
			set cell_program([get_int programmap [expr {$p-1}] 1]) $p
		}
		set old_pg 1
		for {set c 1} {$c <= $::pgcs($vts,$pgc,numcells)} {incr c} {
			if {[info exists cell_program($c)]} {
				set old_pg $cell_program($c)
			} else {
				set cell_program($c) $old_pg
			}
		}
		set cellpostable $::pgcs($vts,$pgc,cellpostable)
		set numcells $::pgcs($vts,$pgc,numcells)
	}
	set buttonsfound false
	set menulabels {}
	set menucommands {}
	set ::menubuttons::menutargets {}
	set cell 0
	for {set c 1} {$c <= $numcells} {incr c} {
		set idx [expr {($c-1) * 4}]
		set this_vobid [get_int cellpostable $idx 2]
		incr idx 3
		set this_cellid [get_int cellpostable $idx 1]
		if {[info exists ::buttons::curbutsinfo($this_vobid,$this_cellid,numbuts)]} {
			set this_numbuts $::buttons::curbutsinfo($this_vobid,$this_cellid,numbuts)
			lappend menulabels "Program $cell_program($c), Cell $c:  VOB ID $this_vobid, Cell ID $this_cellid ($this_numbuts buttons)"
			lappend menucommands "set ::menubuttons::vobu_offset 0; ::menubuttons::display_cell $w $type $vts $lu $pgc $c $cell_program($c) $this_vobid $this_cellid 1024 true $mode 0"
			lappend ::menubuttons::menutargets "$cell_program($c) $c $this_vobid $this_cellid"
			if {$mode == "trace" && $cell == 0} {
				set cell $c
				set vobid $this_vobid
				set cellid $this_cellid
				set numbuts $this_numbuts
			}
			if {$this_numbuts > 0} {
				set buttonsfound true
 				if {$mode != "trace" && $cell == 0} {
					set cell $c
					set vobid $this_vobid
					set cellid $this_cellid
					set numbuts $this_numbuts
				}
			}
		}
	}
	if {! $buttonsfound} {
		if {$type == "menu"} {
			show_msg "Menu buttons: This menu has no buttons."
		} else {
			show_msg "Menu buttons: There are no BOVs in this title."
		}
		if {$mode == "trace"} {
			return 0
		}
		set cell 1
		set vobid [get_int cellpostable 0 2]
		set cellid [get_int cellpostable 3 1]
		set numbuts 0
	}

	if {$show_cell >= 1} {
		set cell $show_cell
		set idx [expr {($cell-1) * 4}]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if {[info exists ::buttons::curbutsinfo($vobid,$cellid,numbuts)]} {
			set numbuts $::buttons::curbutsinfo($vobid,$cellid,numbuts)
			if {$numbuts == 0} {
				show_msg "Menu buttons: This menu has no buttons"
				if {$mode != "trace"} bell
				return 0
			}
		}
	}

	### GUI
	button .b
	set font [.b cget -font]
	destroy .b
	set ::menubuttons::minysize [expr {[font metrics $font -ascent] - [font metrics $font -descent] +4}]

	set nullimg {
		#define null_width 1
		#define null_height 1
		static unsigned char null_bits[] = {
			0x00};
	}
	image create bitmap ::menubuttons::bitmapnull -background {} -data $nullimg

	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	if {$mode == "info"} {
		wm protocol $w WM_DELETE_WINDOW {set ::menubuttons::menubuttons_done 1}
		bind $w <Escape> {set ::menubuttons::menubuttons_done 1}
	} elseif {[string range $mode 0 4] == "blank"} {
		wm protocol $w WM_DELETE_WINDOW bell
		bind $w <Escape> bell
	} else {
		wm protocol $w WM_DELETE_WINDOW {set ::tr::wait_step 0; set ::menubuttons::menubuttons_done 1}
		#wm protocol $w WM_DELETE_WINDOW bell
		bind $w <Escape> {set ::menubuttons::menubuttons_done 1}
	}
	bind $w <Return> {set ::menubuttons::menubuttons_done 1}

	frame $w.fsc
	pack  $w.fsc -fill x -padx 4 -pady 4
	if {$mode == "trace"} {
		label $w.fsc.this -bd 1 -relief sunken -height 1 -textvariable ::menubuttons::lab1_var
	} else {
		menubutton $w.fsc.this -menu $w.fsc.this.select_cell -textvariable ::menubuttons::lab1_var
		if {$::config(preview_screen_divisor) == 2.0} {
			$w.fsc.this configure -anchor w
		}
		menu $w.fsc.this.select_cell
		for {set i 0} {$i < [llength $menulabels]} {incr i} {
			set breakat 30
			if {[llength $menulabels] > 90} {set breakat 40}
			if {[llength $menulabels] > 160} {set breakat 51}
			if {[expr {$i % $breakat}] == 0 && $i != 0} {set break 1} else {set break 0}
			$w.fsc.this.select_cell add radiobutton -label [lindex $menulabels $i] \
					-command [lindex $menucommands $i] -columnbreak $break \
					-variable ::menubuttons::cell_select_menu_item -value [expr {$i+1}]
		}
		set ::menubuttons::cell_select_menu_item $cell
		if {$numcells > 1 && $::config(preview_screen_divisor) != 2.0} {
			frame $w.fsc.pnf -bd 1 -relief raised
			pack  $w.fsc.pnf -side left -pady 1 -fill y
			button $w.fsc.pnf.prev -text "<" -command {::menubuttons::select_prev_next_cell -1} \
					-bd 1 -padx 1 -pady 0 -highlightthickness 1
			pack   $w.fsc.pnf.prev -side left -fill y
			button $w.fsc.pnf.next -text ">" -command {::menubuttons::select_prev_next_cell 1} \
					-bd 1 -padx 1 -pady 0 -highlightthickness 1
			pack   $w.fsc.pnf.next -side left -fill y
		}
	}
	pack  $w.fsc.this -fill both -ipady 0

	label $w.labinfo -bd 1 -relief sunken -height 1 \
		-textvariable ::menubuttons::lab2_var
	pack  $w.labinfo -fill x -padx 4 -pady 2 -ipady 2

	pack [frame $w.videopad1] -fill y -expand 1
	pack [frame $w.video -bd 2 -relief ridge] -padx 4 -pady 0
	pack [frame $w.videopad2] -fill y -expand 1

	# used only to set window dimentions
	pack [canvas $w.video.f -width [expr {round(720 / $::config(preview_screen_divisor))}] \
			-height [expr {round(540 / $::config(preview_screen_divisor))}] -bd 0]
	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x ;## -expand 1

	if {$mode == "info"} {
		if {$::config(tk) == "aqua"} {
			label $w.f2.labcmd -bd 2 -relief raised -height 2 -textvariable ::menubuttons::command
		} else {
			button $w.f2.labcmd -height 2 -textvariable ::menubuttons::command
		}
		# -command set later
		pack  $w.f2.labcmd -fill x -padx 4 -pady 2
	} else {
		label $w.f2.labcmd -bd 2 -relief sunken -height 2 -textvariable ::menubuttons::command
		pack  $w.f2.labcmd -fill x -padx 4 -pady 2
	}

	frame $w.f2.f
	pack  $w.f2.f -fill x ;## -expand 1
	# TODO: enable the checkbox when in editor mode?
	checkbutton $w.f2.f.auto -state disabled -text "Auto Action" -variable ::menubuttons::autoaction
	pack   $w.f2.f.auto -side left
	menubutton $w.f2.f.cp -text "Clipboard" -width 16 -menu $w.f2.f.cp.menu -pady 2
	pack   $w.f2.f.cp -side right -padx 3 -pady 1
	menu $w.f2.f.cp.menu
	$w.f2.f.cp.menu add command -label "Copy to clipboard" -command {::menubuttons::copy_current_command_to_clip 1}
	$w.f2.f.cp.menu add command -label "Append to clipboard" -command {::menubuttons::copy_current_command_to_clip 2}
	$w.f2.f.cp.menu add command -label "Copy as text" -command {::menubuttons::copy_current_command_to_clip 3}
	$w.f2.f.cp.menu add separator
	$w.f2.f.cp.menu add command -label "Append to scratchbook" -command {::menubuttons::append_current_command_to_scratchbook}

	if {$mode == "info"} {
		$w.f2.f.cp.menu add separator
		$w.f2.f.cp.menu add command -label "Paste" -command "::menubuttons::paste_current_command $type $vts $vobid $cellid"
		button $w.f2.f.follow -text "Jump to Target" -width 16 -pady 0	;# -command set later
		pack   $w.f2.f.follow -side right -pady 1
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 0 -fill x ;## -expand 1
	label  $w.fb.lab3 -text "Group"
	radiobutton $w.fb.g1 -variable ::buttons::groupnum -value 1 -text "1" -command \
			"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid 1024 false $mode 0"
	radiobutton $w.fb.g2 -variable ::buttons::groupnum -value 2 -text "2" -command \
			"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid 1024 false $mode 0"
	radiobutton $w.fb.g3 -variable ::buttons::groupnum -value 3 -text "3" -command \
			"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid 1024 false $mode 0"
#	if {$mode == "info"} {
		pack   $w.fb.lab3 -side left
		pack   $w.fb.g1 -side left
		pack   $w.fb.g2 -side left
		pack   $w.fb.g3 -side left
		set butlabtext "   Button"
#	} else {
#		set butlabtext "Button"
#	}
	label  $w.fb.lab1 -text $butlabtext
	pack   $w.fb.lab1 -side left
	label  $w.fb.lab2 -textvariable ::buttons::default_button -padx 0 -width 3 -anchor w
	pack   $w.fb.lab2 -side left
	button $w.fb.prev -text "<" -width 2 -pady 0 ;# commands set later
	pack   $w.fb.prev -side left
	button $w.fb.next -text ">" -width 2 -pady 0
	pack   $w.fb.next -side left

	if {$::config(use_preview)} {
		frame  $w.fb.pad
		label  $w.fb.labf -text "Image"
		set ::menubuttons::vobu_offset 0
		button $w.fb.startfr -text "|<" -width 2 -pady 0	;# commands set later
		button $w.fb.prevfr -text "<" -width 2 -pady 0
		button $w.fb.nextfr -text ">" -width 2 -pady 0
		frame  $w.fb.pad2 -width 2
		pack   $w.fb.pad -side left -fill x -expand 1
		pack   $w.fb.labf -side left
		pack   $w.fb.startfr -side left
		pack   $w.fb.prevfr -side left
		pack   $w.fb.nextfr -side left
		pack   $w.fb.pad2 -side left
	}

	frame  $w.fb2
	pack   $w.fb2 -padx 2 -pady 4 -fill x ;## -expand 1
	if {$::config(use_preview)} {
		button $w.fb2.pv -text "Preview" -width 13
		if {$type == "menu"} {
			$w.fb2.pv configure -command \
					"::preview::preview2 menu $vts,$lu,$pgc $cell false false 0"
		} else {
			$w.fb2.pv configure -command \
					"::preview::preview2 pgc $vts,$pgc $cell false false 0"
		}
		pack   $w.fb2.pv -side left -padx 2
	}
	if {$mode == "info"} {
		button $w.fb2.edit -text "Edit" -width 13
		pack   $w.fb2.edit -side left -padx 2
	}
	frame  $w.fb2.pad
	pack   $w.fb2.pad -side left -fill x -expand 1
	if {$mode != "trace"} {
		if {$mode == "info"} {
			button $w.fb2.ok -text "Close" -width 13 -command {set ::menubuttons::menubuttons_done 1}
			pack   $w.fb2.ok -side left -padx 2
		} else {
			button $w.fb2.skip -text "No Button!" -width 13 -command {set ::menubuttons::menubuttons_done -1}
			pack   $w.fb2.skip -side left -padx 2
			::balloon::set_balloon $w.fb2.skip "Kill PGC but do not simulate a button.\nUse it ONLY for non-looping, non-infinite menus!"
			if {$mode == "blank_infinite"} {
				$w.fb2.skip configure -state disabled
			}
			button $w.fb2.ok -text "Kill playback" -width 13 -command {set ::menubuttons::menubuttons_done 1}
			pack   $w.fb2.ok -side left -padx 2
			::balloon::set_balloon $w.fb2.ok "Kill PGC playback using\nselected menu button."
			button $w.fb2.cancel -text "Cancel" -width 13 -command {set ::menubuttons::menubuttons_done 0}
			pack   $w.fb2.cancel -side left -padx 2
		}
	} else {
		button $w.fb2.skip -text "Skip" -width 13 -command {set ::menubuttons::menubuttons_done 0}
		pack   $w.fb2.skip -side left -padx 2

		button $w.fb2.ok2 -text "Select & Break" -width 13 -command {set ::tr::wait_step 0; set ::menubuttons::menubuttons_done 1} -foreground red
		pack   $w.fb2.ok2 -side left -padx 2
		::balloon::set_balloon $w.fb2.ok2 "Use the current button,\nand stop Trace's Run mode."

		button $w.fb2.ok -text "Select" -width 13 -command {set ::menubuttons::menubuttons_done 1}
		pack   $w.fb2.ok -side left -padx 2
		::balloon::set_balloon $w.fb2.ok "Use the currently\nselected button."
	}
	if {$::config(platform) == "mac"} {
		pack [frame $w.fb2.macpad] -side left -padx 6
	}

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	if {! [info exists ::menubuttons::win_geometry]} {
		wm geometry $w $::config(menubuttons_win_geometry)
	} else {
		wm geometry $w $::menubuttons::win_geometry
	}
	update
	wm maxsize $w [expr {[winfo reqwidth $w] + 300}] [winfo reqheight $w]
	set ::menubuttons::vobu_offset 0
	display_cell $w $type $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid $sprm8 true $mode 0

	if {! [winfo exists $w]} {return 0}
	wm deiconify $w
	focus $w
	if {$mode != "trace" && $::config(debug)==0} {
		grab set $w	;# TODO: prohibit some functions when in Trace mode (ie. step)
	}

	set done false
	while {! $done} {
		set ::menubuttons::menubuttons_done 0
		tkwait variable ::menubuttons::menubuttons_done
		# Don't exit if it's a NOP command!
		if {$mode == "trace" && $::menubuttons::menubuttons_done == 1 \
				&& [string first {[00 00 00 00 00 00 00 00]} $::menubuttons::command] >= 0} {
			set txt "A NOP command in a menu button is ignored."
			set txt "$txt\nPlayback continues normally."
			set txt "$txt\n\nSelect another button, or use the Skip button"
			set txt "$txt\nto simulate a normal playback, without button"
			set txt "$txt\nselection, and continue..."
			tk_messageBox -title "PgcEdit: Trace" -icon info \
					-message $txt -parent . -type ok
			catch {focus $w}
			#grab set $w
		} else {
			set done true
		}
	}

	if {$::menubuttons::menubuttons_done == 1 && $::buttons::default_button > 0} {
		catch {$w.video.f.b$::buttons::default_button configure -background green \
				-foreground black -relief sunken}
		catch {image delete ::menubuttons::preview_but($::buttons::default_button)}
		if {$mode == "trace"} {
			after 500 "set ::menubuttons::menubuttons_done 1"
			tkwait variable ::menubuttons::menubuttons_done
		}
	}
	if {$::menubuttons::menubuttons_done == 0 && $::buttons::action_button > 0 \
			&& [winfo exists $w.video.f.b$::buttons::action_button]} {
		catch {$w.video.f.b$::buttons::action_button configure -background green \
				-foreground black -relief sunken}
		catch {image delete ::menubuttons::preview_but($::buttons::action_button)}
		if {$mode == "trace"} {
			after 500 "set ::menubuttons::menubuttons_done 0"
			tkwait variable ::menubuttons::menubuttons_done
		}
	}

	catch {set ::menubuttons::win_geometry [wm geometry $w]}
	catch {destroy $w}
	catch {image delete ::menubuttons::preview_ppm}
	for {set i 1} {$i <= 32} {incr i} {
		if {[catch {image delete ::menubuttons::preview_but($i)}]} break
	}
	if {$mode == "trace"} {
		set ::tr::sprm(8) [expr {$::buttons::default_button * 1024}]
	}
	if {$::menubuttons::menubuttons_done == 1} {
		return $::buttons::default_button
	} elseif {$::menubuttons::menubuttons_done == 0} {
		return $::buttons::action_button
	} elseif {$::menubuttons::menubuttons_done < 0} {
		return -1
	} else {
		return 0
	}
}

proc select_prev_next_cell {direction} {
	set newcell [expr $::menubuttons::cell_select_menu_item + $direction]
	if {$newcell < 1 || $newcell > [.menubuttons.fsc.this.select_cell index end]} {bell; return}
	eval [.menubuttons.fsc.this.select_cell entrycget $newcell -command]
	set ::menubuttons::cell_select_menu_item $newcell
}

proc display_cell {w type vts lu pgc cell program vobid cellid sprm8 initbutton mode vobu_direction} {
	$w.f2.labcmd configure -background $::config(systembuttonface)
	if {$mode == "info"} {
		$w.fb2.edit configure -command "::menued::launch_menued $type $vts $lu $pgc $cell $vobid $cellid \$::buttons::default_button \$::buttons::groupnum"
		$w.f2.f.cp.menu entryconfigure end -command "::menubuttons::paste_current_command $type $vts $vobid $cellid"
	}
	if {$::config(use_preview)} {
		if {$type == "menu"} {
			$w.fb2.pv configure -command \
					"::preview::preview2 menu $vts,$lu,$pgc $cell false false 0"
		} else {
			$w.fb2.pv configure -command \
					"::preview::preview2 pgc $vts,$pgc $cell false false 0"
		}
		$w.fb.startfr configure -command \
				"set ::menubuttons::vobu_offset 0; ::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 false $mode 0"
		$w.fb.prevfr configure -command \
				"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 false $mode -1"
		bind $w.fb.prevfr <3> "::menubuttons::new_background_image $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 false $mode -5 $w.fb.prevfr"
		$w.fb.nextfr configure -command \
				"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 false $mode 1"
		bind $w.fb.nextfr <3> "::menubuttons::new_background_image $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 false $mode 5 $w.fb.nextfr"
	}
	if {[info exists ::buttons::butsinfo($type,$vts,$vobid,$cellid,edited)] &&
			$::buttons::butsinfo($type,$vts,$vobid,$cellid,edited)} {
		wm title .menubuttons "Menu buttons  *"
	} else {
		wm title .menubuttons "Menu buttons"
	}

	if {![info exists ::buttons::curbutsinfo($vobid,$cellid,numbuts)]} {
		set txt "Error:"
		set txt "$txt\nUnable to find cell wirh Vob/Cell ID $vobid/$cellid in VOB file."
		set txt "$txt\nThe VOB file don't match the IFO file."
		tk_messageBox -title "Menu Buttons Viewer/Editor" -icon error \
			-message $txt -parent .
		return 0
	}
	set numbuts $::buttons::curbutsinfo($vobid,$cellid,numbuts)
	if {$type == "menu"} {
		if {$vts == 0} {set vtslabel "VMGM"} else {set vtslabel "VTSM $vts"}
		set bcd [string range $::menus($vts,$lu,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 4}] [expr {(($cell - 1) * 24) + 7}]]
	} else {
		set vtslabel "VTST $vts"
		set bcd [string range $::pgcs($vts,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 4}] [expr {(($cell - 1) * 24) + 7}]]
	}
	set cell_time [::pgced::BCDtime $bcd false]
	if {[string range $cell_time 0 3] == "00:0"} {
		set cell_time [string range $cell_time 4 end]
	} elseif {[string range $cell_time 0 2] == "00:"} {
		set cell_time [string range $cell_time 3 end]
	} elseif {[string range $cell_time 0 0] == "0"} {
		set cell_time [string range $cell_time 1 end]
	}

	set scrndim [get_screen_dims $type $vts]
	set xdim [lindex $scrndim 0]
	set ydim [lindex $scrndim 1]
	set ispal [lindex $scrndim 2]
	set aspect169 [lindex $scrndim 3]

	set ::menubuttons::panscan false
	set ::menubuttons::letterbox false

	if {[info exists ::buttons::curbutsinfo($vobid,$cellid,$::buttons::groupnum,grouptype)]} {
		set group_descr [grouptype_descr $::buttons::curbutsinfo($vobid,$cellid,$::buttons::groupnum,grouptype)]
		if {$aspect169} {
			set height 405.0
			if {[string first "normal" $group_descr]<0} {
				if {[string first "pan&scan" $group_descr]>=0 && [string first "letterbox" $group_descr]>=0} {
					set videomodes [lindex [::streams::video_attributes [::streams::get_video_attribute_value]] 6]
					if {[string first "Pan&Scan" $videomodes]>=0} {
						set height 540.0
						set ::menubuttons::panscan true
					}
					if {[string first "Letterbox" $videomodes]>=0} {
						set height 540.0
						set ::menubuttons::letterbox true
						set ::menubuttons::image_divisor_y [expr {$::config(preview_screen_divisor) * ($ydim/405.0)}]
					}
				} else {
					if {[string first "pan&scan" $group_descr]>=0} {
						set height 540.0
						set ::menubuttons::panscan true
					}
					if {[string first "letterbox" $group_descr]>=0} {
						set height 540.0
						set ::menubuttons::letterbox true
						set ::menubuttons::image_divisor_y [expr {$::config(preview_screen_divisor) * ($ydim/405.0)}]
					}
				}
			}
		} else {
			set height 540.0
		}
	} else {
		if {$aspect169} {
			set height 405.0
		} else {
			set height 540.0
		}
	}

	set ::menubuttons::resize_divisor_x [expr {$::config(preview_screen_divisor) * ($xdim/720.0)}]
	set ::menubuttons::resize_divisor_y [expr {$::config(preview_screen_divisor) * ($ydim/$height)}]
	if {$ispal} {
		set ::menubuttons::buttons_divisor_y [expr {$::config(preview_screen_divisor) * (576/$height)}]
	} else {
		set ::menubuttons::buttons_divisor_y [expr {$::config(preview_screen_divisor) * (480/$height)}]
	}

	#set ::menubuttons::lab1_var "$vtslabel / LU $lu / PGC $pgc, PG $program, Cell $cell: VOB ID $vobid, Cell ID $cellid, $cell_time (NAV LBA=$::buttons::curbutsinfo($vobid,$cellid,navpack))"
	set ::menubuttons::lab1_var "$vtslabel / LU $lu / PGC $pgc, PG $program, Cell $cell: VOB ID $vobid, Cell ID $cellid, $cell_time"
	::balloon::set_balloon $w.fsc.this "Cell's first NAV pack (or first NAV pack with buttons) LBA in VOB: $::buttons::curbutsinfo($vobid,$cellid,navpack)"

	$w.fb.g1 configure -command "::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid 1024 false $mode 0"

	if {$numbuts == 0} {
		if {[string range $mode 0 4] == "blank"} {
			$w.fb2.ok configure -state disabled
			bind $w <Return> bell
		}
		set ::menubuttons::lab2_var "No buttons"

		destroy $w.video.f
		pack [canvas $w.video.f -width [expr {round(720 / $::config(preview_screen_divisor))}] \
				-height [expr {round($height / $::config(preview_screen_divisor))}] -background gray -bd 0]
		display_preview $w.video.f $type $vts $lu $pgc $cell $vobid $cellid 0 $vobu_direction
		set ::menubuttons::autoaction 0
		$w.f2.f.auto configure -disabledforeground [$w.fb.lab3 cget -disabledforeground]
		$w.fb.lab3 configure -state disabled
		$w.fb.g1 configure -state disabled
		$w.fb.g2 configure -state disabled
		$w.fb.g3 configure -state disabled
		$w.fb.prev configure -state disabled
		$w.fb.next configure -state disabled
		$w.fb.lab1 configure -state disabled
		$w.fb.lab2 configure -state disabled
		set ::buttons::default_button 0
		set ::menubuttons::command ""

		if {$::config(tk) != "aqua"} {
			catch {$w.f2.labcmd configure -command bell -relief sunken}
		} else {
			catch {$w.f2.labcmd configure -relief sunken}
			bind $w.f2.labcmd <1> "bell"
		}
		return
	} else {
		if {[string range $mode 0 4] == "blank"} {
			$w.fb2.ok configure -state normal
			bind $w <Return> {set ::menubuttons::menubuttons_done 1}
		}
		$w.fb.lab1 configure -state normal
		$w.fb.lab2 configure -state normal
	}
	if {$::buttons::curbutsinfo($vobid,$cellid,numgroups) < 3} {
		$w.fb.g3 configure -state disabled
		if {$::buttons::groupnum >= 3} {set ::buttons::groupnum 2}
	} else {
		$w.fb.g3 configure -state normal -command \
			"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid 1024 false $mode 0"
	}
	if {$::buttons::curbutsinfo($vobid,$cellid,numgroups) < 2} {
		$w.fb.g2 configure -state disabled
		if {$::buttons::groupnum == 2} {set ::buttons::groupnum 1}
	} else {
		$w.fb.g2 configure -state normal -command \
			"::menubuttons::display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid 1024 false $mode 0"
	}

	$w.f2.f.auto configure -disabledforeground black
	$w.fb.lab3 configure -state normal
	$w.fb.g1 configure -state normal
	$w.fb.prev configure -state normal
	$w.fb.next configure -state normal

	set ::buttons::action_button $::buttons::curbutsinfo($vobid,$cellid,forceactionbutn)
	set ::menubuttons::lab2_var "Group $::buttons::groupnum ($group_descr), $numbuts buttons, force select button $::buttons::curbutsinfo($vobid,$cellid,forceselectbutn), force action button $::buttons::action_button"

	if {$mode == "trace"} {
		$w.fb.g1 configure -state disabled
		$w.fb.g2 configure -state disabled
		$w.fb.g3 configure -state disabled
	}
	$w.fb.prev configure -command \
		"::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid \[expr {\$::buttons::default_button - 1}\] true true $mode"
	$w.fb.next configure -command \
		"::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid \[expr {\$::buttons::default_button + 1}\] true true $mode"
	if {$mode == "info"} {
		$w.f2.f.follow configure -command \
			"::menubuttons::jump_to_target_inside_PGC $w $type $vts $lu $pgc $cell $program \$::menubuttons::command info -1"
	} elseif {$mode == "trace"} {
		if {$type == "menu"} {
			set still_time [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 2}] 1]
		} else {
			set still_time [get_int ::pgcs($vts,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 2}] 1]
		}
		if {$still_time < 255} {
			$w.fb2.skip configure -state normal
			set txt "Wait to end of cell's playback\nwithout selecting any button.\n"
			if {$::buttons::action_button > 0} {
				set txt "${txt}(Will use action button $::buttons::action_button)"
			} else {
				set txt "${txt}(Will execute cell/post commands)"
			}
			::balloon::set_balloon $w.fb2.skip $txt

		} else {
			$w.fb2.skip configure -state disabled
			::balloon::set_balloon $w.fb2.skip "This cell's Still Time is infinite.\nYou must select a button."
		}
	}

	destroy $w.video.f
	pack [canvas $w.video.f -width [expr {round(720 / $::config(preview_screen_divisor))}] \
			-height [expr {round($height / $::config(preview_screen_divisor))}] -background gray -bd 0]
	display_preview $w.video.f $type $vts $lu $pgc $cell $vobid $cellid $numbuts $vobu_direction

	if {$::config(use_preview)} {
		bind $w.video.f <ButtonPress-1> \
				"for {set but $numbuts} {\$but > 0} {incr but -1} {destroy $w.video.f.b\$but}"
		bind $w.video.f <ButtonRelease-1> \
				"for {set but $numbuts} {\$but > 0} {incr but -1} \
				{::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid \$but false false $mode}; ::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid \$::buttons::default_button true false $mode"
	}
	update
	for {set but $numbuts} {$but > 0} {incr but -1} {
		display_button $w $type $vts $lu $pgc $cell $vobid $cellid $but false false $mode
	}
	if {$initbutton} {
		if {$::buttons::curbutsinfo($vobid,$cellid,forceselectbutn) != 0 && $sprm8 >= 0} {
			if {$::buttons::curbutsinfo($vobid,$cellid,forceselectbutn) <= \
					$::buttons::curbutsinfo($vobid,$cellid,numbuts)} {
				set ::buttons::default_button $::buttons::curbutsinfo($vobid,$cellid,forceselectbutn)
			} else {
				set txt "Warning: The Force select button number is out of range!\n"
				set txt "Value (in the VOB file): $::buttons::curbutsinfo($vobid,$cellid,forceselectbutn)\n"
				set txt "Valid range for this menu: 1 to $::buttons::curbutsinfo($vobid,$cellid,numbuts) (or 0 to use SPRM 8 instead)\n\n"
				set txt "${txt}The button 1 will be selected by default."
				tk_messageBox -title "PgcEdit: Menu Buttons" -icon warning \
						-message $txt -parent $w -type ok
				set ::buttons::default_button 1
			}
		} elseif {$sprm8 >= 1024 || $sprm8 <= 1024} {
			if {$sprm8 < 0} {set sprm8 [expr $sprm8 * -1]}
			set button [expr {$sprm8 / 1024}]
			if {$button <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {
				set ::buttons::default_button $button
			} else {
				set txt "Warning: The SPRM 8: highlighted menu button number is out of range!\n"
				set txt "${txt}SPRM 8 button value: $sprm8 (button $button)\n"
				set txt "${txt}Valid button number range for this menu: 1 to $::buttons::curbutsinfo($vobid,$cellid,numbuts)\n\n"
				set txt "${txt}The button 1 will be selected by default."
				tk_messageBox -title "PgcEdit: Menu Buttons" -icon warning \
						-message $txt -parent $w \
						-type ok
				set ::buttons::default_button 1
			}
		} else {
			set txt "Warning: The SPRM(8:highlighted menu button) value is 0!\n"
			set txt "${txt}Valid button number range for this menu: 1 to $::buttons::curbutsinfo($vobid,$cellid,numbuts)\n\n"
			set txt "${txt}The button 1 will be selected by default."
			tk_messageBox -title "PgcEdit: Menu Buttons" -icon warning \
					-message $txt -parent $w \
					-type ok
			set ::buttons::default_button 1
		}
	}
	if {$vobu_direction == 0} {set highlight_target true} else {set highlight_target false}
	display_button $w $type $vts $lu $pgc $cell $vobid $cellid $::buttons::default_button true $highlight_target $mode

	if {[info exists ::buttons::curbutsinfo($vobid,$cellid,numsets)] && \
			$::buttons::curbutsinfo($vobid,$cellid,numsets) > 1 && \
			! [info exists ::buttons::butsinfo($type,$vts,$vobid,$cellid,warn_multiple_sets_done)]} {
		set txt "WARNING: PgcEdit has detected $::buttons::curbutsinfo($vobid,$cellid,numsets) sets of buttons in this cell."
		set txt "$txt\nCurrently, PgcEdit supports only the first set."
		set txt "$txt\nIt is possible that the button commands and/or other button"
		set txt "$txt\nparameters are different in the other sets."
		set txt "$txt\nDon't trust the button viewer in this case!"
		set txt "$txt\n\nNote:  This warning will not be shown again for this cell."
		set txt "$txt\n\nNote for expert users:"
		set txt "$txt\nSet 1 begins at LBA $::buttons::curbutsinfo($vobid,$cellid,navpack)."
		for {set set 2} {$set <= [min $::buttons::curbutsinfo($vobid,$cellid,numsets) 30]} {incr set} {
			set txt "$txt\nSet $set begins at LBA $::buttons::curbutsinfo($vobid,$cellid,$set,lba), "
			set txt "${txt}with highlight status $::buttons::curbutsinfo($vobid,$cellid,$set,hl_status)."
		}
		if {$::buttons::curbutsinfo($vobid,$cellid,numsets) > 30} {
			set txt "${txt}\n..."
		}
		tk_messageBox -title "Menu Buttons" -icon warning -message $txt -parent .menubuttons
		set ::buttons::butsinfo($type,$vts,$vobid,$cellid,warn_multiple_sets_done) 1
	}
}

proc display_button {w type vts lu pgc cell vobid cellid but default highlight_target mode} {
	if {! [winfo exists $w]} return
	set group $::buttons::groupnum
	if {$but > 0 && $but <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {
		set coords $::buttons::curbutsinfo($vobid,$cellid,$group,$but,coords)
		set x_start [expr {round([lindex $coords 0] / $::config(preview_screen_divisor))}]
		set y_start [expr {round([lindex $coords 1] / $::menubuttons::buttons_divisor_y)}]
		set x_end   [expr {round([lindex $coords 3] / $::config(preview_screen_divisor))}]
		set y_end   [expr {round([lindex $coords 4] / $::menubuttons::buttons_divisor_y)}]

		set x_size  [expr {$x_end - $x_start - 1}]
		set y_size  [expr {$y_end - $y_start - 1}]
		if  {$y_size < $::menubuttons::minysize} {
			set y_size $::menubuttons::minysize
			if {[expr {$y_start + $::menubuttons::minysize + 4}] > [winfo reqheight $w.video.f]} {
				set y_start [expr {[winfo reqheight $w.video.f] - ($y_size+4)}]
			}
		}

		if {[winfo exists $w.video.f.b$but]} {destroy $w.video.f.b$but}
		if {[catch {image inuse ::menubuttons::preview_ppm}] || \
					![image inuse ::menubuttons::preview_ppm]} {
			button $w.video.f.b$but -image ::menubuttons::bitmapnull -bd 1 -padx 0 -pady 0 \
					-compound center -highlightthickness 0 -relief raised -text $but \
					-width $x_size -height $y_size -command "::menubuttons::display_button \
						$w $type $vts $lu $pgc $cell $vobid $cellid $but true true $mode"
			if {$::buttons::curbutsinfo($vobid,$cellid,$group,$but,autoaction)} {
				$w.video.f.b$but configure -relief sunken
			}
		} else {
			image create photo ::menubuttons::preview_but($but) -width $x_size -height $y_size

			set x1 [expr {$x_start+1}]
			if {$x1 >= [image width  ::menubuttons::preview_ppm]} {set x1 [expr {[image width  ::menubuttons::preview_ppm]-1}]}
			set y1 [expr {$y_start+1}]
			if {$y1 >= [image height ::menubuttons::preview_ppm]} {set y1 [expr {[image height ::menubuttons::preview_ppm]-1}]}
			set x2 [expr {$x_start+$x_size+3}]
			if {$x2 >= [image width  ::menubuttons::preview_ppm]} {set x2 [expr {[image width  ::menubuttons::preview_ppm]-1}]}
			set y2 [expr {$y_start+$y_size+3}]
			if {$y2 >= [image height ::menubuttons::preview_ppm]} {set y2 [expr {[image height ::menubuttons::preview_ppm]-1}]}
			::menubuttons::preview_but($but) copy ::menubuttons::preview_ppm -from $x1 $y1 $x2 $y2
			button $w.video.f.b$but -image ::menubuttons::preview_but($but) -bd 1 -padx 0 -pady 0 \
					-compound center -highlightthickness 0 -relief flat -text $but \
					-width $x_size -height $y_size -command "::menubuttons::display_button \
						$w $type $vts $lu $pgc $cell $vobid $cellid $but true true $mode"
		}
		::balloon::set_balloon $w.video.f.b$but \
				"Button $but\n[::disasm::disasm $::buttons::curbutsinfo($vobid,$cellid,$group,$but,command)]"

		# try to avoid a Tk bug with the 'after 100' delay
		if {$mode == "trace"} {
			bind $w.video.f.b$but <2>			"::menubuttons::display_button $w $type $vts $lu $pgc $cell \
					$vobid $cellid $but true true $mode; after 100 {set ::menubuttons::menubuttons_done 1}"
			bind $w.video.f.b$but <Control-1>	"::menubuttons::display_button $w $type $vts $lu $pgc $cell \
					$vobid $cellid $but true true $mode; after 100 {set ::menubuttons::menubuttons_done 1}"
		} elseif {$mode == "info"} {
			bind $w.video.f.b$but <2>         "::menubuttons::display_button $w $type $vts $lu $pgc $cell \
					$vobid $cellid $but true true $mode; eval \[$w.f2.f.follow cget -command\]" ;# jump to target
			bind $w.video.f.b$but <Control-1> "::menubuttons::display_button $w $type $vts $lu $pgc $cell \
					$vobid $cellid $but true true $mode; eval \[$w.f2.f.follow cget -command\]"
			bind $w.video.f.b$but <3> "::menued::launch_menued $type $vts $lu $pgc $cell $vobid $cellid $but \$::buttons::groupnum"
		}

		if {$default} {
			set ::buttons::default_button $but
			if {[catch {image inuse ::menubuttons::preview_ppm}] || \
					![image inuse ::menubuttons::preview_ppm]} {
				for {set i 1} {$i <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {incr i} {
					$w.video.f.b$i configure -background [$w.video cget -background] -foreground red
				}
				$w.video.f.b$but configure -background lightgreen -foreground black
			} else {
				for {set i 1} {$i <= $::buttons::curbutsinfo($vobid,$cellid,numbuts)} {incr i} {
					$w.video.f.b$i configure -background red -foreground $::config(system_window_color)
				}
				$w.video.f.b$but configure -background green -foreground $::config(system_window_color)
			}
			set ::menubuttons::autoaction $::buttons::curbutsinfo($vobid,$cellid,$group,$but,autoaction)
			set ::menubuttons::command "[string toupper \[$::buttons::curbutsinfo($vobid,$cellid,$group,$but,command)\]]\n[::disasm::disasm $::buttons::curbutsinfo($vobid,$cellid,$group,$but,command)]"
			::menued::validate_command $::buttons::curbutsinfo($vobid,$cellid,$group,$but,command) .menubuttons.f2.labcmd
			if {$::config(tk) != "aqua"} {
				catch {$w.f2.labcmd configure -command "::menued::edit_menu_button .menubuttons $type $vts $vobid $cellid $group $but" -relief raised}
			} else {
				catch {$w.f2.labcmd configure -relief raised}
				bind $w.f2.labcmd <1> "::menued::edit_menu_button .menubuttons $type $vts $vobid $cellid $group $but"
			}

			set udlr $::buttons::curbutsinfo($vobid,$cellid,$group,$but,udlr)
			bind $w <Up>    "::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid [lindex $udlr 0] true true $mode"
			bind $w <Down>  "::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid [lindex $udlr 1] true true $mode"
			bind $w <Left>  "::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid [lindex $udlr 2] true true $mode"
			bind $w <Right> "::menubuttons::display_button $w $type $vts $lu $pgc $cell $vobid $cellid [lindex $udlr 3] true true $mode"

			if {$highlight_target} {
				if {$type == "menu"} {
					::main::highlight_jump_cmd menu $vts,$lu,$pgc $::menubuttons::command -1 false true
				} else {
					::main::highlight_jump_cmd pgc $vts,$pgc $::menubuttons::command -1 false true
				}
			}
		}
		place $w.video.f.b$but -x [expr {$x_start+2}] -y [expr {$y_start+2}]
	}
}


proc display_preview {w type vts lu pgc cell vobid cellid numbuts vobu_direction} {
	if {!$::config(use_preview)} return
	if {$vobid==0 || $cellid==0} return
	if {! [info exists ::buttons::curbutsinfo($vobid,$cellid,navpack)]} return	;# TODO: Afficher quand mme une frame?
	set lba $::buttons::curbutsinfo($vobid,$cellid,navpack)
	set old_vobu_offset $::menubuttons::vobu_offset
	incr ::menubuttons::vobu_offset $vobu_direction
	if {$::menubuttons::vobu_offset != 0} {
		set lbabin [d2c $lba 4]
		if {$type == "menu"} {
			if {$vts == 0} {
				set vobutable [string range $::raw_sectors($vts,VMGM_VOBU_ADMAP) 4 end]
			} else {
				set vobutable [string range $::raw_sectors($vts,VTSM_VOBU_ADMAP) 4 end]
			}
		} else {
			set vobutable [string range $::raw_sectors($vts,VTS_VOBU_ADMAP) 4 end]
		}
		set vobutablepos -2
		while {[expr {$vobutablepos%4}] != 0} {
			set vobutablepos [string first $lbabin $vobutable $vobutablepos]
			if {$vobutablepos == -1} break
		}
		if {$vobutablepos != -1} {
			set newidx [expr {$vobutablepos + ($::menubuttons::vobu_offset * 4)}]
			if {$newidx >= 0 && $newidx < [string length $vobutable]} {
				set lba [get_int vobutable $newidx 4]
			} else {
				bell
				set ::menubuttons::vobu_offset $old_vobu_offset
				catch {$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw}
				return
			}
		}
	}
	if {$type == "menu"} {
		set first_vobu_start [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr {(($cell - 1) * 24) +  8}] 4]
		set last_vobu_start  [get_int ::menus($vts,$lu,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 16}] 4]
	} else {
		set first_vobu_start [get_int ::pgcs($vts,$pgc,cellpbtable) [expr {(($cell - 1) * 24) +  8}] 4]
		set last_vobu_start  [get_int ::pgcs($vts,$pgc,cellpbtable) [expr {(($cell - 1) * 24) + 16}] 4]
	}
	if {$lba < $first_vobu_start || $lba > $last_vobu_start} {
		show_msg "LBA mismatch in VOB/IFO: Menu buttons LBA is $lba while cell is between LBAs $first_vobu_start and $last_vobu_start!"
		bell
		set ::menubuttons::vobu_offset $old_vobu_offset
		catch {$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw}
		return
	}
	if {$type == "menu"} {
		if {$vts == 0} {
			set vobfile  [file nativename [file join $::dvddir "VIDEO_TS.VOB"]]
		} else {
			set vobfile  [file nativename [file join $::dvddir "VTS_[format %02d $vts]_0.VOB"]]
		}
	} else {
		set vobfile  [file nativename [file join $::dvddir "VTS_[format %02d $vts]_1.VOB"]]
	}
	if {! [file exists $vobfile]} return
	set progname [file nativename [file join $::config(bindir) "PgcEditPreview.exe"]]
	set tempfile [file nativename [file join $::config(tempdir) "pgcedit_preview_[pid].ppm"]]
	catch {file delete $tempfile}
	set divisor_y $::menubuttons::resize_divisor_y
	if {$::menubuttons::panscan} {
		set divisor_x [expr {$::menubuttons::resize_divisor_x * 3.0 / 4}]
	} else {
		set divisor_x $::menubuttons::resize_divisor_x
	}
	if {$::menubuttons::letterbox && ! $::menubuttons::panscan} {
		set divisor_y [expr {$::menubuttons::image_divisor_y}]
	}
	# PgcEdit_preview 1.7 command line:
	# PgcEdit_preview "VOBfile.VOB" LBA VOB_ID Cell_ID Number_of_buttons X_resize_factor \
	#     Y_resize_factor bool_save_as_PPM_instead_of_BMP "outfile.[bmp|ppm]"
	# Where VOB_ID Cell_ID Number_of_buttons are used to check if the right nav pack
	# is passed in argument.  -1 disables the check for that particular option.
	if {$::config(debug)>0} {
		puts "> $::config(preview_launcher) \"$progname\" \"$vobfile\" $lba $vobid $cellid -1 $divisor_x $divisor_y 1 \"$tempfile\""
	}
	if {$::config(preview_launcher) == ""} {
		set rc [catch {exec -- $progname "$vobfile" $lba $vobid $cellid -1 $divisor_x $divisor_y 1 "$tempfile"} errstr]
	} else {
		set rc [catch {exec -- $::config(preview_launcher) $progname "$vobfile" $lba $vobid $cellid -1 $divisor_x $divisor_y 1 "$tempfile" 2>/dev/null} errstr]
	}
	if {$rc} {
		bell
		set ::menubuttons::vobu_offset $old_vobu_offset
		catch {$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw}
		return
	}
	if {![file exists $tempfile]} {
		bell
		set ::menubuttons::vobu_offset $old_vobu_offset
		catch {$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw}
		return
	}

	if {$::menubuttons::panscan} {
		set height        [expr {round(540 / $::config(preview_screen_divisor))}]
		set visiblewidth  [expr {round(720 / $::menubuttons::resize_divisor_x)}]
		set imagewidth    [expr {round(720 / $divisor_x)}]
		set border [expr {($imagewidth-$visiblewidth)/2}]
		if {! [catch {image create photo ::menubuttons::tmp_ppm -file $tempfile}]} {
			if {! [catch {image create photo ::menubuttons::preview_ppm -width $visiblewidth -height $height}]} {
				::menubuttons::preview_ppm copy ::menubuttons::tmp_ppm -from $border 0 [expr {$border+$visiblewidth}] $height
				image delete ::menubuttons::tmp_ppm
				$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw
			}
		}
	} elseif {$::menubuttons::letterbox} {
		# one pixel blank image, used to fill the background of 16:9 letterboxed menus.
		set fname [file join $::config(tempdir) black1pixel_[pid].ppm]
		if {![file exists $fname]} {
			saveFile $fname binary "P6 1 1 255\n\040\040\040"
		}
		image create photo ::menubuttons::black1pixel -format ppm -file $fname
		if {! [catch {image create photo ::menubuttons::tmp_ppm -file $tempfile}]} {
			set visibleheight [expr {round(540 / $::config(preview_screen_divisor))}]
			set imageheight   [image height ::menubuttons::tmp_ppm]
			set width         [expr {round(720 / $::config(preview_screen_divisor))}]
			set border [expr {($visibleheight-$imageheight)/2}]
			if {! [catch {image create photo ::menubuttons::preview_ppm -width $width -height $visibleheight}]} {
				::menubuttons::preview_ppm copy ::menubuttons::black1pixel -zoom $width $visibleheight
				::menubuttons::preview_ppm copy ::menubuttons::tmp_ppm -to 0 $border
				image delete ::menubuttons::tmp_ppm
				$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw
			}
		}
		catch {file delete $fname}
	} else {
		if {! [catch {image create photo ::menubuttons::preview_ppm -file $tempfile}]} {
			$w create image 2 2 -image ::menubuttons::preview_ppm -anchor nw
		}
	}
	catch {file delete $tempfile}
}

proc new_background_image {w type vts lu pgc cell program vobid cellid sprm8 initbutton mode vobu_direction widget} {
	$widget configure -relief sunken
	update idletasks
	after 5 "set ::pause true"
	tkwait variable ::pause
	display_cell $w $type $vts $lu $pgc $cell $program $vobid $cellid $sprm8 $initbutton $mode $vobu_direction
	$widget configure -relief raised
}

proc get_screen_dims {type vtsnum} {
	if {$type == "menu"} {
		if {$vtsnum == 0} {
			set mat_table $::raw_sectors(0,VMGM_MAT)
			set vattr [get_int mat_table 0x100 2]
		} else {
			set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
			set vattr [get_int mat_table 0x100 2]
		}
	} else {
		set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
		set vattr [get_int mat_table 0x200 2]
	}
	if {[expr {$vattr & 0X1000}] == 0} {
		set h 480
	} else {
		set h 576
	}
	switch [expr {($vattr & 0X0018) / 8}] {
		0 {set w 720}
		1 {set w 704}
		2 {set w 352}
		3 {set w 352; set h [expr {$h / 2}]}
	}
	set ispal [expr {($vattr & 0x1000) / 0x1000}]
	set aspect [expr {$vattr & 0x0C00}]
	if {$aspect == [expr {(0x0C00)}]} {set aspect 1} else {set aspect 0}
	return [list $w $h $ispal $aspect]
}

proc copy_current_command_to_clip {mode} {
	if {$::menubuttons::command == ""} {
		bell
		return
	}
	switch $mode {
		1 {
			set cmd [list [string range $::menubuttons::command 1 23]]
		}
		2 {
			set old [clipboard get]
			set cmd [lappend old [string range $::menubuttons::command 1 23]]
		}
		3 {
			set cmd [string map {"\n" "  "} $::menubuttons::command]
		}
	}
	clipboard clear
	clipboard append $cmd
}

proc paste_current_command {type vts vobid cellid} {
	if {[catch {set str [clipboard get]}]} {
		bell
		return ""
	}
	set hex [::ced::extract_hexcommand $str]
	if {$hex == ""} {
		bell
	} elseif {[::menued::confirm_edit_several_sets $type $vts $vobid $cellid .menubuttons]} {
		set group $::buttons::groupnum
		set but   $::buttons::default_button
		if {$::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups) > 1} {
			set equal true
			set cmd1 $::buttons::butsinfo($type,$vts,$vobid,$cellid,1,$but,command)
			for {set g 2} {$g <= $::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups)} {incr g} {
				if {$::buttons::butsinfo($type,$vts,$vobid,$cellid,$g,$but,command) != $cmd1} {
					set equal false
				}
			}
			if {$equal} {
				set r yes
			} else {
				set txt "Modify the command for the $::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups) groups?"
				if {$equal} {
					set default yes
					set txt "$txt\nRecommended: The commands in all groups are identical."
				} else {
					set default no
					set txt "$txt\nWarning: THE COMMANDS IN ALL GROUPS ARE DIFFERENT!"
				}
				set r [tk_messageBox -title "Menu Button Editor" -icon question \
						-message $txt -parent . -type yesnocancel -default $default]
			}
			if {$r == "yes"} {
				set startgroup 1
				set endgroup $::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups)
			} elseif {$r == "no"} {
				set startgroup $group
				set endgroup   $group
			} else {
				return
			}
		} else {
			set startgroup 1
			set endgroup   1
		}
		for {set g $startgroup} {$g <= $endgroup} {incr g} {
			set ::buttons::butsinfo($type,$vts,$vobid,$cellid,$g,$but,command) $hex
			set ::buttons::curbutsinfo($vobid,$cellid,$g,$but,command) $hex
		}
		set ::menubuttons::command "\[$hex\]\n[::disasm::disasm $hex]"
		::menued::validate_command $hex .menubuttons.f2.labcmd
		wm title .menubuttons "Menu buttons  *"

		set ::buttons::butsinfo($type,edited) 1
		set ::buttons::butsinfo($type,$vts,edited) 1
		set ::buttons::butsinfo($type,$vts,$vobid,edited) 1
		set ::buttons::butsinfo($type,$vts,$vobid,$cellid,edited) 1
		::main::dvd_modified true true

		set index [lindex $::editor_cmds_from 1]
		::main::highlight_jump_cmd $type $index \[$hex\] 0 false true
	}
}

proc append_current_command_to_scratchbook {} {
	if {$::menubuttons::command == ""} {
		bell
		return
	}
	lappend ::scratchbook_buffer [string range $::menubuttons::command 1 23]
}

proc jump_to_target_inside_PGC {w type vts lu pgc cell pg cmd mode linenum} {
	if {$::menubuttons::command == ""} {
		bell
		return 0
	}
	set cmd [string range $cmd 1 23]
	set jumpfound 0
	set targetptt  -1
	set targetpg   -1
	set targetcell -1
	# Link commands
	set cmdchk [lindex $cmd 0][string range [lindex $cmd 1] 1 1]
	switch $cmdchk {
		"201" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr {$cell + 1}]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr {$cell - 1}]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr {$pg + 1}]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr {$pg - 1}]
				}
			}
		}
		"205" {
			# LinkPTT
			set targetptt [expr {[x2d [string range $cmd 20 20][string range $cmd 22 23]] & 0x03FF}]
			if {[info exists ::pgcs($vts,$pgc,pttlist)]} {
				foreach j $::pgcs($vts,$pgc,pttlist) {
					if {[lindex $j 1] == $targetptt} {
						set targetpg [lindex $j 0]
						break
					}
				}
			}
		}
		"206" {
			# LinkPGN
			set targetpg [x2d [lindex $cmd 7]]
		}
		"207" {
			# LinkCN
			set targetcell [x2d [lindex $cmd 7]]
		}
	}
	# Set command with Link subcommand
	set cmdchk [string range [lindex $cmd 0] 0 0][string range [lindex $cmd 1] 1 1]
	switch $cmdchk {
		"61" - "71" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr {$cell + 1}]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr {$cell - 1}]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr {$pg + 1}]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr {$pg - 1}]
				}
			}
		}
		"66" - "76" {
			# LinkPGN
			set targetpg [x2d [lindex $cmd 7]]
		}
		"67" - "77" {
			# LinkCN
			set targetcell [x2d [lindex $cmd 7]]
		}
	}
	# SetCLnk, CSetCLnk, CSetLnk commands
	set cmdchk [string range [lindex $cmd 0] 0 0]
	switch $cmdchk {
		"8" - "9" - "A" - "B" - "C" - "D" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr {$cell + 1}]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr {$cell - 1}]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr {$pg + 1}]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr {$pg - 1}]
				}
			}
		}
	}
	if {$targetpg != -1} {
		set hl_btnn [expr {([x2d [lindex $cmd 6][lindex $cmd 7]] / 1024) * 1024}]
		if {$hl_btnn == 0} {set hl_btnn 1024}
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetpg == [lindex $entry 0]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item [expr {$i + 1}]
				set ::menubuttons::vobu_offset 0
				display_cell $w $type $vts $lu $pgc [lindex $entry 1] $targetpg [lindex $entry 2] [lindex $entry 3] $hl_btnn true $mode 0
				break
			}
		}
		if {! $jumpfound} bell
	}
	if {$targetcell != -1} {
		set hl_btnn [expr {([x2d [lindex $cmd 6][lindex $cmd 7]] / 1024) * 1024}]
		if {$hl_btnn == 0} {set hl_btnn 1024}
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetcell == [lindex $entry 1]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item [expr {$i + 1}]
				set ::menubuttons::vobu_offset 0
				display_cell $w $type $vts $lu $pgc $targetcell [lindex $entry 0] [lindex $entry 2] [lindex $entry 3] $hl_btnn true $mode 0
				break
			}
		}
		if {! $jumpfound} bell
	}

	if {! $jumpfound} {
		if {$type == "menu"} {
			set h [::main::highlight_jump_cmd menu $vts,$lu,$pgc \[$cmd\] $linenum true false]
		} else {
			set h [::main::highlight_jump_cmd pgc $vts,$pgc \[$cmd\] $linenum true false]
		}
		if {$h == "" || ($h != "" && ![lindex $h 4])} {
			set jumpfound 0
		} else {
			set jumpfound 1
		}
		if {$jumpfound} {
			set ::menubuttons::menubuttons_done 1
		}
	}
	return $jumpfound
}

proc save_buts {{filename ""}} {
	if {! [info exists ::dvddir]} {bell; return 1}
	if {(![info exists ::menued::scan_done] || !$::menued::scan_done) && \
			![info exists ::warning_scan_bov] && $filename==""} {
		set r [tk_messageBox -type yesnocancel -default yes -message \
			"The DVD has not been scanned for BOVs yet\nDo you want to scan it now?  (recommended)" \
			-title "PgcEdit: Dump Buttons/BOVs" -icon question -parent .]
		if {$r == "cancel"} {
			return -1
		}
		if {$r == "yes"} {
			set ::config(check_multiple_BOVs_sets) true
			::menued::get_buttons_all_title_domains $::dvddir true "Dump Buttons/BOVs" true
		}
	}

	if {$filename == ""} {
		if {[file isdirectory $::backupdir]} {
			set defaultdir $::backupdir
		} else {
			set defaultdir $::dvddir
		}
		set filename [tk_getSaveFile -filetypes {{"PgcEdit Buttons files" {*.but}}} \
			-parent . -initialdir $defaultdir -initialfile "menubuttons.but" \
			-title "PgcEdit: Export Buttons/BOVs" -parent .]
		if {$filename == ""} {
			return -1
		}
		if {[string toupper [file extension $filename]] != ".BUT"} {
			set filename $filename.but
		}
	}

	show_msg "Saving buttons information..."
	set errnum [catch {set out [open $filename w]} err]
	if {$errnum} {
		errmsg $errnum "opening file for write" $err
		return $errnum
	}
	fconfigure $out -translation auto

	set errnum [catch {puts $out "# PgcEdit Buttons dump file"} err]
	if {!$errnum} {set errnum [catch {puts $out "# Automatically saved by PgcEdit v$::program_version"} err]}
	if {!$errnum} {set errnum [catch {puts $out "# DVD: \"$::dvddir\""} err]}
	if {!$errnum} {set errnum [catch {puts $out "# DVD-TEXT General Name: \"[::utils::get_General_Name]\""} err]}
	if {!$errnum} {set errnum [catch {puts $out "# Provider ID: \"[::utils::get_provider_ID]\""} err]}
	if {!$errnum} {set errnum [catch {puts $out "# Number of VTS: $::pgcs(numvts)"} err]}

	if {!$errnum} {set errnum [catch {puts $out "\n################ Number of menu buttons (by VTS,LU,PGC) ################"} err]}
	if {!$errnum} {
		foreach n [lsort -dictionary [array names ::menus *,numbuttons]] {
			if {!$errnum} {set errnum [catch {puts $out "set nbuts($n) $::menus($n)"} err]}
		}
	}
	if {!$errnum} {set errnum [catch {puts $out "\n################ Number of BOVs (by VTS,PGC) ################"} err]}
	if {!$errnum} {
		foreach n [lsort -dictionary [array names ::pgcs *,numbuttons]] {
			if {!$errnum} {set errnum [catch {puts $out "set nbovs($n) $::pgcs($n)"} err]}
		}
	}
	if {!$errnum} {
		set old_type ""
		set old_vts ""
		foreach n [lsort -dictionary [array names ::buttons::butsinfo *]] {
			set indexes [split $n ,]
			set new_type [lindex $indexes 0]
			set new_vts  [lindex $indexes 1]
			if {$old_type!=$new_type || $old_vts!=$new_vts} {
				if {$old_type!=$new_type} {
					if {$new_type == "menu"} {
						if {!$errnum} {set errnum [catch {puts $out "\n################ MENU BUTTONS (by VTS,VOBID,CellID) ################"} err]}
					} else {
						if {!$errnum} {set errnum [catch {puts $out "\n################ BOVS (by VTS,VOBID,CellID) ################"} err]}
					}
					if {!$errnum} {set errnum [catch {puts $out "\n######## VTS $new_vts"} err]}
				} elseif {$old_vts!=$new_vts} {
					if {!$errnum} {set errnum [catch {puts $out "\n######## VTS $new_vts"} err]}
				}
				set old_type $new_type
				set old_vts  $new_vts
			}
			if {!$errnum} {set errnum [catch {puts $out "set but($n) [list $::buttons::butsinfo($n)]"} err]}
		}
	}
	set errnum [catch {puts $out "\n################ EDITED FLAGS (by Type,VTS,VOBID,CellID) ################"} err]
	if {!$errnum} {
		set old_type ""
		set old_vts ""
		set old_vid ""
		set old_cid ""
		foreach n [lsort -dictionary [array names ::buttons::butsinfo *,numbuts]] {
			if {$::buttons::butsinfo($n) > 0} {
				set indexes [split $n ,]
				set new_type [lindex $indexes 0]
				set new_vts  [lindex $indexes 1]
				set new_vid  [lindex $indexes 2]
				set new_cid  [lindex $indexes 3]
				if {$old_type != $new_type} {
					if {!$errnum} {set errnum [catch {puts $out "\nset but($new_type,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "\nset but($new_type,$new_vts,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,$new_cid,edited) 1"} err]}
				} elseif {$old_vts != $new_vts} {
					if {!$errnum} {set errnum [catch {puts $out "\nset but($new_type,$new_vts,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,$new_cid,edited) 1"} err]}
				} elseif {$old_vid != $new_vid} {
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,edited) 1"} err]}
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,$new_cid,edited) 1"} err]}
				} elseif {$old_cid != $new_cid} {
					if {!$errnum} {set errnum [catch {puts $out "set but($new_type,$new_vts,$new_vid,$new_cid,edited) 1"} err]}
				}
				set old_type $new_type
				set old_vts  $new_vts
				set old_vid  $new_vid
				set old_cid  $new_cid
			}
		}
	}
	if {[info exists ::menued::scan_done]} {
		if {!$errnum} {set errnum [catch {puts $out "\n######## Flag: Scan for BOVs has been made ########"} err]}
		if {!$errnum} {set errnum [catch {puts $out "set scan_done $::menued::scan_done"} err]}
	}

	close $out
	if {$errnum} {
		errmsg $errnum "writing file" $err
		catch {file delete $filename}
	} else {
		show_msg ""
	}
	return $errnum
}

proc load_buts {{filename ""}} {
	if {! [info exists ::dvddir]} {bell; return 1}
	if {$filename == ""} {
		set loadanyway false
		if {[file isdirectory $::backupdir]} {
			set defaultdir $::backupdir
		} else {
			set defaultdir $::dvddir
		}
		set filename [tk_getOpenFile -filetypes \
			{{"PgcEdit Buttons files" {*.but}} {"buttoninfo.txt files" {*.txt}}} \
			-parent . -initialdir $defaultdir -initialfile "menubuttons.but" \
			-title "PgcEdit: Import Buttons/BOVs" -parent .]
		if {$filename == ""} {
			return -1
		}
	} else {
		set loadanyway true
	}

	if {[string toupper [file extension $filename]] == ".TXT"} {
		return [open_buttoninfo $filename]
	}

	if {[catch {source $filename} errmsg]} {
		tk_messageBox -message "Error importing buttons file\n$\"filename\":\n$errmsg" \
			-title "PgcEdit: Import Buttons/BOVs" -icon error -parent .
		return 1
	}

	show_msg "Importing menu buttons/BOVs from [file nativename $filename]"

	# verify if the buttons backup is compatible with the current VOB files
	set ok 1
	set fh ""
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		foreach type {menu pgc} {
			set err false
			set old_vobfile ""
			if {[info exists but($type,$vts,totalbuttons)] && $but($type,$vts,totalbuttons) != 0} {
				for {set vid 1} {$vid <= $but($type,$vts,numvobids)} {incr vid} {
					for {set cid 1} {$cid <= $but($type,$vts,$vid,numcellids)} {incr cid} {
						if {$but($type,$vts,$vid,$cid,numbuts) > 0} {
							set lba $but($type,$vts,$vid,$cid,navpack)
							set poslist [::menued::get_nav_pack_position $type $vts $lba $::dvddir]
							set vobfile [lindex $poslist 0]
							set sector  [lindex $poslist 1]
							if {$vobfile == ""} {
								if {!$loadanyway} {
									set txt "Some VOB files with buttons or BOVs are missing!"
									set txt "$txt\n\nDo you want to continue to import the buttons/BOVs anyway?"
									set txt "$txt\n(This can be useful to examine the DVD with the IFOs only."
									set txt "$txt\nHowever, you will not be able to save the buttons.)"
									set r [tk_messageBox -title "PgcEdit: Import Buttons/BOVs" -message $txt \
											-type yesnocancel -default no -icon question]
									if {$r == "yes"} {
										set loadanyway true
									}
									if {$r == "cancel"} {
										return 1
									}
								}
								set err true
								set errmsg "File not found"
								break
							}
							if {$vobfile != $old_vobfile} {
								if {![file exists $vobfile]} {
									set err true
									set errmsg "File not found"
									break
								}
								if {$fh != ""} {
									close $fh
									set old_vobfile ""
								}
								set fh [::menued::openbinfile $vobfile r]
								if {$fh == ""} {
									set err true
									set errmsg "Can't open file"
									break
								}
								set old_vobfile $vobfile
							}
							set navpack [::menued::load_nav_pack $vobfile $fh $sector]
							if {$navpack == ""} {
								set err true
								set errmsg "Nav pack not valid or LBA $lba out of range"
								break
							}
							set vobid  [get_int navpack 0x41F 2]
							set cellid [get_int navpack 0x422 1]
							if {$vid != $vobid || $cid != $cellid} {
								set err true
								set errmsg "VOB/Cell ID mismatch at LBA $lba"
								if {$type == "pgc"} {
									set errmsg "$errmsg (sector $sector in file)"
								}
								break
							}
						}
					}
					if {$err} break
				}
			}
			if {$fh != ""} {
				close $fh
				set fh ""
			}
			if {$err && !$loadanyway} {
				show_msg "Error importing buttons in VTS $vts! [file tail $vobfile]: $errmsg"
				set txt "Warning!\nThe buttons definitions of VTS $vts are not compatible with the VOB file:"
				set txt "$txt\n[file tail $vobfile]: $errmsg"
				set txt "$txt\n\nThe structure of the VOB file has probably changed since the backup."
				set txt "$txt\nIt is recommended to abort completely the import buttons/BOVs operation."
				set txt "$txt\nIf you continue,"
				if {$type == "menu"} {
					if {$vts == 0} {
						set txt "$txt the buttons will NOT be imported in the VMGM domain."
					} else {
						set txt "$txt the buttons will NOT be imported in the VTSM $vts domain."
					}
				} else {
					set txt "$txt the BOVs will NOT be imported in the VTST $vts domain."
				}
				set txt "$txt\n\nContinue anyway?"
				set r [tk_messageBox -title "PgcEdit: Import Buttons/BOVs" -message $txt \
						-type okcancel -default cancel -icon warning]
				if {$r == "cancel"} {
					return 1
				}
				foreach i [array names but "$type,$vts,*"] {
					unset -nocomplain but($i)
				}
				if {$type == "menu"} {
					foreach i [array names nbuts "$vts,*"] {
						unset -nocomplain nbuts($i)
					}
				} else {
					foreach i [array names nbovs "$vts,*"] {
						unset -nocomplain nbovs($i)
					}
				}
			}
		}
	}
	if {$fh != ""} {close $fh}
	foreach n [lsort -dictionary [array names but *]] {
		set ::buttons::butsinfo($n) $but($n)
	}
	foreach n [lsort -dictionary [array names nbuts *,numbuttons]] {
		set ::menus($n) $nbuts($n)
	}
	foreach n [lsort -dictionary [array names nbovs *,numbuttons]] {
		set ::pgcs($n) $nbovs($n)
	}
	if {[info exists scan_done]} {
		set ::menued::scan_done $scan_done
	}

	if {[llength [array names but *]] > 1} {
		::main::init_tree_menu_no_move
		if {!$loadanyway} {
			::main::dvd_modified true false
		}
		show_msg "Menu buttons/BOVs imported OK."
	}
	return 0
}

proc open_buttoninfo {filename} {
	show_msg "Parsing $filename"
	set data [readFile $filename auto]
	if {$data==""} {return -1}
	set in [split [string trim $data \n] \n]

	unset -nocomplain ::buttons::butsinfo
	set ::buttons::butsinfo(menu,multiplesets) false
	set ::buttons::butsinfo(pgc,multiplesets) false
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		foreach type {menu pgc} {
			set ::buttons::butsinfo($type,$vts,errorcode) 0
			set ::buttons::butsinfo($type,$vts,errormsg)  "Loaded from ButtonInfo.txt file."
			set ::buttons::butsinfo($type,$vts,numvobids) 0
		}
		set ::buttons::butsinfo(pgc,$vts,scanned) 1
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set ::menus($vts,$lu,$pgc,numbuttons) 0

				set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
				set cellpbtable  $::menus($vts,$lu,$pgc,cellpbtable)

				for {set cell 1} {$cell <= $::menus($vts,$lu,$pgc,numcells)} {incr cell} {
					set cellposidx [expr ($cell-1)*4]
					set vid [get_int cellpostable $cellposidx 2]
					incr cellposidx 3
					set cid [get_int cellpostable $cellposidx 1]
					if {$::buttons::butsinfo(menu,$vts,numvobids)  < $vid} {
						set ::buttons::butsinfo(menu,$vts,numvobids) $vid
					}
					if {![info exists ::buttons::butsinfo(menu,$vts,$vid,numcellids)] || \
								$::buttons::butsinfo(menu,$vts,$vid,numcellids)<$cid} {
						set ::buttons::butsinfo(menu,$vts,$vid,numcellids) $cid
					}
					set ::buttons::butsinfo(menu,$vts,$vid,$cid,navpack) \
							[get_int cellpbtable [expr (($cell-1)*24)+8] 4]
				}
			}
		}
		for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
			if {! [info exists ::buttons::butsinfo(menu,$vts,$vid,numcellids)]} {
				set ::buttons::butsinfo(menu,$vts,$vid,numcellids) 0
			}
			for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
				set ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts) 0
				set ::buttons::butsinfo(menu,$vts,$vid,$cid,numgroups) 0
				if {![info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,navpack)]} {
					set ::buttons::butsinfo(menu,$vts,$vid,$cid,navpack) -1	;# missing cell ID
				}
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set ::pgcs($vts,$pgc,numbuttons) 0

				set cellpostable $::pgcs($vts,$pgc,cellpostable)
				set cellpbtable  $::pgcs($vts,$pgc,cellpbtable)

				for {set cell 1} {$cell <= $::pgcs($vts,$pgc,numcells)} {incr cell} {
					set cellposidx [expr ($cell-1)*4]
					set vid [get_int cellpostable $cellposidx 2]
					incr cellposidx 3
					set cid [get_int cellpostable $cellposidx 1]
					if {$::buttons::butsinfo(pgc,$vts,numvobids)  < $vid} {
						set ::buttons::butsinfo(pgc,$vts,numvobids) $vid
					}
					if {![info exists ::buttons::butsinfo(pgc,$vts,$vid,numcellids)] || \
								$::buttons::butsinfo(pgc,$vts,$vid,numcellids)<$cid} {
						set ::buttons::butsinfo(pgc,$vts,$vid,numcellids) $cid
					}
					set ::buttons::butsinfo(pgc,$vts,$vid,$cid,navpack) \
							[get_int cellpbtable [expr (($cell-1)*24)+8] 4]
				}
			}
			for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
				if {! [info exists ::buttons::butsinfo(pgc,$vts,$vid,numcellids)]} {
					set ::buttons::butsinfo(pgc,$vts,$vid,numcellids) 0
				}
				for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
					set ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts) 0
					set ::buttons::butsinfo(pgc,$vts,$vid,$cid,numgroups) 0
					if {![info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,navpack)]} {
						set ::buttons::butsinfo(pgc,$vts,$vid,$cid,navpack) -1	;# missing cell ID
					}
				}
			}
		}
	}

	set idx 0
	set totalbuttons 0
	set v 0
	while {$idx < [llength $in]} {
		set line [string trim [lindex $in $idx]]
		if {[string length $line]==0 || [string range $line 0 0]=="#"} {
			incr idx
			continue
		}
		set nummatches [scan $line "%d %d %d %d %d" vts lu pgc cell numbuttons]
		if {$nummatches != 5} {
			show_msg "ButtonInfo.txt: Wrong file format!"
			bell
			return -1
		}
		incr idx
		if {$vts > $::pgcs(numvts)} {
			tk_messageBox  -title "Open ButtonInfo" -icon error \
					-message "VTS number ($vts) out of range in line \"$line\".\nThis buttoninfo.txt doesn't match the current IFOs."
			for {set but 1} {$but <= $numbuttons} {incr but} {incr idx}
			continue
		}
		if {$lu == 0} {
			set type pgc
			if {$pgc > $::pgcs($vts,numpgcs)} {
				tk_messageBox  -title "Open ButtonInfo" -icon error \
						-message "PGC number ($pgc) out of range in line \"$line\".\nThis buttoninfo.txt doesn't match the current IFOs."
				for {set but 1} {$but <= $numbuttons} {incr but} {incr idx}
				continue
			}
			set cellpostable $::pgcs($vts,$pgc,cellpostable)
			set ::pgcs($vts,$pgc,numbuttons) [expr $numbuttons+$::pgcs($vts,$pgc,numbuttons)]
			set numcells $::pgcs($vts,$pgc,numcells)
		} else {
			set type menu
			if {$lu > $::menus($vts,numlus)} {
				tk_messageBox  -title "Open ButtonInfo" -icon error \
						-message "LU number ($lu) out of range in line \"$line\".\nThis buttoninfo.txt doesn't match the current IFOs."
				for {set but 1} {$but <= $numbuttons} {incr but} {incr idx}
				continue
			}
			if {$pgc > $::menus($vts,$lu,numpgcs)} {
				tk_messageBox  -title "Open ButtonInfo" -icon error \
						-message "PGC number ($pgc) out of range in line \"$line\".\nThis buttoninfo.txt doesn't match the current IFOs."
				for {set but 1} {$but <= $numbuttons} {incr but} {incr idx}
				continue
			}
			set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
			set ::menus($vts,$lu,$pgc,numbuttons) [expr $numbuttons+$::menus($vts,$lu,$pgc,numbuttons)]
			set numcells $::menus($vts,$lu,$pgc,numcells)
		}

		if {$cell>$numcells} {
			tk_messageBox  -title "Open ButtonInfo" -icon error \
					-message "Cell number ($cell) out of range in line \"$line\".\nThis buttoninfo.txt doesn't match the current IFOs."
			for {set but 1} {$but <= $numbuttons} {incr but} {incr idx}
			continue
		}

		# retrieve VOB/Cell ID of the current cell
		set cellposidx [expr ($cell-1)*4]
		set vid [get_int cellpostable $cellposidx 2]
		incr cellposidx 3
		set cid [get_int cellpostable $cellposidx 1]

		set ::buttons::butsinfo($type,$vts,$vid,$cid,numbuts) $numbuttons
		set ::buttons::butsinfo($type,$vts,$vid,$cid,numgroups) 1
		set ::buttons::butsinfo($type,$vts,$vid,$cid,hlstarttime) -1
		set ::buttons::butsinfo($type,$vts,$vid,$cid,hlendtime) -1
		set ::buttons::butsinfo($type,$vts,$vid,$cid,selendtime) -1
		set ::buttons::butsinfo($type,$vts,$vid,$cid,forceselectbutn) 0
		set ::buttons::butsinfo($type,$vts,$vid,$cid,forceactionbutn) 0
		set ::buttons::butsinfo($type,$vts,$vid,$cid,numselectablebuts) 0
		set ::buttons::butsinfo($type,$vts,$vid,$cid,1,grouptype) 7

		for {set but 1} {$but <= $numbuttons} {incr but} {
			if {! [info exists buttonsinvts($type,$vts)]} {
				set buttonsinvts($type,$vts) 0
			}
			incr buttonsinvts($type,$vts)
			set cmd [string range [lindex $in $idx] 1 23]
			set rest [string range [lindex $in $idx] 26 end]
			set nummatches [scan $rest "%d %d %d %d  %d %d %d %d  %d" posx posy width height  up dn lf rg  auto]
			incr idx

			set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,command) $cmd

			if {$nummatches < 4 || ($posx==0 && $posy==0 && $width==0 && $height==0)} {
				set posx [expr ((($but-1)/10)*150)+50]
				set posy [expr ((($but-1)%10)*44)+20]
				set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,coords) "$posx $posy - [expr $posx+130] [expr $posy+40]"
			} else {
				set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,coords) "$posx $posy - [expr $posx+$width] [expr $posy+$height]"
			}

			if {$nummatches < 8} {
				set prev [expr $but-1]
				if {$prev==0} {set prev $numbuttons}
				set next [expr $but+1]
				if {$next>$numbuttons} {set next 1}
				set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,udlr)	"$prev $next $prev $next"
			} else {
				if {$up==0} {set up $but}
				if {$dn==0} {set dn $but}
				if {$lf==0} {set lf $but}
				if {$rg==0} {set rg $but}
				set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,udlr)	"$up $dn $lf $rg"
			}

			if {$nummatches < 9} {
				set auto 0
			}
			set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,autoaction) $auto

			set ::buttons::butsinfo($type,$vts,$vid,$cid,1,$but,colorscheme) 1

			incr totalbuttons
		}
	}
	foreach type {menu pgc} {
		for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
			if {[info exists buttonsinvts($type,$vts)]} {
				set ::buttons::butsinfo($type,$vts,totalbuttons) $buttonsinvts($type,$vts)
			} else {
				set ::buttons::butsinfo($type,$vts,totalbuttons) 0
			}
		}
	}
	::main::init_tree_menu_no_move
	show_msg "$totalbuttons menu buttons or BOVs loaded from $filename"
	set ::menued::scan_done true
	return $totalbuttons
}

}	;# end of namespace ::menubuttons
