#############################################################################
# readdvd.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# read DVD IFOs in memory.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################


## Variables usage:
##
## pgcs(numvts)								;# number of VTS in the DVD
## pgcs($vtsnum,numpgcs)					;# number of PGCs in this VTS
## pgcs($vtsnum,$pgcnum)					;# list of commands in this PGC
## pgcs($vtsnum,$pgcnum,Title)				;# Title number for this PGC
## pgcs($vtsnum,$pgcnum,TTN)				;# Title number within titleset
## pgcs($vtsnum,$pgcnum,header)				;# Header part of the PGC
## pgcs($vtsnum,$pgcnum,numprograms)		;# Number of Programs in the PGC
## pgcs($vtsnum,$pgcnum,programmap)			;# PGC program map, in binary
## pgcs($vtsnum,$pgcnum,numcells)			;# Number of Cells in the PGC
## pgcs($vtsnum,$pgcnum,cellpbtable)		;# Cell Playback Table, in binary
## pgcs($vtsnum,$pgcnum,cellpostable)		;# Cell Position Table, in binary
## pgcs($vtsnum,$pgcnum,pttlist)			;# List of lists with {program chapter ttu}
##											;# formatted as {{pg# ptt# ttn#} {pg# ptt# ttn#} ...}
##
## ttns($vtsnum,$ttn)						;# Entry PGC for this TTN (title in VTS)
## pgcs(Title,$title) [list $vts $pgc]		;# Target PGC for this Title
##
##
## menus(numvts)							;# number of VTS in the DVD
## menus($vtsnum,numlus)					;# number of LUs in this VTS (0=no menu)
## menus($vtsnum,lus_are_identical)			;# Bool: There is the same number of PGCs in the
##											;# different LUs, and same commands in the PGCs.
## menus($vtsnum,$lunum,lang)				;# language code for this LU
## menus($vtsnum,$lunum,numpgcs)			;# number of PGCs in this LU
## menus($vtsnum,$lunum,$pgcnum,mask)		;# category mask value for this PGC
## menus($vtsnum,$lunum,$pgcnum,masktext)	;# mask in human readeable form, or ""
## menus($vtsnum,$lunum,$pgcnum)			;# list of commands in this PGC
## menus($vtsnum,$lunum,$pgcnum,header)		;# Header part of the PGC
## menus($vtsnum,$lunum,$pgcnum,numprograms);# Number of Programs in the PGC
## menus($vtsnum,$lunum,$pgcnum,programmap)	;# PGC program map, in binary
## menus($vtsnum,$lunum,$pgcnum,numcells)   ;# Number of Cells in the PGC
## menus($vtsnum,$lunum,$pgcnum,cellpbtable);# Cell Playback Table, in binary
## menus($vtsnum,$lunum,$pgcnum,cellpostable);# Cell Position Table, in binary
## menus($vtsnum,$lunum,$pgcnum,numbuttons)	;# total number of buttons for this PGC.
##
## raw_sectors($vtsnum,<tablename>)			;# raw IFO data, splitted by tables
## VMG tables names: VMGM_MAT VMG_TT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT
##					 VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP
## VTS tables names: VTSI_MAT VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI
##					 VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP
##
## dvd_modified								;# flag: the DVD has been modified.
##
## Notes:
## - Tcl arrays are one dimention only.  The indexes are __string__ values.
##   Therefore, you cannot use spaces around the comas in the index names!
## - the array indexes that are not prefixed with $ are constant text strings.
## - The vtsnum, pgcnum, lunum and commands indexes are numbered starting at 1
## - vtsnum 0 is the VMG (VIDEO_TS.IFO)
## - the lists of commands are Tcl lists with the commands in hex form,
##   with blank between bytes: "00 00 00 00 00 00 00 00".
##   Elements beginning with * are pre, post and cell separators
##
##
## Buttons informations.  Available ONLY if the DVD is NOT encrypted!
## The Menu VOBs are always parsed.  BOVs are parsed on demmand!
## Note: vts = 0 for VMG.
##
## For each VTS, the following variables are set:
## ::buttons::butsinfo($type,$vts,errorcode)	;# 0=OK
## ::buttons::butsinfo($type,$vts,errormsg)
## ::buttons::butsinfo($type,$vts,numvobids)
## For each VOB ID
## ::buttons::butsinfo($type,$vts,$vobid,numcellids)
## For each Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,navpack)	;# first NAV pack LBA for current VOB/Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numbuts)	;# number of buttons (in first group)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceselectbutn)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceactionbutn)
## For eacg group
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,grouptype)	;# 1=normal, 2=letterbox, 4=pan&scan (ored together)
## For each button
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,autoaction) ;# 0 or 1
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,colorscheme) ;# 0 to 3
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,coords)	;# Button position: {x_start y_start - x_end y_end}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,udlr)		;# Adjacent buttons: {up down left right}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,command)	;# Button command: {XX XX XX XX XX XX XX XX}

namespace eval ::readdvd {

namespace export readdvd readdvdfile parse_PGCI_UT

#############################################################################
## Procedure:  get_hex_s
## Get a hex string from inside the binary data.  Pad bytes with spaces.
proc {get_hex_s} {binarystring_upvar start_offset length} {

	upvar $binarystring_upvar binarystring

	set s [string range $binarystring $start_offset [expr {$start_offset + $length - 1}]]
	binary scan $s H* hex
	set len [string length $hex]
	set out ""
	for {set i 0} {$i<$len} {incr i 2} {
		set out "$out[string range $hex $i [expr {$i+1}]] "
	}
	return [string trim $out]
}

#############################################################################
## Procedure:  getCommandsFromPGC
## return a list with all commands of the provided PGC in hexadecimal form.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
## Elements beginning with * are pre, post and cell separators
proc {getCommandsFromPGC} {data_upvar pgcindex type vts lu pgc} {

	upvar $data_upvar data
	if {$pgcindex >= [string length $data]} {
		return {{* pre} {* post} {* cell}}
	}

	set cmdtable_start [get_int data [expr {$pgcindex+0XE4}] 2]
	if {$cmdtable_start == 0} {
		set num_pre_commands  0
		set num_post_commands 0
		set num_cell_commands 0
	} else {
		# remove PGC header, and keep only the command table and whatever follow...
		set cmds_table [string range $data [expr {$pgcindex+$cmdtable_start}] end]
		if {[string length $cmds_table] < 8} {
			set num_pre_commands -1
			set num_post_commands 0
			set num_cell_commands 0
		} else {
			# get number of pre, post and cell commands
			set num_pre_commands  [get_int cmds_table 0 2]
			set num_post_commands [get_int cmds_table 2 2]
			set num_cell_commands [get_int cmds_table 4 2]
			tolog "Found $num_pre_commands pre, $num_post_commands post and $num_cell_commands cell commands."
		}
		if {[expr $num_pre_commands+$num_post_commands+$num_cell_commands] > 128 || \
					$num_pre_commands == -1} {
			if {$num_pre_commands > -1} {
				set total_num_cmds [expr ([get_int cmds_table 6 2]-7)/8]
				if {$total_num_cmds > 128} {
					tolog "WARNING!  Too many commands!"
					set total_num_cmds 128
				}
			}

			if {$type == "pgc"} {
				if {$vts == 0} {
					set label "First-Play PGC"
				} else {
					set label "VTST $vts, PGC $pgc"
				}
			} else {
				if {$vts == 0} {
					set label "VMGM LU $lu, PGC $pgc"
				} else {
					set label "VTSM $vts, LU $lu, PGC $pgc"
				}
			}
			set skip false
			if {$num_pre_commands == -1} {
				set msg "ERROR: Wrong start of command table in $label !"
				set msg "$msg\nIt is beyond the end of the PGCI table !\nAssuming 0 pre/post/cell commands."
				set num_pre_commands  0
				set skipid pgc_wrong_start_of_cmd_table
				set skip [info exists ::skip_dialog_during_session($skipid)]
			} else {
				set msg "Warning: Too many commands in $label !"
				if {$num_cell_commands > 128} {
					set msg "$msg\nFixing number of cell-commands from $num_cell_commands"
					set num_cell_commands [expr $total_num_cmds - ($num_pre_commands+$num_post_commands)]
					if {$num_cell_commands < 0} {set num_cell_commands 0}
					set msg "$msg to $num_cell_commands."
				}
				if {$num_post_commands > 128} {
					set msg "$msg\nFixing number of post-commands from $num_post_commands"
					set num_post_commands [expr $total_num_cmds - ($num_pre_commands+$num_cell_commands)]
					if {$num_post_commands < 0} {set num_post_commands 0}
					set msg "$msg to $num_post_commands."
				}
				if {$num_pre_commands > 128} {
					set msg "$msg\nFixing number of pre-commands from $num_pre_commands"
					set num_pre_commands [expr $total_num_cmds - ($num_post_commands+$num_cell_commands)]
					if {$num_pre_commands < 0} {set num_pre_commands 0}
					set msg "$msg to $num_pre_commands."
				}
				set skipid pgc_too_many_commands
				set skip [info exists ::skip_dialog_during_session($skipid)]
			}
			set ::dvd_modified true
			if {!$skip} {
				set msg "$msg\nPlease save the DVD."
				set r [tk_messageBox -title "PgcEdit: Open DVD" -icon error -message $msg \
						-parent . -type "OK to all|OK"]
				if {$r == "ok to all"} {
					set ::skip_dialog_during_session($skipid) true
				}
			}
		}
	}

	set idx 8
	lappend outlist "* pre"
	for {set i 0} {$i < $num_pre_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* post"
	for {set i 0} {$i < $num_post_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* cell"
	for {set i 0} {$i < $num_cell_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	return $outlist
}

#############################################################################
## Procedure:  getHeaderFromPGC
## return the header part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getHeaderFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	return [string range $data $pgcindex [expr {$pgcindex + 0XEB}]]
}

#############################################################################
## Procedure:  getProgramMapFromPGC
## return the program map part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getProgramMapFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numprograms [get_int data [expr {$pgcindex + 2}] 1]
	if {$numprograms == 0} {return [list 0 ""]}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XE6}] 2]}]
	set end [expr {$start + $numprograms - 1}]
	if {[expr {$end % 2}] == 0} {incr end}	;# must be padded to word boundary
	return [list $numprograms [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPBTableFromPGC
## return the cells Playback table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPBTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr {$pgcindex + 3}] 1]
	tolog "Found $numcells cells in Cell playback table."
	if {$numcells == 0} {return [list 0 ""]}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XE8}] 2]}]
	set end [expr {$start + ($numcells * 24) - 1}]
	return [list $numcells [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPosTableFromPGC
## return the cells Position table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPosTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr {$pgcindex + 3}] 1]
	tolog "Found $numcells cells in Cell position table"
	if {$numcells == 0} {return ""}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XEA}] 2]}]
	set end [expr {$start + ($numcells * 4) - 1}]
	return [string range $data $start $end]
}

#############################################################################
## Procedure:  parse_PGCI_UT
## process the data in *_PGCI_UT
proc parse_PGCI_UT {pgci_ut_raw_data_upvar vtsnum menusvar} {
	tolog "Parsing PGCI_UT (VTS $vtsnum)."
	upvar $pgci_ut_raw_data_upvar pgci_ut_raw_data

	# get number of Language Units
	if {$pgci_ut_raw_data == ""} {
		set ${menusvar}($vtsnum,numlus) 0
		return true
	} else {
		set numlus [get_int pgci_ut_raw_data 0 2]
		set ${menusvar}($vtsnum,numlus) $numlus
	}

	set lu_pgcs(1) 0

	# for all Language units
	for {set lu 1} {$lu <= $numlus} {incr lu} {

		# get LU's language, number of PGCs and offset
		set lang [string range $pgci_ut_raw_data [expr {$lu * 8}] [expr {($lu * 8) + 1}]]
		if {$lang == "\000\000" || $lang == "\777\777" || $lang == "  "} {set lang "--"}
		set ${menusvar}($vtsnum,$lu,lang) $lang

		tolog "Parsing LU $lu ($lang)"

		if {[string length $pgci_ut_raw_data] < [expr {($lu * 8) + 8}]} {
			show_warning  -title "Parsing PGCI_UT" -icon error -parent . -message \
				"WARNING: LU pointer out of range in VTS $vtsnum.\nRemoving the whole menu domain!"
			set ${menusvar}($vtsnum,numlus) 0
			return false
		}
		set offset [get_int pgci_ut_raw_data [expr {($lu * 8) + 4}] 4]

		set ludata [string range $pgci_ut_raw_data $offset end]

		set numpgcs [get_int ludata 0 2]
		set ${menusvar}($vtsnum,$lu,numpgcs) $numpgcs

		#dprint 2 "\nMenu language unit $lu: \"$lang\""
		#dprint 2 "offset to VMGM_LU = 0X[format %08X $offset] ($offset)"

		set lu_pgcs($lu) $numpgcs

		# for each PGC
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
			tolog "Parsing LU $lu, PGC $pgc."
			# get and store category mask
			set category [get_int ludata [expr {$pgc * 8}] 1]
			set ${menusvar}($vtsnum,$lu,$pgc,mask) $category
			switch [expr {$category & 0X0F}] {
				0 {set masktext ""}
				2 {set masktext "TitleM"}
				3 {set masktext "RootM"}
				4 {set masktext "SubPicM"}
				5 {set masktext "AudioM"}
				6 {set masktext "AngleM"}
				7 {set masktext "ChapterM"}
				default {set masktext "UNKNOWN MENU"}
			}
			set ${menusvar}($vtsnum,$lu,$pgc,masktext) $masktext

			# get and store PGC offset
			set pgc_offset [get_int ludata [expr {($pgc * 8) + 4}] 4]

			# get and store PGC's commands
			set ${menusvar}($vtsnum,$lu,$pgc) [getCommandsFromPGC ludata $pgc_offset menu $vtsnum $lu $pgc]

			# get and store the other PGC parts
			set ${menusvar}($vtsnum,$lu,$pgc,header) [getHeaderFromPGC ludata $pgc_offset]
			set tmp [getProgramMapFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numprograms) [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,programmap)  [lindex $tmp 1]

			set tmp [getCellsPBTableFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numcells)    [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,cellpbtable) [lindex $tmp 1]

			set ${menusvar}($vtsnum,$lu,$pgc,cellpostable) \
					[getCellsPosTableFromPGC ludata $pgc_offset]
		}
	}
	# verify that there is the same number of PGCs in each LUs,
	# and that PGC commands are identical!
	set numpgcs $lu_pgcs(1)
	set lus_are_identical true
	if {$numlus > 1} {
		if {$vtsnum == 0} {set domain "VMGM"} else {set domain "VTSM $vtsnum"}
		for {set lu 2} {$lu <= $numlus} {incr lu} {
			if {$lu_pgcs($lu) != $numpgcs} {
				set txt "Caution: There is a different number of PGCs"
				set txt "$txt\nin the menu LUs of the $domain domain."
				set txt "$txt\n\nThis is somewhat unusual, and may cause"
				set txt "$txt\nproblems with the 'Set Menu Type' and"
				set txt "$txt\n'Copy LU Commands' functions."
				tk_messageBox -icon info -title "PgcEdit: Read DVD" -parent . \
						-type ok -message $txt
				set lus_are_identical false
				break
			}
		}
		if {$lus_are_identical} {
			for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
				set lu1cmds [subst \$${menusvar}($vtsnum,1,$pgc)]
				for {set lu 2} {$lu <= $numlus} {incr lu} {
					if {[subst \$${menusvar}($vtsnum,$lu,$pgc)] != $lu1cmds} {
						set txt "Warning: There are different commands"
						set txt "$txt\nin the various PGCs of the menu LUs"
						set txt "$txt\nof the $domain domain."
						set txt "$txt\n\nThis is somewhat unusual, and may cause"
						set txt "$txt\nproblems with the 'Set Menu Type' and"
						set txt "$txt\n'Copy LU Commands' functions."
						tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
								-type ok -message $txt
						set lus_are_identical false
						break
					}
				}
				if {!$lus_are_identical} break
			}
		}
		tolog "Identical commands in all LUs: $lus_are_identical"
	}
	set ${menusvar}($vtsnum,lus_are_identical) $lus_are_identical
	return true
}

#############################################################################
## Procedure:  parse_VTS_PGCI
## process the data in the VTS_PGCI of a VTS IFO
proc parse_VTS_PGCI {pgci_data_upvar vtsnum pgcsvar} {
	tolog "Parsing VTS_PGCI."

	upvar $pgci_data_upvar pgci_data

	if {$pgci_data == ""} {
		set ${pgcsvar}($vtsnum,numpgcs) 1
		set ${pgcsvar}($vtsnum,1,TTN) 1
		set ${pgcsvar}($vtsnum,1) {{* pre} {30 08 00 00 01 00 00 00} {* post} {* cell}}
		set ${pgcsvar}($vtsnum,1,header) [string repeat \000 236]
		set_int ${pgcsvar}($vtsnum,1,header) 2 1 1
		set_int ${pgcsvar}($vtsnum,1,header) 3 1 1
		set ${pgcsvar}($vtsnum,1,numprograms) 1
		set ${pgcsvar}($vtsnum,1,programmap) \001
		set ${pgcsvar}($vtsnum,1,numcells) 1
		set ${pgcsvar}($vtsnum,1,cellpbtable) \002[string repeat \000 23]
		set ${pgcsvar}($vtsnum,1,cellpostable) \000\001\000\001
		set ${pgcsvar}($vtsnum,1,pttlist) {{1 1 1}}
		set ::raw_sectors($vtsnum,VTS_PGCI) [string repeat \000 16]
		set_int ::raw_sectors($vtsnum,VTS_PGCI) 0 1 2
		set_int ::raw_sectors($vtsnum,VTS_PGCI) 8 129 1
		return
	}
	set numpgcs [get_int pgci_data 0 2]
	set ${pgcsvar}($vtsnum,numpgcs) $numpgcs

	# for each PGC
	for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
		tolog "Parsing PGC $pgc"
		# get and store PGC's TTN
		set ${pgcsvar}($vtsnum,$pgc,TTN) [expr {[get_int pgci_data [expr {$pgc * 8}] 1] & 0x7F}]	;# bug fixed

		# get PGC offset
		set pgc_offset [get_int pgci_data [expr {($pgc * 8) + 4}] 4]
		# get and store PGC's commands
		set ${pgcsvar}($vtsnum,$pgc) [getCommandsFromPGC pgci_data $pgc_offset pgc $vtsnum 0 $pgc]

		set ${pgcsvar}($vtsnum,$pgc,header) [getHeaderFromPGC pgci_data $pgc_offset]

		set tmp [getProgramMapFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numprograms) [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,programmap)  [lindex $tmp 1]

		set tmp [getCellsPBTableFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numcells)    [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,cellpbtable) [lindex $tmp 1]

		set ${pgcsvar}($vtsnum,$pgc,cellpostable) [getCellsPosTableFromPGC pgci_data $pgc_offset]

		set ${pgcsvar}($vtsnum,$pgc,pttlist) {}	;# bug fixed
	}
	# HD-DVD incompatibility!  The TTN number is wrong!
	set firstttn 1000
	for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
		if {[subst \$${pgcsvar}($vtsnum,$pgc,TTN)] < $firstttn} {
			set firstttn [subst \$${pgcsvar}($vtsnum,$pgc,TTN)]
		}
	}
	if {$firstttn != 1} {
		set ttnoffset [expr $firstttn -1]
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
			set ${pgcsvar}($vtsnum,$pgc,TTN) [expr [subst \$${pgcsvar}($vtsnum,$pgc,TTN)] - $ttnoffset]
		}
	}
}

#############################################################################
## Functions to fix the end byte pointers of the tables

proc real_endbyte {tablename start_address end_address data_upvar} {
	upvar $data_upvar data
	if {$end_address == 0} {
		return 0
	}
	set endbyte [expr $end_address - $start_address]
	if {$endbyte <= 0} {
		return 0
	}
	if {$endbyte >= [string length [string range $data $start_address end]]} {
		return [expr [string length [string range $data $start_address end]]-1]
	}

	switch $tablename {
		VMGM_MAT {
			set endbyte [real_endbyte_VMGM_MAT $endbyte $data]
			set_int data 0x80 $endbyte 4
		}
		VMG_TT_SRPT {
			set endbyte [real_endbyte_VMG_TT_SRPT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VMGM_PGCI_UT {
			set endbyte [real_endbyte_PGCI_UT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VMG_PTL_MAIT {
			# TODO
		}
		VMG_VTS_ATRT {
			# Rebuilt from scratch when the DVD is saved
		}
		VMG_TXTDT_MG {
			# TODO
		}
		VMGM_C_ADT {
			set endbyte [real_endbyte_C_ADT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}

		VTSI_MAT {
			if {$endbyte < 983} {
				set endbyte 983
				set_int data 128 983 4
			}
		}
		VTS_PTT_SRPT {
			set endbyte [real_endbyte_VTS_PTT_SRPT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VTS_PGCI {
			set endbyte [real_endbyte_VTS_PGCI $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VTSM_PGCI_UT {
			set endbyte [real_endbyte_PGCI_UT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VTS_TMAPTI {
			set endbyte [real_endbyte_VTS_TMAPTI $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VTSM_C_ADT {
			set endbyte [real_endbyte_C_ADT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
		VTS_C_ADT {
			set endbyte [real_endbyte_C_ADT $endbyte [string range $data $start_address end]]
			set_int data [expr $start_address+4] $endbyte 4
		}
	}
	return $endbyte
}

proc real_endbyte_VMG_TT_SRPT {orig data} {
	if {$data == ""} {return 0}
	set numtitles [get_int data 0 2]	;# last PGC in domain
	set endbyte [get_int data 4 4]
	set lastbyte [expr {($numtitles * 12) + 7}]
	if {$lastbyte > $endbyte} {
		set endbyte $lastbyte
	}
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}
proc real_endbyte_VMGM_MAT {orig data} {
	if {$data == ""} {return 0}
	set endbyte [get_int data 0x80 4]
	set pgc_offset [get_int data 0x84 4]
	set endbyte [endbyteofpgc $data $pgc_offset $endbyte]
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}

proc real_endbyte_VTS_PGCI {orig data} {
	if {$data == ""} {return 0}
	set numpgcs [get_int data 0 2]	;# last PGC in domain
	set endbyte [get_int data 4 4]
	set pgc_offset [get_int data [expr {($numpgcs * 8) + 4}] 4]
	set endbyte [endbyteofpgc $data $pgc_offset $endbyte]
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}

proc checknumpgcs {data} {
	if {$data == ""} {return 0}
	set numpgcs [get_int data 0 2]
	set endbyte [get_int data 4 4]
	for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
		set pgc_offset [get_int data [expr {($pgc * 8) + 4}] 4]
		if {$pgc_offset >= [string length $data]} {
			return [expr $pgc-1]
		}
		set tmp [expr $pgc_offset + 0xE4]
		if {$tmp >= [string length $data]} {
			return [expr $pgc-1]
		}
		set offset [get_int data [expr $pgc_offset + 0xE4] 2]
		if {$offset > 0} {
			set tablestart [expr $pgc_offset + $offset+6]
			if {$tablestart >= [string length $data]} {
				return [expr $pgc-1]
			}
		}
		set rc [catch {set tmp [endbyteofpgc $data $pgc_offset $endbyte]}]
		if {$rc} {
			return [expr $pgc -1]
		}
	}
	return $numpgcs
}

proc real_endbyte_PGCI_UT {orig data} {
	if {$data == ""} {return 0}
	set numlus [get_int data 0 2]
	set endbyte [get_int data 4 4]
	set lustart [get_int data [expr ($numlus*8)+4] 4]

	set numpgcs [get_int data $lustart 2]
	# get PGC offset
	set pgc_offset [expr $lustart + [get_int data [expr $lustart + ($numpgcs*8) + 4] 4]]
	set endbyte [endbyteofpgc $data $pgc_offset $endbyte]
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}
proc endbyteofpgc {data pgc_offset endbyte} {
	set new_endbyte $endbyte
	if {$new_endbyte < [expr $pgc_offset+0xEB]} {
		set new_endbyte [expr $pgc_offset+0xEB]
	}
	if {$pgc_offset > 500000} {
		return $new_endbyte
	}
	# verify end of command table
	set offset [get_int data [expr $pgc_offset + 0xE4] 2]
	if {$offset > 0} {
		set tablestart [expr $pgc_offset + $offset]
		set end [expr $pgc_offset + $offset + [get_int data [expr $tablestart+6] 2]]
		if {$new_endbyte < $end} {
			set new_endbyte $end
		}
	}

	set numpgs [get_int data [expr $pgc_offset+2] 1]
	set numcells [get_int data [expr $pgc_offset+3] 1]
	if {$numcells > 0} {
		# verify end of program map
		set offset [get_int data [expr $pgc_offset + 0xE6] 2]
		if {$offset > 0} {
			set tablestart [expr $pgc_offset + $offset]
			set end [expr $tablestart + $numpgs - 1]
			if {$new_endbyte < $end} {
				set new_endbyte $end
			}
		}
		# verify end of cell playback table
		set offset [get_int data [expr $pgc_offset + 0xE8] 2]
		if {$offset > 0} {
			set tablestart [expr $pgc_offset + $offset]
			set end [expr $tablestart + ($numcells*24) - 1]
			if {$new_endbyte < $end} {
				set new_endbyte $end
			}
		}
		# verify end of cell position table
		set offset [get_int data [expr $pgc_offset + 0xEA] 2]
		if {$offset > 0} {
			set tablestart [expr $pgc_offset + $offset]
			set end [expr $tablestart + ($numcells*4) - 1]
			if {$new_endbyte < $end} {
				set new_endbyte $end
			}
		}
	}
	if {$new_endbyte > $endbyte} {
		set endbyte $new_endbyte
	}
	return $endbyte
}

proc real_endbyte_VTS_PTT_SRPT {orig data} {
	if {$data == ""} {return 0}
	set numttus [get_int data 0 2]
	if {$numttus == 0} {return 0}
	set endbyte [get_int data 4 4]
	if {$endbyte == 0} {return 0}
	if {$endbyte > [expr 0x00FFFFFF]} {
		set endbyte [expr $endbyte & 0x00FFFFFF]
	}
	set lastttustart [get_int data [expr ($numttus*4)+4] 4]
	set minend [expr $lastttustart+3]
	if {$minend > $endbyte} {
		set endbyte $minend
	}
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}
proc real_endbyte_VTS_TMAPTI {orig data} {
	if {$data == ""} {return 0}
	set numpgcs [get_int data 0 2]
	if {$numpgcs == 0} {return 0}
	set endbyte [get_int data 4 4]
	if {$endbyte == 0} {return 0}
	set tmstart [get_int data [expr ($numpgcs*4)+4] 4]
	set numentries [get_int data [expr $tmstart+2] 2]
	set lastbyte [expr $tmstart+($numentries*4)+3]
	if {$lastbyte > $endbyte} {
		set endbyte $lastbyte
	}
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}
proc real_endbyte_C_ADT {orig data} {
	if {$data == ""} {return 0}
	set numvobs [get_int data 0 2]
	if {$numvobs == 0} {return 0}
	set endbyte [get_int data 4 4]
	if {$endbyte == 0} {return 0}
	set lastbyte [expr ($numvobs*12) + 7]
	if {$lastbyte > $endbyte} {
		set endbyte $lastbyte
	}
	if {$endbyte >= [string length $data]} {return $orig}
	return $endbyte
}

#############################################################################
## Procedure:  parse_VTS_PTT_SRPT
## process the data in the VTS_PTT_SRPT of a VTS IFO
## Build chapter list for each titles
proc parse_VTS_PTT_SRPT {vts_ptt_srpt_data_upvar vtsnum pgcsvar ttnsvar} {
	tolog "Parsing VTS_PTT_SRPT (VTS $vtsnum)"

	upvar $vts_ptt_srpt_data_upvar data

	set numttus 0
	if {$data != ""} {
		set numttus [get_int data 0 2]
		set maxpgcn -1
		for {set ttu 1} {$ttu <= $numttus} {incr ttu} {
			tolog "Processing TTU $ttu."
			set ptt 1
			set idx [get_int data [expr {4 + ($ttu * 4)}] 4]
			if {$ttu < $numttus} {
				set endidx [expr {[get_int data [expr {4 + (($ttu+1) * 4)}] 4] -1}]
			} else {
				set endidx [get_int data 4 4]
			}
			# build table of entry PGCs for each TTU
			set pgcn [get_int data $idx 2]
			set ${ttnsvar}($vtsnum,$ttu) $pgcn

			while {$idx < $endidx} {
				set pgcn [get_int data $idx 2]
				if {$pgcn > $maxpgcn} {
					set maxpgcn $pgcn
				}
				set pg  [get_int data [expr {$idx+2}] 2]
				#lappend PTTlist($pgcn) "$pg $ptt"
				lappend PTTlist($pgcn) "$pg $ptt $ttu" ;# Added TTU in 0.5.0beta11

				#if {$ttu != ${pgcsvar}($vtsnum,$pgcn,TTN)} {
				#	WARNING!  TTU do not match TTN
				#}
				incr idx 4
				incr ptt
			}
		}
		for {set pgcn 1} {$pgcn <= $maxpgcn} {incr pgcn} {
			if {[info exists PTTlist($pgcn)]} {
				set ${pgcsvar}($vtsnum,$pgcn,pttlist) $PTTlist($pgcn)
			} else {
				set ${pgcsvar}($vtsnum,$pgcn,pttlist) {}
			}
		}
	} else {
		set txt "Error: Missing VTS_PTT_SRPT table (chapter table) in VTS $vtsnum."
		set txt "$txt\nThis error is NOT fixed automatically by PgcEdit!"
		set txt "$txt\nVTS $vtsnum is probably a fake titleset added only to confuse the rippers!"
		set txt "$txt\nVerify if VTS $vtsnum is called from the VMG, and, if it's not the case, remove it."
		set txt "$txt\nYou can probably do that automatically with DVD -> Delete Uncalled PGCs."
		tk_messageBox -title "Open DVD: Parse VTS_PTT_SRPT table" -message $txt -icon error
	}
	set ${pgcsvar}($vtsnum,numttus) $numttus
	return $numttus
}

#############################################################################
## Procedure:  read_VMG_IFO
## read the data of the VIDEO_TS.IFO of the DVD directory supplied
proc read_VMG_IFO {filename sectorsvar pgcsvar menusvar} {
	tolog "Reading VMG IFO."
	set error 0
	set data [readFile $filename binary]
	if {$data == ""} {
		set errmsg "Can't open [file tail $filename]\nor file empty!"
		set error 1
	}
	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VMG"} {
		set errmsg "Error checking IFO file type:\nFile [file tail $filename] is not of correct type DVDVIDEO-VMG."
		set error 1
	}
	if {$error} {
		set error 0
		set msg "$errmsg\n\nDo you want to try to load the BUP file instead?"
		set r [tk_messageBox -title "Read DVD" -message $msg -icon warning \
				-type yesnocancel -default yes -skip LoadBUPinstead]
		if {$r == "cancel"} {
			return 1000
		} elseif {$r == "yes"} {
			set filename [string replace $filename end-2 end BUP]
			set data [readFile $filename binary]
			if {$data == ""} {
				tolog "Can't read [file tail $filename]"
				return 1
			}
			if {[string range $data 0 11] != "DVDVIDEO-VMG"} {
				errmsg 1000 "checking IFO file type" \
						"File \"[file tail $filename]\" is not of correct type DVDVIDEO-VMG."
				return 1000
			}
		}
	}

	#dprint 2 ""
	##### split the VIDEO_TS.IFO file data in main tables parts
	tolog "Computing start and end addresses of the VMG tables."
	set start_address(0) 0

	set end_address(0) [get_int data 0X80 4]	;# should work, but...
	set fp_pgc_start [get_int data 0X84 4]


	### WARNING: BUG in IfoEdit: the end address of the VMGM_MAT is not exact!!!
	### IfoEdit never change this address, even when the command table is growing!
	### We must therefore calculate the real end address (what a pitty!).
	if {$fp_pgc_start != 0} {
		set cmdtable_start [get_int data [expr {$fp_pgc_start+0XE4}] 2]
		if {$cmdtable_start != 0} {
			set numpre  [get_int data [expr {$fp_pgc_start+$cmdtable_start}] 2]
			set numpost [get_int data [expr {$fp_pgc_start+$cmdtable_start+2}] 2]
			set numcell [get_int data [expr {$fp_pgc_start+$cmdtable_start+4}] 2]
			set cmdtable_len [expr {8+($numpre*8)+($numpost*8)+($numcell*8)}]
			set cmdtable_end [expr {$fp_pgc_start+$cmdtable_start+$cmdtable_len-1}]
		} else {
			set numpre  0
			set numpost 0
			set numcell 0
			set cmdtable_len 0
			set cmdtable_end $end_address(0)
		}
	} else {
		# Yes, there are DVDs WITHOUT FP-PGC!
		# A new FP-PGC is created by PgcEdit in parse_VMG_IFO
		set cmdtable_end $end_address(0)
	}
	if {$cmdtable_end != $end_address(0)} {
		set msg "Warning: The value for \"End byte of VMGI_MAT\" is incorrect!"
		set msg "$msg\n(This value is in the VIDEO_TS.IFO's VMGM_MAT at offset 0X80)"
		set msg "$msg\n\nValue in file: $end_address(0)"
		set msg "$msg\nCorrect value: $cmdtable_end"
		set msg "$msg\n\nIfoEdit doesn't upgrade this value when you add or remove"
		set msg "$msg\ncommands in the First-Play PGC."
		set msg "$msg\n\nThe correct value will be updated when you save the DVD."
		show_warning -title "PgcEdit: Open DVD" -icon error -message $msg -parent .
		# finally, fix the error!
		set end_address(0) $cmdtable_end
		tolog "Fixing End byte of VMGI_MAT: $end_address(0) -> $cmdtable_end"
		set ::dvd_modified true
	}

	for {set i 1} {$i < 8} {incr i} {
		set start_address($i) [expr {[get_int data [expr {(0XC0 + (4 * $i))}] 4] * 0X800}]
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VMGM_VOBU_ADMAP has the end_byte at position 0, and
			# VMG_TXTDT_MG at pos 16! All other tables at pos 4.
			switch $i {
				5 {set end_byte_pos 16}
				7 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr {$start_address($i) + \
					[get_int data [expr {$start_address($i)+$end_byte_pos}] 4]}]
		}
	}
	set i 0
	foreach p {VMGM_MAT VMG_TT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT \
			VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP} {
		tolog "Extracting $p."
		# Try to fix the problems with wrong end of table pointers
		set realendaddress [real_endbyte $p $start_address($i) $end_address($i) data]
		if {$realendaddress > 0} {
			set ifoaddress [expr $end_address($i) - $start_address($i)]
			if {$ifoaddress < $realendaddress} {
				set msg "Warning: the table $p of the VMG has a wrong end byte address!"
				set msg "${msg}\nValue in IFO: $ifoaddress, computed value: $realendaddress."
				set msg "${msg}\nPlease save the DVD to fix it!"
				tk_messageBox -title "Read DVD" -message $msg -icon warning
				set ::dvd_modified true
			}
			set end_address($i) [expr $start_address($i) + $realendaddress]
		}

		if {$end_address($i) > 0} {
			set ${sectorsvar}(0,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}(0,$p) ""
		}
		incr i
	}
	# remove VMGM_VOBU_ADMAP and VMGM_C_ADT tables when there are no entries
	# in the table (Needed for DVDShrink)
	if {[string length [subst \$${sectorsvar}(0,VMGM_C_ADT)]] == 8} {
		set ${sectorsvar}(0,VMGM_C_ADT) ""
		tolog "Removing empty VMGM_C_ADT table."
	}
	if {[string length [subst \$${sectorsvar}(0,VMGM_VOBU_ADMAP)]] == 4} {
		set ${sectorsvar}(0,VMGM_VOBU_ADMAP) ""
		tolog "Removing empty VMGM_VOBU_ADMAP table."
	}

	parse_VMG_IFO $sectorsvar $pgcsvar $menusvar

	return 0
}

#############################################################################
## Procedure:  parse_VMG_IFO
## process the data of the VIDEO_TS.IFO
proc parse_VMG_IFO {sectorsvar pgcsvar menusvar} {
	##### Number of title sets
	tolog "Parsing VMG IFO."
	set num_vts [get_int ${sectorsvar}(0,VMGM_MAT) 0X3E 2]
	set ${pgcsvar}(numvts) $num_vts
	set ${menusvar}(numvts) $num_vts	;#!!!!!!
	tolog "Found $num_vts VTSs"

	##### VMGM_MAT/FP-PGC
	#dprint 2 "\nFirst Play PGC"
	set fp_pgc_start [get_int ${sectorsvar}(0,VMGM_MAT) 0X84 4]
	tolog "Parsing FP-PGC."

	# Fix the No FP-PGC error
	if {$fp_pgc_start == 0} {
		tolog "Creating missing FP-PGC."
		set msg "Warning: There is no First-Play PGC in your VMG!"
		set msg "$msg\n(The First-Play PGC start byte value is 0)"
		set msg "$msg\n\nA new, blank First-Play PGC will be created."
		show_warning -title "PgcEdit: Open DVD" -icon error -message $msg -parent .

		# get frame rate from VMGM domain streams attributes
		set dsa [get_int ${sectorsvar}(0,VMGM_MAT) 0x100 2]
		if {$dsa==0} {
			# get video standard
			set videostandard [::streams::get_video_standard "New First-Play PGC" false "Please select the video standard for the new First-Play PGC"]
			if {$videostandard=="NTSC"} {
				set fps 192
			} else {
				set fps 64
			}
		} else {
			if {[expr $dsa & 0x1000]==0} {
				set fps 192
			} else {
				set fps 64
			}
		}

		# build a new FP-PGC from scratch
		set ${pgcsvar}(0,0) {{* pre} {30 02 00 00 00 01 00 00} {* post} {* cell}}
		set tmp [string repeat "\000" 236]
		set_int tmp 4 $fps 4	;# fix Playback Time to legal value (PAL or NTSC)
		set ${pgcsvar}(0,0,header) $tmp
		# fix the FP-PGC Start Byte in VMGM_MAT
		set_int ${sectorsvar}(0,VMGM_MAT) 0X84 \
				[expr {[get_int ${sectorsvar}(0,VMGM_MAT) 0X80 4]+1}] 4

		set ::dvd_modified true
	} else {
		# Get the FP-PGC commands and header.
		set ${pgcsvar}(0,0) [getCommandsFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start pgc 0 0 0]
		set ${pgcsvar}(0,0,header) [getHeaderFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start]
		set ${pgcsvar}(0,0,pttlist) ""
	}
	# Init the FP-PGC values.
	set ${pgcsvar}(0,numpgcs) 1
	set ${pgcsvar}(0,0,numprograms)  0
	set ${pgcsvar}(0,0,programmap)   ""
	set ${pgcsvar}(0,0,numcells)     0
	set ${pgcsvar}(0,0,cellpbtable)  ""
	set ${pgcsvar}(0,0,cellpostable) ""

	##### VMGM_PGCI_UT
	#dprint 2 "\nVMGM_PGCI_UT"
	parse_PGCI_UT ${sectorsvar}(0,VMGM_PGCI_UT) 0 ${menusvar}

	return 0
}

proc parse_VMG_TT_SRPT {sectorsvar pgcsvar ttnsvar} {
	##### get the table of titles from the VMG_TT_SRPT
	set data [subst $${sectorsvar}(0,VMG_TT_SRPT)]
	if {$data != ""} {
		set numtitles [get_int data 0 2]
		tolog "Found $numtitles Titles."
		for {set t 1} {$t <= $numtitles} {incr t} {
			set t_idx [expr {14+(($t-1)*12)}]
			# HD-DVD bug: vtsnum and ttnnum are 0!
			set vtsnum [get_int data $t_idx 1]
			set ttnnum [get_int data [expr {$t_idx+1}] 1]
			if {[catch {set numpgcs [subst $${pgcsvar}($vtsnum,numpgcs)]}]} {
				set numpgcs 0
			}
			set found 0
			for {set i 1} {$i <= $numpgcs} {incr i} {
				if {[subst $${pgcsvar}($vtsnum,$i,TTN)] == $ttnnum && \
							[expr [get_int ::raw_sectors($vtsnum,VTS_PGCI) [expr {$i*8}] 1] & 0x80]>0} {
					if {[info exists ${pgcsvar}($vtsnum,$i,Title)]} {
						set t1 [subst \$${pgcsvar}($vtsnum,$i,Title)]
						tolog "WARNING!  Titles $t1 and $t are assigned to the same PGC (VTS $vtsnum, PGC $i.)"
						set txt "Warning in VMG_TT_SRPT table!"
						set txt "$txt\nThe Titles $t1 and $t are assigned to the same PGC (VTS $vtsnum, PGC $i.)"
						set txt "$txt\nThis strange authoring might be illegal, but it's not sure."
						if {$t == $numtitles} {
							set txt "$txt\n\nDo you want to remove Title $t from the VMG_TT_SRPT table?"
							set txt "$txt\nDon't forget to save the DVD if you want to permanently keep the fix."
						} else {
							set txt "$txt\n\nSince Title $t is not the last one in the table, PgcEdit cannot remove it."
							set txt "$txt\n\nIf you want, you should remove it with IfoEdit."
						}
						set txt "$txt\nYou will have to verify that is is not called with Info -> Find Jumps to Nowhere."
						set txt "$txt\nIf there are jumps to Title $t, replace them with jumps to Title $t1."
						if {$t == $numtitles} {
							set r [tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
								-type yesno -default no -message $txt]
							if {$r == "yes"} {
								incr numtitles -1
								set_int ${sectorsvar}(0,VMG_TT_SRPT) 0 $numtitles 2
								set ${sectorsvar}(0,VMG_TT_SRPT) [string range \
										[subst \$${sectorsvar}(0,VMG_TT_SRPT)] \
										0 [expr ($numtitles*12)+7]]
								set ::dvd_modified true
							}
						} else {
							tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
								-type ok -message $txt
						}
					} else {
						catch {set ${pgcsvar}($vtsnum,$i,Title) $t}
					}
					set ${pgcsvar}(Title,$t) [list $vtsnum $i]
					set ${pgcsvar}(Title,NumTitles) $t
					set found 1
					break
				}
			}
			if {!$found} {
				set txt "Error in Title Play Map table (VMG_TT_SRPT):"
				set txt "$txt\nTitle $t is assigned to a non existing TTN (VTS $vtsnum, TTN $ttnnum)"
				set txt "$txt\n(or TTN $ttnnum has no Entry PGC defined in the VTS_PGCITI table of VTS $vtsnum)!"
				tolog $txt
				if {$t == $numtitles} {
					set txt "$txt\n\nDo you want to remove Title $t from the Title Play Map table?"
					set txt "$txt\nYou will have to verify that is is not called with Info -> Find Jumps to Nowhere."
					set txt "$txt\nIf there are jumps to Title $t, you will need to remove them."
					set txt "$txt\nDon't forget to save the DVD if you want to permanently keep the fix."
					set r [tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
						-type yesno -default yes -message $txt]
					if {$r == "yes"} {
						incr numtitles -1
						set_int ${sectorsvar}(0,VMG_TT_SRPT) 0 $numtitles 2
						set ${sectorsvar}(0,VMG_TT_SRPT) [string range \
								[subst \$${sectorsvar}(0,VMG_TT_SRPT)] \
								0 [expr ($numtitles*12)+7]]
						set ::dvd_modified true
					}
				} else {
					set txt "$txt\n\nThis problem can be caused by a new form of copy protection."
					set txt "$txt\nPgcEdit can stop parsing the Title Play Map Table right now, but it's dangerous."
					set txt "$txt\nIf you stop the parsing now, PgcEdit will still try to load the DVD, but some titles"
					set txt "$txt\nmight be unavailables.  Take care!"
					set txt "$txt\n\nStop the parsing of the Title Play Map Table now?"
					set r [tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
						-type yesno -default yes -message $txt]
					if {$r == "yes"} {
						incr numtitles -1
						set_int ${sectorsvar}(0,VMG_TT_SRPT) 0 $numtitles 2
						set ${sectorsvar}(0,VMG_TT_SRPT) [string range \
								[subst \$${sectorsvar}(0,VMG_TT_SRPT)] \
								0 [expr ($numtitles*12)+7]]
						set ::dvd_modified true
						break
					}
				}
			}
		}
		# check if all TTNs have an entry PGC
		set oktoall1 0
		for {set vts 1} {$vts <= [subst \$${pgcsvar}(numvts)]} {incr vts} {
			set numttns 0
			for {set pgc 1} {$pgc <= [subst \$${pgcsvar}($vts,numpgcs)]} {incr pgc} {
				if {[subst \$${pgcsvar}($vts,$pgc,TTN)]>$numttns} {
					set numttns [subst \$${pgcsvar}($vts,$pgc,TTN)]
				}
			}
			if {[subst \$${pgcsvar}($vts,numttus)] < $numttns} {
				set ${pgcsvar}($vts,numttus) $numttns
			}
			for {set ttn 1} {$ttn <= $numttns} {incr ttn} {
				if {! [info exists ${ttnsvar}($vts,$ttn)]} {
					set ptt_assigned 0
					for {set pgc 1} {$pgc <= [subst \$${pgcsvar}($vts,numpgcs)]} {incr pgc} {
						if {[subst \$${pgcsvar}($vts,$pgc,TTN)] == $ttn} {
							set ${ttnsvar}($vts,$ttn) $pgc
							set tmp [subst \$${sectorsvar}($vts,VTS_PGCI)]
							set_int tmp [expr $pgc*8] [expr 0x80+$ttn] 1
							set ${sectorsvar}($vts,VTS_PGCI) $tmp

							if {$ptt_assigned < 1} {
								if {[subst \$${pgcsvar}($vts,$pgc,numcells)] > 0 && $ptt_assigned != 1} {
									set ${pgcsvar}($vts,$pgc,pttlist) [list [list 1 1 $ttn]]
									set ptt_assigned 1
								} else {
									set ${pgcsvar}($vts,$pgc,pttlist) [list [list 1 1 $ttn]]
									set ptt_assigned -1
								}
							}

							::main::dvd_modified true false
							break
						}
					}
					if {! $oktoall1} {
						set txt "ERROR: Invalid DVD structure.  Problem in cell playback table of VTS $vts!"
						set txt "$txt\nThere is no entry PGC for the TTN $ttn of the VTS $vts!  This is illegal."
						set txt "$txt\n\nPgcEdit will assign the entry PGC of TTN $ttn to PGC $pgc."
						set txt "$txt\n\nPlease save the DVD!"
						set r [tk_messageBox -message $txt -title "Open DVD: Parse VMG_TT_SRPT" \
								-icon error -type "OK to all|OK" -default ok]
						if {$r == "ok to all"} {set oktoall1 1}
					}
				}
			}
		}

		# Check if at least one Title is defined per VTS and creates one new Title if necessary.
		set oktoall2 0
		for {set vts 1} {$vts <= [subst \$${pgcsvar}(numvts)]} {incr vts} {
			set titlefound false
			for {set pgc 1} {$pgc <= [subst \$${pgcsvar}($vts,numpgcs)]} {incr pgc} {
				if {[info exists ${pgcsvar}($vts,$pgc,Title)]} {
					set titlefound true
					break
				}
			}
			if {! $titlefound} {
				incr numtitles
				if {$numtitles > 99} {
					set txt "ERROR: Invalid DVD structure.  Problem in the VMG_TT_SRPT table!"
					set txt "$txt\nThere is no Title in VTS $vts!"
					set txt "$txt\nThis is illegal, and can cause problems when burning the DVD."
					set txt "$txt\n\nPgcEdit CANNOT create Title $numtitles, assigned to the entry PGC of TTN 1 of VTS $vts,"
					set txt "$txt\nas the limit of 99 Titles in the DVD has been reached!"
					set txt "$txt\n\nPlease try to remove the bad VTS $vts.  If it is totally uncalled, you can probably do"
					set txt "$txt\nthat easily with the DVD -> Delete Uncalled PGCs function."
					tk_messageBox -message $txt -title "Open DVD: Parse VMG_TT_SRPT" -icon error
				} else {
					if {! $oktoall2} {
						set txt "ERROR: Invalid DVD structure.  Problem in the VMG_TT_SRPT table!"
						set txt "$txt\nThere is no Title in VTS $vts!"
						set txt "$txt\nThis is illegal, and can cause problems when burning the DVD."
						set txt "$txt\n\nPgcEdit will create Title $numtitles, assigned to the entry PGC of TTN 1 of VTS $vts."
						set txt "$txt\n\nPlease save the DVD!"
						set r [tk_messageBox -message $txt -title "Open DVD: Parse VMG_TT_SRPT" \
								-icon error -type "OK to all|OK" -default ok]
						if {$r == "ok to all"} {set oktoall2 1}
					}
					set ${pgcsvar}(Title,NumTitles) $numtitles

					set entrypgc [subst \$${ttnsvar}($vts,1)]
					set ${pgcsvar}($vts,$entrypgc,Title) $numtitles
					set ${pgcsvar}(Title,$t) [list $vts $entrypgc]

					set_int data 0 $numtitles 2
					set t_idx [expr {8+(($numtitles-1)*12)}]
					set data [string range $data 0 [expr $t_idx-1]][string repeat \000 12]
					if {[subst \$${pgcsvar}($vts,numpgcs)] > $entrypgc && \
						[subst \$${pgcsvar}($vts,$entrypgc,TTN)] == [subst \$${pgcsvar}($vts,[expr $entrypgc+1],TTN)]} {
						set_int data $t_idx 96 1
					} else {
						set_int data $t_idx 32 1
					}
					set_int data [expr $t_idx+1] 1 1
					set_int data [expr $t_idx+2] 1 2
					set_int data [expr $t_idx+6] $vts 1
					set_int data [expr $t_idx+7] 1 1
					set_int data 4 [expr [string length $data]-1] 4
					set ${sectorsvar}(0,VMG_TT_SRPT) $data
					::main::dvd_modified true false
				}
			}
		}
	}
}

#############################################################################
## Procedure:  read_VTS_IFO
## process the data in one VTS_xx_0.IFO of the DVD directory supplied
proc read_VTS_IFO {filename vtsnum sectorsvar pgcsvar menusvar ttnsvar} {
	# Open and read original IFO file in memory
	set error 0
	set data [readFile $filename binary]
	if {$data == ""} {
		set errmsg "Can't open [file tail $filename]\nor file empty!"
		set error 1
	}
	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VTS"} {
		set errmsg "Error checking IFO file type:\nFile [file tail $filename] is not of correct type DVDVIDEO-VTS."
		set error 1
	}
	if {$error} {
		set error 0
		set msg "$errmsg\n\nDo you want to try to load the BUP file instead?"
		set r [tk_messageBox -title "Read DVD" -message $msg -icon warning \
				-type yesnocancel -default yes -skip LoadBUPinstead]
		if {$r == "cancel"} {
			return 1000
		} elseif {$r == "yes"} {
			set filename [string replace $filename end-2 end BUP]
			set data [readFile $filename binary]
			if {$data == ""} {
				tolog "Can't read [file tail $filename]"
				return 1
			}
			if {[string range $data 0 11] != "DVDVIDEO-VTS"} {
				errmsg 1000 "checking IFO file type" \
						"File \"[file tail $filename]\" is not of correct type DVDVIDEO-VTS."
				return 1000
			}
		}
	}
	set filelen [expr [file size $filename] / 0x800]

	##### split the VTS_xx_0.IFO file data in main tables parts
	tolog "Computing start and end addresses of the tables."
	set start_address(0) 0
	set end_address(0) [get_int data 0X80 4]
	for {set i 1} {$i < 9} {incr i} {
		set sec [get_int data [expr {(0XC4 + (4 * $i))}] 4]
		if {$sec > $filelen} {
			set sec 0
			set start_address($i) 0
		} else {
			set start_address($i) [expr {$sec * 0X800}]
		}
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VTSM_VOBU_ADMAP and VTS_VOBU_ADMAP have the end_byte at position 0, not 4!
			switch $i {
				6 -
				8 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr {$start_address($i) + \
					[get_int data [expr {$start_address($i)+$end_byte_pos}] 4]}]
		}
	}

	# verify number of PGCs in tables
	set numpgcs2 [get_int data $start_address(2) 2] ;# VTS_PGCITI
	set numpgcs  [checknumpgcs [string range $data $start_address(2) end]]
	if {$numpgcs != $numpgcs2} {
		tolog "WARNING!  Number of PGCs in VTS_PGCITI: $numpgcs2.  Found $numpgcs valid PGCs!"
		if {[info exists ::skip_dialog_during_session(invalid_pgcs)]} {
			set r yes
		} else {
			set msg "WARNING: Invalid PGCs in the title domain of VTS $vtsnum"
			set msg "$msg\nMaximum number of valid PGCs: $numpgcs"
			set msg "$msg\nNumber of PGCs declared in VTS_PGCITI: $numpgcs2"
			set msg "$msg\n\nDo you want to remove the invalid PGCs (recommended)?"
			set r [tk_messageBox -title "Read DVD" -message $msg -icon warning \
					-type "Yes to All|Yes|No" -default yes]
			if {$r == "yes to all"} {
				set r yes
				set ::skip_dialog_during_session(invalid_pgcs) true
			}
		}
		if {$r == "yes"} {
			tolog "Removing invalid PGCs..."
			if {$start_address(2)>0} {
				set_int data $start_address(2) $numpgcs 2
			}
			if {$start_address(1)>0} {
				if {[get_int data $start_address(1) 2] > $numpgcs} {
					set_int data $start_address(1) $numpgcs 2
				}
			}
			if {$start_address(4)>0} {
				if {[get_int data $start_address(4) 2] > $numpgcs} {
					set_int data $start_address(4) $numpgcs 2
				}
			}
			set ::dvd_modified true
			tolog "Invalid PGCs removed OK."
		}
	}

	set i 0
	foreach p {VTSI_MAT VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI \
			VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP} {
		tolog "Extracting $p."
		# Try to fix the problems with wrong end of table pointers
		set realendaddress [real_endbyte $p $start_address($i) $end_address($i) data]
		if {$realendaddress > 0} {
			set ifoaddress [expr $end_address($i) - $start_address($i)]
			if {$ifoaddress < $realendaddress} {
				tolog "WARNING!  Wrong End of table address: in IFO: $ifoaddress, computed value: $realendaddress."
				set msg "Warning: the table $p of VTS $vtsnum has a wrong end byte address!"
				set msg "${msg}\nValue in IFO: $ifoaddress, computed value: $realendaddress."
				set msg "${msg}\nPlease save the DVD to fix it!"
				tk_messageBox -title "Read DVD" -message $msg -icon warning
				set ::dvd_modified true
			}
			set end_address($i) [expr $start_address($i) + $realendaddress]
		}

		if {$end_address($i) > 0} {
			set ${sectorsvar}($vtsnum,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}($vtsnum,$p) ""
		}
		incr i
	}
	# remove VTSM_VOBU_ADMAP and VTSM_C_ADT tables when there are no entries
	# in the table (Needed for DVDShrink)
	if {[string length [subst \$${sectorsvar}($vtsnum,VTSM_C_ADT)]] == 8} {
		tolog "Removing empty VTSM_C_ADT table"
		set ${sectorsvar}($vtsnum,VTSM_C_ADT) ""
	}
	if {[string length [subst \$${sectorsvar}($vtsnum,VTSM_VOBU_ADMAP)]] == 4} {
		tolog "Removing empty VTSM_VOBU_ADMAP table."
		set ${sectorsvar}($vtsnum,VTSM_VOBU_ADMAP) ""
	}

	# Fix the number of VOB IDs in DVDShrink reauthored movies
	set data [subst $${sectorsvar}($vtsnum,VTS_C_ADT)]
	if {[string length $data] >= 8} {
		set endbyte [get_int data 4 4]
		set numvodids [get_int data 0 2]
		set realnvids 0
		set lastvid 0
		for {set idx 8} {$idx<$endbyte} {incr idx 12} {
			if {$lastvid != [get_int data $idx 2]} {
				set lastvid [get_int data $idx 2]
				incr realnvids
			}
		}
		if {$numvodids != $realnvids} {
			set txt "WARNING:\nThe number of VOB IDs stored in the VTS_C_ADT table of VTS $vtsnum ($numvodids)"
			set txt "$txt\ndon't match the real number of VOB IDs in the table ($realnvids)."
			set txt "$txt\n\n(This is a known bug of DVDShrink in reauthor mode)"
			set txt "$txt\n\nThe table will be fixed with the correct number of VOB IDs ($realnvids)."
			if {$sectorsvar == "::raw_sectors"} {
				set txt "$txt\nPlease save the DVD!"
			}
			show_warning -title "PgcEdit: Open DVD" -icon warning -message $txt -parent .
			set_int data 0 $realnvids 2
			set ${sectorsvar}($vtsnum,VTS_C_ADT) $data
			set ::dvd_modified true
			tolog "Fixed wrong number of VOBs in VTS_C_ADT: $numvodids for $realnvids VOBs present."
		}
	}
	unset data

	parse_VTS_IFO $vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar
	return 0
}

#############################################################################
## Procedure:  parse_VTS_IFO
## process the data in one VTS_xx_0.IFO
proc parse_VTS_IFO {vtsnum sectorsvar pgcsvar menusvar ttnsvar} {

	tolog "Parsing VTS $vtsnum IFO."

	##### VTS_PGCI
	#dprint 2 "\nVTS_PGCI"
	parse_VTS_PGCI ${sectorsvar}($vtsnum,VTS_PGCI) $vtsnum $pgcsvar

	##### VTS_PTT_SRPT
	parse_VTS_PTT_SRPT ${sectorsvar}($vtsnum,VTS_PTT_SRPT) $vtsnum $pgcsvar $ttnsvar

	##### VTSM_PGCI_UT
	#dprint 2 "\nVTSM_PGCI_UT"
	if {! [parse_PGCI_UT ${sectorsvar}($vtsnum,VTSM_PGCI_UT) $vtsnum $menusvar]} {
		set ${sectorsvar}($vtsnum,VTSM_PGCI_UT) ""
	}

	return 0
}

#############################################################################
## Procedure:  readdvd
## read the DVD in memory, and refresh the interface
## args: directoryname:  Path to the VIDEO_TS directory.
##       sectorsvar, pgcsvar, menusvar: name of variables to store the IFOs in.
proc readdvd {directoryname sectorsvar pgcsvar menusvar ttnsvar} {
	set ::log ""
	tolog "Reading DVD \"$directoryname\"\n"

	set zerovobfiles {}

	# unset old global variables
	unset -nocomplain $sectorsvar $pgcsvar $menusvar $ttnsvar
	unset -nocomplain ::skip_dialog_during_session

	set ::dvd_modified false

	# Fix filenames with lower case characters
	set lowerfound 0
	set files [lsort [glob -directory $directoryname -nocomplain -tails -types f V*.IFO V*.BUP V*.VOB]]
	set bat "[string range [file nativename [file normalize $directoryname]] 0 1]\nCD \"[file nativename [file normalize $directoryname]]\"\n"
	foreach i $files {
		if {! [string is upper [string map ". {} _ {} { } {} 0 {} 1 {} 2 {} 3 {} 4 {} 5 {} 6 {} 7 {} 8 {} 9 {}" $i]]} {
			if {!$lowerfound} {
				set txt "Warning: $i is not in upper case.\n\nFix all IFO, BUP and VOB filenames to upper case?"
				set r [tk_messageBox -title "Read DVD" -message $txt -icon warning -type yesno -default yes -skip ifo_bup_vob_to_upper_case]
				if {$r == "yes"} {
					set lowerfound 1
				} else {
					break
				}
			}
			if {$lowerfound} {
				if {$::tcl_platform(platform) == "windows"} {
					set bat "${bat}REN \"$i\" \"[string toupper $i]\"\n"
				} else {
					if {! [catch {file rename [file join $directoryname $i] [file join $directoryname [string toupper _$i]]}]} {
						file rename [file join $directoryname [string toupper _$i]] [file join $directoryname [string toupper $i]]
					}
				}
			}
		}
	}
	if {$lowerfound && $::tcl_platform(platform) == "windows"} {
		set batfile [file join $::config(tempdir) PgcEdit_toupper_[pid].bat]
		catch {saveFile $batfile auto $bat}
		catch {exec -- $batfile}
		catch {file delete $batfile}
	}

	# Open and read VMG IFO file
	show_msg "Reading VMG IFO." 1 100
	set errcode [read_VMG_IFO [file join $directoryname "VIDEO_TS.IFO"] \
			$sectorsvar $pgcsvar $menusvar]
	if {$errcode > 0} {
		return $errcode
	} else {
		set filename [file join $directoryname "VIDEO_TS.VOB"]
		if {[file exists $filename] && [file size $filename]==0} {
			set zerovobfiles 0
		}

		set numvts [subst $${pgcsvar}(numvts)]

		# compare stored number of VTS and real number of VTS_XX_0.IFO files.
		for {set vtsnum2 1} {true} {incr vtsnum2} {
			if {! [file exists [file join $directoryname "VTS_[format %02i $vtsnum2]_0.IFO"]]} {
				incr vtsnum2 -1
				break
			}
		}
		if {$vtsnum2 != $numvts} {
			tolog "WARNING!  Number of VTS in VMG IFO: $numvts.  Titlesets on disc: $vtsnum2."
		}
		if {$vtsnum2 < $numvts} {
			set txt "WARNING:\nThe number of VTSs stored in VIDEO_TS.IFO ($numvts)"
			set txt "$txt\ndon't match the real number of Title Sets stored on disk ($vtsnum2)."
			set txt "$txt\n\nYou should manually copy all files related to VTS [expr $vtsnum2+1]"
			if {[expr $vtsnum2+1] < $numvts} {set txt "${txt}-$numvts"}
			set txt "${txt},\nor restore the backup."
			set txt "${txt}\n\nPgcEdit can also assumes that there is only $vtsnum2 VTSs."
			set txt "${txt}\nWarning: If you do so, there is a risk of navigation errors!"
			set txt "${txt}\nCheck your DVD carefully! (Use Info -> Calls Cross References)"
			set txt "${txt}\nShould I assume $vtsnum2 VTSs?"
			set r [tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
					-type yesno -default yes -message $txt]
			if {$r == "yes"} {
				set numvts $vtsnum2
				set ${pgcsvar}(numvts) $vtsnum2
				set ${menusvar}(numvts) $vtsnum2
			}
		}
		if {$vtsnum2 > $numvts} {
			set txt "WARNING:\nThe number of VTSs stored in VIDEO_TS.IFO ($numvts)"
			set txt "$txt\ndon't match the real number of Title Sets stored on disk ($vtsnum2)."
			set txt "$txt\n\nSince all files related to VTS [expr $numvts+1]"
			if {[expr $numvts+1] < $vtsnum2} {set txt "${txt}-$vtsnum2"}
			set txt "${txt} are unreferenced,"
			set txt "${txt}\nit is necessary to remove them from the compilation."
			set txt "${txt}\n\nIf you really need this Title Set, try to import it with one"
			set txt "$txt\nof the \"Import title clip\" or the \"Import VTS Titles\" and"
			set txt "$txt\n\"Import Menu\" functions."
			set txt "${txt}\n\nMove all unreferenced VTS to the special backup folder\n\"PgcEdit_backup_unreferenced_VTS\"?  (Recommended!)"
			set r [tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
					-type yesno -default yes -message $txt]
			if {$r == "yes"} {
				if {$::config(backup_in_root)} {
					set backupdir [file join [file dirname $directoryname] PgcEdit_backup_unreferenced_VTS]
				} else {
					set backupdir [file join $directoryname PgcEdit_backup_unreferenced_VTS]
				}
				set rc 0
				if {! [file exists $backupdir]} {
					set rc [catch {file mkdir $backupdir}]
				}
				if {$rc} {
					set txt "Can't create backup folder"
					set txt "${txt}\n\"$backupdir\"!"
					set txt "${txt}\nThe target folder is probably read-only."
					tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . \
							-type ok -message $txt
				} else {
					for {set vtsnum [expr $numvts+1]} {$vtsnum<100} {incr vtsnum} {
						catch {file rename -force [file join $directoryname "VTS_[format %02i $vtsnum]_0.IFO"] $backupdir}
						catch {file rename -force [file join $directoryname "VTS_[format %02i $vtsnum]_0.BUP"] $backupdir}
						for {set v 0} {$v<10} {incr v} {
							catch {file rename -force [file join $directoryname "VTS_[format %02i $vtsnum]_$v.VOB"] $backupdir}
						}
					}
				}
			}
		}

		# read all VTS
		for {set vtsnum 1} {$vtsnum <= $numvts} {incr vtsnum} {
			tolog "\nReading VTS $vtsnum IFO."
			show_msg "Reading VTS $vtsnum IFO." $vtsnum $numvts
			set errcode [read_VTS_IFO \
					[file join $directoryname "VTS_[format %02i $vtsnum]_0.IFO"] \
					$vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar]
			set filename [file join $directoryname "VTS_[format %02i $vtsnum]_0.VOB"]
			if {[file exists $filename] && [file size $filename]==0} {
				lappend zerovobfiles $vtsnum
			}
			# missing VTS
			if {$errcode != 0} {
				tolog "ERROR: Could not open the VTS $vtsnum IFO file."
				break
			}
		}
		tolog "\nParsing VMG_TT_SRPT table."
		parse_VMG_TT_SRPT $sectorsvar $pgcsvar $ttnsvar
	}

	if {$errcode != 0} {
		set numvts [expr $vtsnum -1]
		set ${pgcsvar}(numvts) $numvts
		set ${menusvar}(numvts) $numvts
		set txt "ERROR: Cannot read VTS $vtsnum!"
		set txt "$txt\nAs this problem can be caused by a new form of copy protection,"
		set txt "$txt\nPgcEdit will ignore VTS $vtsnum and all subsequent VTS."
		set txt "$txt\n\nYou should try to save the DVD, and if you get a warning about"
		set txt "$txt\nsomething wrong in VMG_TT_SRPT, answer Yes.  Then, reload"
		set txt "$txt\nthe DVD, and answer Yes if there is a wrong number of VTSs in"
		set txt "$txt\nVIDEO_TS.IFO."
		set txt "$txt\n\nAnyway, be extremely cautions, and verify carefully the navigation!"
		set txt "$txt\n(Use Info -> Find Jumps to Nowhere to verify if everything is OK.)"
		tk_messageBox -icon error -title "PgcEdit: Read DVD" -parent . -message $txt
		set errcode 0
	}

	# Read menus buttons informations
	if {[info exists ::buttons::dvd_reopened]} {
		set txt "DVD reopened OK.\n\nDo you want to parse the menu buttons again?\n"
		set txt "$txt\nYou MUST answer yes if a VOB file has been modified!"
		if {[tk_messageBox -icon question -title "PgcEdit: Reopen DVD" -parent . \
				-type yesno -default yes -message $txt -skip "Reopen DVD: Parse menu buttons again?"] == "yes"} {
			set parsebuttons true
			set ::buttons::dvd_reopened 1
		} else {
			set parsebuttons false
			for {set vts 0} {$vts <= $numvts} {incr vts} {
				::menued::add_menu_buttons_marker $vts
			}
			set ::buttons::dvd_reopened -1
		}
	} else {
		set parsebuttons true
	}
	if {$parsebuttons} {
		tolog "\nReading menu button informations in VOBs..."
		namespace eval ::buttons {}
		namespace delete ::buttons
		namespace eval ::buttons {}
		unset -nocomplain ::buttons::butinfo
		::menued::get_buttons_all_menu_domains $directoryname
	}

	if {[llength $zerovobfiles] > 0 && $errcode == 0} {
		set txt "WARNING: There are empty (0 byte) VOB files in the menu domains:\n"
		foreach i $zerovobfiles {
			if {$i == 0} {
				set txt "${txt}VMGM, "
			} else {
				set txt "${txt}VTSM $i, "
			}
		}
		set txt "[string trim $txt ", "]\nAlthough it is legal, Nero will complain.\n"
		set txt "$txt\nDelete the empty VOB file"
		if {[llength $zerovobfiles] > 1} {set txt "${txt}s"}
		set txt "${txt}, and fix the references in the IFOs now?\n"
		set txt "${txt}(You will need to save the DVD!)"
		if {[tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
				-type yesno -default yes -message $txt] == "yes"} {
			foreach vtsnum $zerovobfiles {
				if {$vtsnum == 0} {
					set name "VIDEO_TS.VOB"
					set tablesubname "VMGM"
					set mattablename VMGM_MAT
				} else {
					set name "VTS_[format %02i $vtsnum]_0.VOB"
					set tablesubname "VTSM"
					set mattablename VTSI_MAT
				}
				if {[deletefile [file join $directoryname $name] "PgcEdit: Read DVD" .]} {
					break
				} else {
					# remove the VTS_C_ADT and VTS_VOBU_ADMAP tables
					tolog "Deleting $name and removing ${tablesubname}_C_ADT and ${tablesubname}_VOBU_ADMAP of VTS $vtsnum."
					set ${sectorsvar}($vtsnum,${tablesubname}_C_ADT) ""
					set ${sectorsvar}($vtsnum,${tablesubname}_VOBU_ADMAP) ""
					# clear domain streams attributes
					for {set i 256} {$i <= 347} {incr i} {
						set_int ::raw_sectors($vtsnum,$mattablename) $i 0 1
					}
				}
				set ::dvd_modified true
			}
		}
	}

	set ::tmap::need_new_tmap_list {}
	unset -nocomplain ::skip_dialog_during_session(invalid_pgcs)

	#. configure -cursor $oldcursor
	show_msg ""
	tolog "\nDVD loaded OK.\n"
	return $errcode
}

proc reload_buttons_infos {vts} {
	update idletasks
	if {$vts == 0} {set vtsname "VMGM"} else {set vtsname "VTSM $vts"}
	set txt "You have modified the $vtsname VOB file.\n"
	set txt "${txt}The menu buttons information will be updated\n"
	set txt "${txt}only when you load the DVD again.\n\n"
	set txt "${txt}Do you want to save the DVD and reload it now?"
	set r [tk_messageBox -title "PgcEdit: Button Information" -icon question \
		-message $txt -parent . -type yesno -default yes -skip reload_buttons_infos]
	if {$r == "yes"} {
		::main::Write_DVD false
		::main::ReRead_DVD false
	}
}

## read a specific IFO file.
## return 0 if file is a VMG VIDEO_TS.IFO
## return the number of the VTS if file is a VTS_XX_0.IFO file
## return -1 upon error
proc readdvdfile {filename sectorsvar pgcsvar menusvar ttnsvar} {
	unset -nocomplain ::skip_dialog_during_session(invalid_pgcs)
	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return -1
	}
	fconfigure $in -translation binary
	set errnum [catch {set data [read $in 12]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return -1
	}
	close $in

	switch $data {
		"DVDVIDEO-VMG" {
			if {[read_VMG_IFO $filename $sectorsvar $pgcsvar $menusvar] != 0} {
				return -1
			} else {
				return 0
			}
		}
		"DVDVIDEO-VTS" {
			if {[catch {set vtsnum [expr {[string trimleft [string range [file tail $filename] 4 5] "0"] * 1}]}] != 0} {
				errmsg 1000 "getting VTS number from filename" "Invalid filename \"[file tail $filename]\""
				return -1
			}
			if {[read_VTS_IFO $filename $vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar] != 0} {
				return -1
			} else {
				return $vtsnum
			}
		}
		default {
			errmsg 1000 "checking IFO file type" \
					"File \"$filename\" is not of correct type DVDVIDEO-VMG or DVDVIDEO-VTS"
			return -1
		}
	}
}

} ;# end namespace ::readdvd
