#############################################################################
# winshellext.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# Set/remove the Windows shell associations.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################


####### freeWrap bug fix by r0lZ:

namespace eval ::freewrap {

# Gets all the commands assocated with a extension
# Example: shell_getCmds file.txt => {open print}
proc shell_getCmds {file} {
  set extension [file extension $file]
  if {[catch {set fileType [registry get "HKEY_CLASSES_ROOT\\$extension" ""]} err_str]} {
	puts $err_str; return; # No assocation or fileType
     }
  if {[catch {set cmds [registry keys "HKEY_CLASSES_ROOT\\$fileType\\shell"]} err_str]} {
	puts $err_str; return ; # No commands assocated with file Type
     }
  return $cmds
}

# Gets the implimentation of command given a file extension
# Example: shell_getCmd_imp test.txt open => C:\WINDOWS\NOTEPAD.EXE %1
proc shell_getCmd_imp {file cmd} {
  set extension [file extension $file]
  if {[catch {set fileType [registry get "HKEY_CLASSES_ROOT\\$extension" ""]} err_str]} {
	puts $err_str; return; # No assocation or fileType
     }
  if {[catch {set imp [registry get "HKEY_CLASSES_ROOT\\$fileType\\shell\\$cmd\\command" ""]} err_str]} {
	puts $err_str; return ; # No commands assocated with file Type
     }
  set ret $imp
  return $ret
}

}
####### End of freeWrap bug fix by r0lZ



namespace eval ::winshellext {


if {[info exists ::freewrap::progname]} {
 if {$::tcl_platform(platform) == "windows"} {
	package require registry

	# Return the filetype for an extension
	# Example: shell_assoc_get .txt => "txtfile"
	# Example: shell_assoc_get .NEVER => ""
	proc shell_assoc_get {extension} {
		if {[catch {set ret [registry get "HKEY_CLASSES_ROOT\\$extension" ""]}]} {
			set ret ""
		}
		return $ret
	}

	# return true if an executable is associated with a given extension
	# The exename argument must match (case insensitive) a part of the command line.
	# The catch is supposed to avoid a bug in shell_getCmd_imp
	proc shell_is_associated {extension exename} {
		if [::freewrap::shell_assoc_exist $extension] {
			foreach c [::freewrap::shell_getCmds $extension] {
				if {![catch {set cmd [::freewrap::shell_getCmd_imp $extension $c]}]} {
					if {[string first [string tolower $exename] [string tolower $cmd]] >= 0} {
						return 1
					}
				}
			}
		}
		return 0
	}

	# remove (all) the association(s) of an executable with a given extension.
	# The exename argument must match (case insensitive) a part of the command line.
	# The if {$c != "shell"} and the catch are supposed to avoid a bug in shell_getCmd_imp
	proc shell_remove_assoc {extension exename} {
		if [shell_is_associated $extension $exename] {
			set cmdnames [::freewrap::shell_getCmds $extension]
			set assoc [shell_assoc_get $extension]
			foreach c $cmdnames {
				if {![catch {set cmd [::freewrap::shell_getCmd_imp $extension $c]}]} {
					if {[string first [string tolower $exename] [string tolower $cmd]] >= 0} {
						registry delete "HKEY_CLASSES_ROOT\\$assoc\\shell\\$c"
					}
				}
			}
		}
		if {[::freewrap::shell_getCmds $extension] == ""} {
			registry delete "HKEY_CLASSES_ROOT\\[shell_assoc_get $extension]"
			registry delete "HKEY_CLASSES_ROOT\\$extension" ""
		}
	}

	# Associate a command with a given extension.
	# - The extension is the FileType extension (with the dot, case insensitive)
	# - The filetypelabel must describe the file type
	#   (used only if the FileType must be created).
	# - The filetypename must be the symbolic filetype name, in one word.
	# - The cmdname must be "open", "print", "quickview" (in lower case)
	#   or any other symbolic command name.
	# - The cmdlabel is the label displayed in the Windows context popup menu
	#   (ignored for "open", "print" and "quickview").
	# - The cmdline is the command line executed by Windows.
	# Examples:
	# ::winshellext::shell_add_assoc .XXX "XXX file" MyProg open "" \
	#		"C:\\program files\\MyProg\\MyProg.exe \"%1\""
	# ::winshellext::shell_add_assoc .XXX "XXX file" MyProg MyProg "Edit with MyProg" \
	#		"C:\\program files\\MyProg\\MyProg.exe \"%1\""
	proc shell_add_assoc {extension filetypelabel filetypename cmdname cmdlabel cmdline} {
		if [::freewrap::shell_assoc_exist $extension] {
			set assoc [shell_assoc_get $extension]
		} else {
			set assoc ""
		}
		if {$assoc == ""} {
			set assoc "${filetypename}${extension}"
			::freewrap::shell_fileExtension_setup $extension $assoc
			::freewrap::shell_fileType_setup $assoc $filetypelabel
		}
		switch $cmdname {
			"open" {
				::freewrap::shell_fileType_open  $assoc $cmdline
			}
			"print" {
				::freewrap::shell_fileType_print $assoc $cmdline
			}
			"quickview" {
				::freewrap::shell_fileType_quickView $assoc $cmdline
			}
			default {
				::freewrap::shell_fileType_addAny_cmd  $assoc $cmdname $cmdline
				::freewrap::shell_fileType_setMenuName $assoc $cmdname $cmdlabel
			}
		}
	}

	# high level routine to associate PgcEdit with the given extension.
	# arguments:
	# extension (IFO, BUP or VOB, without dot, case sensitive)
	proc associate_PgcEdit {extension} {
		if {$::config(assoicated_with_$extension)} {
			switch $extension {
				"IFO" {set descr "DVD Navigation File"}
				"BUP" {set descr "DVD Backup of IFO File"}
				"VOB" {set descr "DVD VOB MPEG Video"}
			}
			if [catch {set actualopencmd [::freewrap::shell_getCmd_imp ".$extension" open]}] {
				set actualopencmd ""
			}
			if {$actualopencmd != ""} {
				set warning "\nWARNING: If you answer \"yes\", the current \"open\" menu command will be overwritten !"
				set warning "$warning\nCurrent open command:\n$actualopencmd"
				set default "no"
			} else {
				set warning "\n(There is no default \"open\" command currently defined for $extension files.)"
				set default "yes"
			}
			set txt "An \"Open DVD with PgcEdit\" menu item will be added on the context menu for $extension files.\n"
			set txt "$txt\nDo you also want to launch PgcEdit when you double-click on a $extension file ?"
			set txt "$txt$warning"
			set r [tk_messageBox -icon question -title "PgcEdit: File association" \
				-type yesnocancel -default $default -message $txt -parent .]
			switch $r {
				"yes" {
					::winshellext::shell_add_assoc ".$extension" $descr PgcEdit \
							PgcEdit "Open DVD with PgcEdit" "$::argv0 \"%1\""
					::winshellext::shell_add_assoc ".$extension" $descr PgcEdit \
							open "" "$::argv0 \"%1\""
				}
				"no" {
					::winshellext::shell_add_assoc ".$extension" $descr PgcEdit \
							PgcEdit "Open DVD with PgcEdit" "$::argv0 \"%1\""
				}
				"cancel" {
					set ::config(assoicated_with_$extension) 0
				}
			}
		} else {
			shell_remove_assoc ".$extension" "PgcEdit.exe"
		}
	}

	# high level routine to create or remove a shortcut icon to PgcEdit
	# arguments:
	# in: desktop, sendto, startmenu, common_startmenu, startmenuprograms, common_startmenuprograms, anydir, or a directory path
	# create: true=create icon, false=remove icon
	proc create_shortcut_to_PgcEdit {in create} {
		set dir [specialdir_to_path $in]
		if {$dir == ""} {return false}
		set dir [file nativename $dir]

		if {$create} {
			::freewrap::shortcut "$dir\\PgcEdit.lnk" \
					-objectPath [file nativename $::argv0] \
					-description "Edit the PGC commands of DVD's IFO files"
		} else {
			deletefile "$dir\\PgcEdit.lnk" "PgcEdit: Remove Shortcut" .
		}
	}

	# return whether there is a shortcut to PgcEdit in the specified location
	# in: desktop, sendto, startmenu, common_startmenu, startmenuprograms, common_startmenuprograms, anydir, or a directory path
	proc is_shortcut_to_PgcEdit {in} {
		set dir [specialdir_to_path $in]
		if {$dir == ""} {return false}
		set dir [file nativename $dir]

		return [file exists "$dir\\PgcEdit.lnk"]
	}

	proc specialdir_to_path {in} {
		switch $in {
			desktop {
				if {[catch {set tmp [registry get "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" "Desktop"]}]} {
					set tmp [::freewrap::getSpecialDir DESKTOP]
				}
				return [file nativename [file normalize $tmp]]
			}
			startmenu {
				if {[catch {set tmp [registry get "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" "Startup"]}]} {
					set tmp [file join [::freewrap::getSpecialDir STARTMENU] "Programs"]
				} else {
					set tmp [file dirname $tmp]
				}
				return [file nativename [file normalize $tmp]]
			}
			common_startmenu {
				set tmp [::freewrap::getSpecialDir COMMON_STARTMENU]
				if {$tmp != "" && [file exists [file join $tmp "Programs"]]} {
					return [file nativename [file normalize $tmp]]
				} else {
					return [specialdir_to_path startmenu]
				}
			}
			startmenuprograms {
				return [file join [specialdir_to_path startmenu] "Programs"]
			}
			common_startmenuprograms {
				return [file join [specialdir_to_path common_startmenu] "Programs"]
			}
			sendto  {
				if {[catch {set tmp [registry get "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" "SendTo"]}]} {
					set tmp [::freewrap::getSpecialDir SENDTO]
				}
				return [file nativename [file normalize $tmp]]
			}
			anydir {
				set initialdir [specialdir_to_path common_startmenu]
				return [file nativename [file normalize [tk_chooseNewDirectory -parent . -initialdir $initialdir \
						-title "Create shortcut icon to PgcEdit in folder:"]]]
			}
			default {return $in}
		}
	}

	proc uninstallPgcEdit {} {
		set r [tk_messageBox -title "Uninstall PgcEdit" -icon question \
			-message "This will uninstall PgcEdit version $::program_version!\n\n\nAre you sure?" \
			-parent . -type okcancel -default cancel]
		if {$r == "cancel"} return
		if [shell_is_associated .IFO PgcEdit.exe] {
			shell_remove_assoc .IFO PgcEdit.exe
		}
		if [shell_is_associated .BUP PgcEdit.exe] {
			shell_remove_assoc .BUP PgcEdit.exe
		}
		if [shell_is_associated .VOB PgcEdit.exe] {
			shell_remove_assoc .VOB PgcEdit.exe
		}
		if [is_shortcut_to_PgcEdit desktop] {
			create_shortcut_to_PgcEdit desktop false
		}
		if [is_shortcut_to_PgcEdit startmenuprograms] {
			create_shortcut_to_PgcEdit startmenuprograms false
		}
		if [is_shortcut_to_PgcEdit common_startmenuprograms] {
			create_shortcut_to_PgcEdit common_startmenuprograms false
		}
		if [is_shortcut_to_PgcEdit sendto] {
			create_shortcut_to_PgcEdit sendto false
		}
		foreach i {pgcedit.cfg pgcedit.tools burn.cfg pgcedit.player.cfg pgcedit.scratchbook.sbk preview.cfg.ini pgcedit.user.tcl hiddendialogs.cfg} {
			deletefile [file join $::config(config_directory) $i] "PgcEdit: Uninstall" .
		}
		foreach i {.pgcedit.cfg .pgcedit.tools .burn.cfg .pgcedit.player.cfg .pgcedit.scratchbook.sbk preview.cfg.ini pgcedit.user.tcl hiddendialogs.cfg} {
			deletefile [file join $::env(HOME) $i] "PgcEdit: Uninstall" .
		}
		deletefile $::config(default_scratchbook_filename) "PgcEdit: Uninstall" .
		deletefile $::config(config_directory) "PgcEdit: Uninstall" .
		set progdir [file nativename [file dirname $::argv0]]
		catch {file delete -force [file join $progdir "bin"]}
		catch {file delete -force [file join $progdir "doc"]}
		catch {file delete -force [file join $progdir "plugins"]}

		show_msg "Bye!"

		set txt "Thanks you for using PgcEdit $::program_version."
		set txt "$txt\n\nUninstall almost done."
		set txt "$txt\nYou should manually delete the PgcEdit.exe file\nand the 'bin' directory and/or the installation folder."
		tk_messageBox -title "Uninstall PgcEdit" -icon info \
				-message $txt -parent .

		catch {exec -- explorer "$progdir" &}

		#bind . <Destroy> {}
		exit 0
	}

	# proc used to encrypt/decrypt PSKill.exe, so that McAfee doesn't complain
	proc encrypt {data} {
		set map {}
		for {set i 1} {$i < 255} {incr i} {
			lappend map [d2c $i 1]
			lappend map [d2c [expr 255-$i] 1]
		}
		return [string map $map $data]
	}

}	;# end if platform == Windows
}	;# end if exists freewrap
}	;# end namespace ::winshellext
