#############################################################################
# main.tcl Copyright 2004, 2005, 2006, 2007 r0lZ
# PgcEdit main window GUI and procedures.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::main {

#############################################################################
## Procedure:  about
## Display the about message
proc about {} {
	set txt "Version $::program_version ($::release_date)"
	set txt "$txt\nCopyright r0lZ, 2004, 2005, 2006, 2007"

	set txt "$txt\n\nPgcEdit is free software (donationware)"
	set txt "$txt\ndistributed under the GPL license (GNU General Public License)"
	set txt "$txt\nUse PgcEdit at your own risk!"

	set txt "$txt\n\nIf you like PgcEdit, please make a donation, to support future development."

	set txt "$txt\n\nUsing Tcl/Tk version $::tcl_patchLevel"
	if {[info exists ::freewrap::progname]} {
		set txt "$txt\nWrapped with FreeWrap $::freewrap::patchLevel"
	}
	set txt "$txt\nRunning on $::tcl_platform(os) $::tcl_platform(osVersion)"

	set txt "$txt\n\nPgcEdit home at VideoHelp: http://www.videohelp.com/~r0lZ/pgcedit/"
	set txt "$txt\nOfficial PgcEdit help and contact forum hosted by Doom9,"
	set txt "$txt\n(http://forum.doom9.org) in category DVD and miniDVD > IFO/VOB Editors."

	set txt "$txt\n\nBased on DVD Informations by Mpucoder: http://www.mpucoder.com/dvd/"
	set txt "$txt\nPreview by Jeanl, based on Jsoto's modification of DVD2AVI by LOLI.J.  Thanks, Jeanl!"

	set txt "$txt\n\nCredits to DVD-Replica for online commands documentation: http://dvd-replica.com/DVD/"

	set txt "$txt\n\nThanks to everybody at Doom9 and Digital Video forums!"
	set txt "$txt\nSpecial thanks to Jeanl and Jsoto for their support,"
	set txt "$txt\nto Mpucoder for sharing his encyclopedic knowledge,"
	set txt "$txt\nto the beta testers and translators, to everybody for the ideas,"
	set txt "$txt\nto the authors of PgcEdit's guides, and to the generous donators."
	set txt "$txt\nIn short: Thanks to everyone!"

	toplevel .about
	wm withdraw .about
	wm title .about "About PgcEdit"
	wm transient .about .
	wm overrideredirect .about 1
	wm resizable .about 0 0

	frame .about.a -borderwidth 1 -relief sunken
	pack  .about.a -side top -fill both -expand 1
	frame .about.a.b -borderwidth 6 -relief ridge
	pack  .about.a.b -side top -fill both -expand 1 -anchor center
	frame .about.a.b.pad1 -height 20
	pack  .about.a.b.pad1 -side top -fill y -expand 1
	label .about.a.b.lab1 -text "PgcEdit" -bd 2 -relief sunken \
			-padx 12 -pady 10 -background #EEEEEE
	pack  .about.a.b.lab1 -side top -padx 20 -anchor center -pady 8
	.about.a.b.lab1 configure -font "\"[lindex [.about.a.b.lab1 cget -font] 0]\" 14"
	label .about.a.b.lab2 -text $txt
	pack  .about.a.b.lab2 -side top -padx 20 -anchor center
	frame .about.a.b.pad2 -height 20
	pack  .about.a.b.pad2 -side top -fill y -expand 1

	update
	if {[winfo reqwidth .about] > [winfo reqheight .about]} {
		set size [winfo reqwidth .about]
	} else {
		set size [winfo reqheight .about]
	}
	wm geometry .about ${size}x$size
	update
	#set wx [expr ([winfo screenwidth  .about]-[winfo reqwidth  .about])/2]
	#set wy [expr ([winfo screenheight .about]-[winfo reqheight .about])/2]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  .about])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight .about])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry .about +$wx+$wy

	wm deiconify .about
	grab set .about
	focus .about
	bind .about <ButtonRelease-1> {destroy .about}
	bind .about <Escape> {destroy .about}
	bind .about <Return> {destroy .about}
}

proc check_new_version {} {
	#set ::program_version 8.2beta1	;# to test!
	if {$::config(check_for_betas)} {
		set url "http://home.scarlet.be/~aige0009/pgcedit/beta/version.txt"
	} else {
		set url "http://download.videohelp.com/r0lZ/pgcedit/version.txt"
	}

	if {[catch {package require http}]} {return 0}

	show_msg "Checking if a new version of PgcEdit is available..."
	set rc [catch {set token [::http::geturl $url]} errstr]
	if {$rc == 0} {
		set rc [::http::ncode $token]
		set newversion [string trim [::http::data $token]]
		set errstr [::http::code $token]
		::http::cleanup $token
		if {$rc == 200} {
			set rc 0
		}
	}
	if {$rc != 0} {
		show_msg "Error checking for new version: $errstr"
		return 0
	}

	set idx [string first BETA [string toupper $newversion]]
	if {$idx>-1} {
		set newversionL [split [string range $newversion 0 [expr $idx-1]] .]
		set b [string range $newversion [expr $idx+4] end]
	} else {
		set newversionL [split $newversion .]
		set b 1000
	}
	while {[llength $newversionL] < 3} {
		lappend newversionL 0
	}
	lappend newversionL $b

	set curversion $::program_version
	set idx [string first BETA [string toupper $curversion]]
	if {$idx>-1} {
		set curversionL [split [string range $curversion 0 [expr $idx-1]] .]
		set b [string range $curversion [expr $idx+4] end]
	} else {
		set curversionL [split $curversion .]
		set b 1000
	}
	while {[llength $curversionL] < 3} {
		lappend curversionL 0
	}
	lappend curversionL $b

	set greater false
	for {set i 0} {$i < 4} {incr i} {
		if {[lindex $newversionL $i] > [lindex $curversionL $i]} {
			set greater true
			break
		}
		if {[lindex $newversionL $i] < [lindex $curversionL $i]} {
			break
		}
	}
	if {$greater} {
		show_msg "PgcEdit version $newversion is available!"
		if {$::tcl_platform(platform) == "windows" && $::config(check_for_betas) && [string first "beta" $newversion] >0} {
			set txt "A new BETA version of PgcEdit is available!"
		} else {
			set txt "A new version of PgcEdit is available!"
		}
		set txt "${txt}\n\nYour version: $::program_version"
		set txt "${txt}\nCurrent version: $newversion"
		set txt "${txt}\n\nNote: please visit also the PgcEdit home page."
		set txt "${txt}\nSome plugins or guides might have been updated also!"
		if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname] && \
				! [catch {package require zvfs}] && ! [catch {package require http}] &&
				[diriswritable [file dirname $::argv0]]} {
			set txt "$txt\n\nPgcEdit can install the new version automatically."
			set txt "$txt\nDo you prefer to download it with your web browser"
			set txt "$txt\nand install it manually?"
			set r [::msgbox::MsgBox -title "PgcEdit updater" -message $txt -parent . \
					-type "Automatic|Manual|Cancel" -default automatic -icon question]
			if {$r == "cancel"} {
				return 0
			}
			set install_mode $r
		} else {
			set txt "${txt}\n\nDo you want to download the new version $newversion now?"
			set r [tk_messageBox -title "PgcEdit updater" -message $txt -parent . \
					-type yesno -default yes -icon question]
			if {$r == "no"} {
				return 0
			}
			set install_mode manual
		}
	} else {
		show_msg "You have the latest version of PgcEdit: v $::program_version ($::release_date)"
		return 0
	}
	if {$::tcl_platform(platform) == "windows"} {
		if {$install_mode == "automatic"} {
			if {[info exists ::dvddir] && $::dvd_modified} {
				set txt "The DVD has been modified."
				set txt "$txt\nPgcEdit will be automatically closed during the update process."
				set txt "$txt\nUnsaved changes will be lost!"
				set txt "$txt\n\nDo you want to save the DVD now?"
				set r [tk_messageBox -title "PgcEdit updater" -message $txt -parent . \
						-type yesno -default yes -icon question]
				if {$r == "yes"} {
					if {$::config(use_preview)} {
						::preview::close_preview
					}
					::main::Write_DVD false
					::utils::save_pgc_comments_with_confirm
					if {$::config(save_setups_on_exit) && $::config(restore_defaults)==false} {
						save_user_prefs
					}
				}
			}
			if {$::config(check_for_betas) && [string first "beta" $newversion] >0} {
				set url "http://home.scarlet.be/~aige0009/pgcedit/beta/PgcEdit_winexe_${newversion}.zip"
			} else {
				set url "http://download.videohelp.com/r0lZ/pgcedit/pgcedit_winexe.zip"
			}
			set tailname [file tail $url]
			show_msg "Downloading PgcEdit v $newversion.  Please wait..." 100
			set rc [catch {set token [::http::geturl $url -binary true]} errstr]
			set errmsg "getting url $url\n$errstr"

			if {$rc == 0} {
				set rc [::http::ncode $token]
				set data [::http::data $token]
				set errstr [::http::code $token]
				if {$rc == 200} {
					set rc 0
				}
				if {$rc >= 300 && $rc < 400} {
					upvar #0 $token state
					set url [lindex $state(meta) [expr [lsearch $state(meta) Location]+1]]
					::http::cleanup $token
					unset data
					set rc [catch {set token [::http::geturl $url -binary true]} errstr]
					if {$rc == 0} {
						set rc [::http::ncode $token]
						set data [::http::data $token]
						set errstr [::http::code $token]
						if {$rc == 200} {
							set rc 0
						}
					}
				}
				set errmsg "http error $rc: $errstr"
				::http::cleanup $token
			}
			if {$rc == 0 && [string length $data] < 2000000} {
				set rc 1
				set errmsg "downloading [file tail $url]"
			}
			if {$rc == 0} {
				set tempdir ""
#				catch {set tempdir [file join $::env(WINDIR) Temp]}
#				if {$tempdir == "" || ! [file isdirectory $tempdir] || ! [diriswritable $tempdir]} {
#					if {[string range [file dirname $::argv0] 1 1]==":"} {
#						set tempdir [file join "[string range [file dirname $::argv0] 0 0]:/" "temp"]
#						if {! [file isdirectory $tempdir]} {
#							catch {file mkdir $tempdir}
#						}
#					}
#				}
				if {[string range [file dirname $::argv0] 1 1]==":"} {
					set tempdir [file join "[string range [file dirname $::argv0] 0 0]:/" "temp"]
					if {! [file isdirectory $tempdir]} {
						catch {file mkdir $tempdir}
					}
				}
				if {$tempdir == "" || ! [file isdirectory $tempdir] || ! [diriswritable $tempdir]} {
					set tempdir [file join [file dirname $::argv0] "temp"]
					if {! [file isdirectory $tempdir]} {
						catch {file mkdir $tempdir}
					}
				}
				if {! [file isdirectory $tempdir] || ! [diriswritable $tempdir]} {
					set tempdir $::config(tempdir)
				}
puts $tempdir
				set zipfilename [file join $tempdir $tailname]
				set rc [saveFile $zipfilename binary $data]
				set errmsg "saving [file nativename $zipfilename]"
			}
			if {$rc == 0} {
				unset data
				set rc [catch {::zvfs::mount $zipfilename /newversionzip} errmsg]
			}
			if {$rc == 0} {
				set data [readFile [file join /newversionzip PgcEdit.exe] binary]
				if {$data == ""} {set rc 1}
				set errmsg "extracting PgcEdit.exe from [file nativename $zipfilename]"
			}
			if {$rc == 0} {
				set tmppgceditexe [file join $tempdir PgcEdit.exe]
				set rc [saveFile $tmppgceditexe binary $data]
				set errmsg "saving [file join $tempdir PgcEdit.exe]"
			}
			if {$rc == 0} {
				unset data
				catch {::zvfs::unmount $zipfilename}
				catch {file delete -force $zipfilename}
				set batchfilename [file join $tempdir Install_PgcEdit.bat]

				set txt "@ECHO OFF"
				set txt "$txt\nset a=1"
				set txt "$txt\n:loop"
				set txt "$txt\nset /a a=%a+1"
				set txt "$txt\nif %a==10000 goto err"
				set txt "$txt\ncopy /B /Y \"[file nativename [file normalize $tmppgceditexe]]\" \"[file nativename [file normalize $::argv0]]\""
				set txt "$txt\nif ERRORLEVEL 1 goto loop"
				set txt "$txt\n:cont"
				set txt "$txt\ndel \"[file nativename [file normalize $tmppgceditexe]]\""
				if {[info exists ::dvddir]} {
					set txt "$txt\n\"[file nativename [file normalize $::argv0]]\" \"[file nativename [file normalize $::dvddir]]\""
				} else {
					set txt "$txt\n\"[file nativename [file normalize $::argv0]]\""
				}
				set txt "$txt\nexit"
				set txt "$txt\n:err"
				set txt "$txt\necho Error copying PgcEdit.exe!"
				set txt "$txt\necho Are you sure you have quitted all instances of PgcEdit?"
				set txt "$txt\npause"
				catch {file delete -force $batchfilename}
				set rc [saveFile $batchfilename auto $txt]
				set errmsg "saving batch file [file nativename $batchfilename]"
			}
			if {$rc == 0} {
				exec -- $batchfilename &
				exit 0
			}
			if {$rc > 0} {
				set txt "PgcEdit cannot be updated automatically."
				set txt "$txt\nReason: Error $errmsg."
				set txt "$txt\n\nDo you want to download the new version of PgcEdit"
				set txt "$txt\nwith your web browser and update PgcEdit manually?"
				set r [tk_messageBox -title "PgcEdit updater" -icon error -message "Error updating PgcEdit:\nError $errmsg."]
				if {$r == "no"} {
					return 0
				}
			}
		}
		set txt "PgcEdit will now launch your web browser and download the new version."
		set txt "$txt\nIt will also open the PgcEdit's installation folder, and close itself."
		set txt "$txt\n\nOpen the ZIP archive, and copy PgcEdit.exe in the installation folder."
		set txt "$txt\nFinally, relaunch PgcEdit."
		tk_messageBox -title "PgcEdit updater" -message $txt -parent . -icon info \
				-skip "Check version: usage"
		catch {exec -- explorer.exe [file nativename [file normalize [file dirname $::argv0]]] &}

		if {$::config(check_for_betas) && [string first "beta" $newversion] >0} {
			set url "http://home.scarlet.be/~aige0009/pgcedit/beta/PgcEdit_winexe_${newversion}.zip"
		} else {
			set url "http://download.videohelp.com/r0lZ/pgcedit/pgcedit_winexe.zip"
		}
		if {[catch {exec -- $::config(web_browser) $url &}]} {
			set ::config(web_browser) ""
		}
		cleanexit 0
	} else {
		if {$::config(platform) == "linux" && [info exists ::freewrap::progname]} {
			if {$::config(check_for_betas) && [string first "beta" $newversion] >0} {
				set url "http://home.scarlet.be/~aige0009/pgcedit/beta/PgcEdit_linux_exe_${newversion}.zip"
			} else {
				set url "http://download.videohelp.com/r0lZ/pgcedit/pgcedit"
			}
		} else {
			if {$::config(check_for_betas) && [string first "beta" $newversion] >0} {
				set url "http://home.scarlet.be/~aige0009/pgcedit/beta/PgcEdit_source_${newversion}.zip"
			} else {
				set url "http://download.videohelp.com/r0lZ/pgcedit/pgcedit_source.zip"
			}
		}
		if {[catch {exec -- $::config(web_browser) $url &}]} {
			set ::config(web_browser) ""
		}
	}
	return 1
}

proc diriswritable {dir} {
	set tempfile [file join $dir "PgcEdit_check_writable.tmp"]
	if {[file exists $tempfile]} {
		if {[catch {file delete $tempfile}]} {
			return false
		}
	}
	if {[catch {set fh [open $tempfile w]}]} {
		return false
	}
	if {$fh == ""} {
		return false
	}
	catch {puts $fh "test"}
	close $fh
	if {[file exists $tempfile]} {
		if {[catch {file delete $tempfile}]} {
			return false
		}
		return true
	}
	return false
}

proc showtextfile {w title file args} {
	set txt "File \"$file\" not found!"
	catch {set txt [string map {"\014" ""} [readFile $file auto]]}
	#showtext $w $title [encoding convertfrom identity $txt] $args
	showtext $w $title $txt $args
}

proc showtext {w title txt args} {
	set winname [string trim $w "."]
	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp
	set currentgrab [grab current]

	set _grab 0
	set _defaultfilename ""
	for {set i 0} {$i < [llength $args]} {incr i} {
		switch -- [lindex $args $i] {
			"grab" -
			"-grab" {
				set _grab 1
			}
			"defaultfilename" -
			"-defaultfilename" {
				incr i
				set _defaultfilename [lindex $args $i]
			}
		}
	}

	if {! [winfo exists $w]} {
		if {! [info exists $searchregexpvar]} {
			set $searchstringvar ""
			set $searchcasevar 0
			set $searchregexpvar "-exact"
		}

		toplevel $w
		wm withdraw $w
		wm title $w $title
		bind $w <Escape> "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"
		wm protocol $w WM_DELETE_WINDOW "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"


		frame $w.a
		pack  $w.a -side top -fill both -expand 1 -padx 4 -pady 4
		text  $w.a.txt -xscrollcommand "$w.a.xs set" \
				-yscrollcommand "$w.a.ys set" \
				-font $::config(fixed_width_font) -wrap none
		scrollbar $w.a.ys -command "$w.a.txt yview"
		scrollbar $w.a.xs -command "$w.a.txt xview" -orient horizontal
		grid  $w.a.txt -column 0 -row 0 -sticky nesw
		grid  $w.a.ys  -column 1 -row 0 -sticky ns
		grid  $w.a.xs  -column 0 -row 1 -sticky ew
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		frame $w.b
		pack  $w.b -side top -pady 2 -fill x
		frame  $w.b.s -bd 2 -relief groove
		pack   $w.b.s -side left -padx 4 -fill x -expand 1
		label  $w.b.s.l -text "Find"
		pack   $w.b.s.l -side left
		entry  $w.b.s.e -textvariable $searchstringvar
		pack   $w.b.s.e -side left -fill x -padx 2 -expand 1
		bind   $w.b.s.e <Return> "::main::searchtext $w -forwards"
		bind   $w.b.s.e $::config(popups_button) "tk_popup $w.searcheditmenu \[winfo pointerx .\] \[winfo pointery .\] 3"
		if {! [winfo exists $w.searcheditmenu]} {
			menu $w.searcheditmenu -tearoff 0
			$w.searcheditmenu add command -label "Copy"  -accelerator "Ctrl+C" -command "clipboard clear; clipboard append \$$searchstringvar"
			$w.searcheditmenu add command -label "Paste" -accelerator "Ctrl+V" -command "if \[catch {set $searchstringvar \[clipboard get\]}\] {bell} else {$w.b.s.e icursor end}"
			$w.searcheditmenu add separator
			$w.searcheditmenu add command -label "Clear" -accelerator "Ctrl+K" -command "set $searchstringvar {}"
		}
		if {$::tcl_platform(platform) == "windows"} {set minwidth 2} else {set minwidth 0}
		button $w.b.s.b1 -text "<" -width 2 -command "::main::searchtext $w -backwards"
		pack   $w.b.s.b1 -side left -padx $minwidth
		button $w.b.s.b2 -text ">" -width 2 -command "::main::searchtext $w -forwards"
		pack   $w.b.s.b2 -side left -padx $minwidth
		frame  $w.b.s.cf
		pack   $w.b.s.cf -side left
		checkbutton $w.b.s.cf.c1 -bd $::config(min_bd) -text "match case" \
				-variable $searchcasevar -pady 0
		pack   $w.b.s.cf.c1 -side top -anchor w
		checkbutton $w.b.s.cf.c2 -bd $::config(min_bd) -text "regular expression" \
				-variable $searchregexpvar -pady 0 -onvalue "-regexp" -offvalue "-exact"
		pack   $w.b.s.cf.c2 -side top -anchor w

		if {$::config(platform) == "mac"} {
			set anchor n
		} else {
			set anchor s
		}
		button $w.b.save -text Save -width 8 -command "::main::savetext $w [list $_defaultfilename]"
		pack   $w.b.save -side left -padx 4 -anchor $anchor
		button $w.b.ok -text Close -width 8 -command "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"
		pack   $w.b.ok -side left -padx 4 -anchor $anchor

		# window geometry
		if {[info exists ::config(${w}_geometry)]} {
			wm geometry $w $::config(${w}_geometry)
		} else {
			update
			set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
			set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
			if {$wx < 10} {set wx 10}
			if {$wy < 22} {set wy 22}
			wm geometry $w +$wx+$wy
		}

		menu $w.editmenu -tearoff 0
		$w.editmenu add command -label "Copy" -accelerator "Ctrl+C" -command "::main::showtext_copy $w"
		$w.editmenu add command -label "Select all" -accelerator "Ctrl+A" -command "::main::showtext_select_all $w"
		$w.editmenu add separator
		$w.editmenu add command -label "Clear all" -command "::main::showtext_clear_all $w"
		bind $w.a.txt $::config(popups_button) "tk_popup $w.editmenu \[winfo pointerx .\] \[winfo pointery .\] 0"
		bind $w <Control-a> "::main::showtext_select_all $w"

		if {$_grab} {
			if {$currentgrab != ""} {
				$w.b.ok configure -command "grab set $currentgrab; destroy $w"
			}
			grab set $w
		}
	}

	$w.a.txt delete 1.0 end
	$w.a.txt insert 1.0 $txt
	$w.a.txt mark set insert 1.0

	wm deiconify $w
	focus $w
}
proc showtext_append {w title txt} {
	if {! [winfo exists $w]} {
		showtext $w $title $txt
	} else {
		$w.a.txt insert end $txt
	}
	$w.a.txt yview moveto 1
	# Don't display the last blank line.
	if {[string range $txt end end] == "\n"} {
		$w.a.txt delete "end - 1 chars" end
		update
		$w.a.txt insert end "\n"
	}
}
proc savetext {w {defaultfilename ""}} {
	# Get input filename interactively
	set filetypes {{"Text files" {*.txt}} {"All files" {*.*}}}
	if {$defaultfilename != ""} {
		set initialpath [file dirname $defaultfilename]
		set initialfile [file tail $defaultfilename]
		set extension   [file extension $defaultfilename]
		if {$extension != "" && [string toupper $extension] != ".TXT"} {
			set filetypes [list [list "[string trim $extension .] files" "*$extension"] {"Text files" {*.txt}} {"All files" {*.*}}]
		}
	} else {
		set initialpath [file dirname $::config(dumpinitialfile)]
		set initialfile ""
		set extension ".txt"
	}
	set filename [tk_getSaveFile \
		-filetypes $filetypes -initialdir $initialpath -initialfile $initialfile \
		-parent $w -title "PgcEdit: save text to file..."]
	if {$filename == ""} {return}
	if {$::tcl_platform(platform) == "windows" && [file extension $filename] == "" && $extension != ""} {
		set filename "${filename}$extension"
	}
	if {$defaultfilename != ""} {
		set ::config(dumpinitialfile) $filename
	}

	saveFile $filename auto [$w.a.txt get 1.0 end]
}
proc searchtext {w direction args} {
	if {$args == ""} {
		set winname [string trim $w "."]
	} else {
		set winname $args
	}

	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp

	set idx [$w.a.txt index insert]
	if {$direction == "-forwards"} {
		set idx "$idx + 1 chars"
	} else {
		set idx "$idx - 1 chars"
	}
	if {[subst $$searchcasevar]} {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-- [subst $$searchstringvar] $idx]} errstr]
	} else {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-nocase -- [subst $$searchstringvar] $idx]} errstr]
	}
	if {$err != 0} {
		show_warning -title "Search error" -icon error \
				-message [string map {": " ":\n"} $errstr] -parent $w -skip "Search error msg"
	} else {
		if {$found != ""} {
			$w.a.txt mark set insert $found
			$w.a.txt see $found
			focus $w.a.txt
		} else {
			bell
		}
	}
}
proc showtext_copy {w} {
	set txt ""
	catch {set txt [$w.a.txt get sel.first sel.last]}
	if {$txt != ""} {
		clipboard clear
		clipboard append $txt
	} else {
		bell
	}
}
proc showtext_select_all {w} {
	focus $w.a.txt
	$w.a.txt tag add sel 1.0 end
}
proc showtext_clear_all {w} {
	$w.a.txt delete 1.0 end
}

#############################################################################
## Procedure:  Create_Backup
## If the backup directory already exists, then ask the user for overwrite, or
## if the ifnotexists arg is true, don't ask and don't overwrite
proc Create_Backup {ifnotexists {quiet true}} {
	set toolname "Create Backup"
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {[file exists $::backupdir]} {
		if {$ifnotexists} {
			return 0
		} else {
			set numvobfiles [glob -nocomplain -tails -directory $::backupdir "*.VOB"]
			set remappinginfo [file exists "$::backupdir/remap_vts_info.tcl"]
			if {$remappinginfo || $numvobfiles>0} {
				set txt "WARNING:  Backup already exists!"
				if {$numvobfiles>0} {
					set txt "$txt\nThere are VOB files in the old backup folder."
				}
				if {$remappinginfo} {
					set txt "$txt\nYou have remapped the titlesets of the DVD\nsince the creation of the last backup."
				}
				set txt "$txt\n\nIt is NOT safe to overwrite this backup folder!"
				set txt "$txt\nYou sgould do an incremental backup instead!"
				set txt "$txt\n\nCreate a new incremental backup?  (recommended)"
				set r [tk_messageBox -title $toolname -icon question \
					-message $txt -parent . -type yesnocancel -default yes]
				if {$r == "cancel"} {return 0}
				if {$r == "yes"} {
					return [Incremental_Backup]
				}
			} else {
				set r [tk_messageBox -title $toolname -icon question \
					-message "Backup already exists!\nOverwrite?" -parent . -type yesno -default no]
				if {$r == "no"} {return 0}
			}
			if {[file exists "$::backupdir/User comments.upc"]} {
				movefile "$::backupdir/User comments.upc" $::dvddir -force $toolname .
			}
			deletefile "$::backupdir/VIDEO_TS.IFO" $toolname .
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				deletefile "$::backupdir/VTS_[format %02i $vts]_0.IFO" $toolname .
			}
			deletefile $::backupdir $toolname .
			set dummy [file exists $::backupdir]	;# wait till directory is deleted
		}
	}
	if {$quiet && ![file writable $::dvddir]} {
		if {[file tail $::dvddir] == "VIDEO_TS"} {
			set root [file dirname $::dvddir]
			if {[string length $root] == 3 && [string range $root 1 2] == ":/"} {
				show_msg "Backup not created: read-only media."
				return 0
			}
		}
	}
	show_msg "Creating backup..." 1 [expr $::pgcs(numvts)+1]
	set err [makedir $::backupdir $toolname .]
	if {! $err} {
		CopyFile "$::dvddir/VIDEO_TS.IFO" $::backupdir -force $toolname .
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			update_progress [expr $vts+1] [expr $::pgcs(numvts)+1]
			CopyFile "$::dvddir/VTS_[format %02i $vts]_0.IFO" $::backupdir -force $toolname .
		}
		if {[file exists "$::dvddir/User comments.upc"]} {
			CopyFile "$::dvddir/User comments.upc" "$::backupdir" -force $toolname .
			movefile "$::dvddir/User comments.upc" "$::backupdir/User comments backup.upc" -force $toolname .
		}

		::menubuttons::save_buts [file join $::backupdir "menubuttons.but"]

		show_msg "Backup created OK."
		return 1
	}
	show_msg "Cannot create backup!"
	return 0
}

#############################################################################
## Procedure:  Incremental_Backup
## If the backup directory already exists, rename it, then call Create_Backup.
proc Incremental_Backup {} {
	set toolname "Incremental Backup"
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {[file exists $::backupdir]} {
		file stat $::backupdir stats
		set newbackupdir "${::backupdir}_[clock format $stats(ctime) -format %y%m%d_%H%M%S]"
		if {[movefile $::backupdir $newbackupdir false $toolname .]} {
			return 0
		}
		if {[Create_Backup false true]} {
			# Conserve user comments
			if {[file exists "$newbackupdir/User comments backup.upc"]} {
				CopyFile "$newbackupdir/User comments backup.upc" $::backupdir -force $toolname .
			}
			if {[file exists "$newbackupdir/User comments.upc"]} {
				CopyFile "$newbackupdir/User comments.upc" $::backupdir -force $toolname .
				CopyFile "$newbackupdir/User comments.upc" "$::backupdir/User comments backup.upc" -force $toolname .
			}
		} else {
			return 0
		}
	} else {
		return [Create_Backup false false]
	}
}

#############################################################################
## Procedure:  Restore_Backup
proc Restore_Backup {{undo_last ""}} {
	set toolname "Restore Backup"
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {! [file exists $::backupdir]} {
		show_warning -title $toolname -icon error \
			-message "PgcEdit_backup not found!" -parent . -skip "Restore backup: backup not found"
		return 0
	}

	# are there any other backup folder more recent than the PgcEdit backup?
	if {$undo_last == ""} {
		set txt "Warning: At least one backup created by another program"
		set txt "$txt\nexists and has been created more recently than the current"
		set txt "$txt\nPgcEdit backup folder:"
		if {[other_more_recent_backup_exists $::backupdir $txt $toolname]} {
			return 0
		}
	}

	# are there any VOB file to restore?
	set vobfiles [lsort [glob -nocomplain -tails -directory $::backupdir "*.VOB"]]
	if {$vobfiles == ""} {
		set vob_action "no vobs"
	} else {
		set txt "There are VOB files in your backup folder."
		set txt "$txt\nDo you want to restore them?  (Recommended)"
		set txt "$txt\n\nCaution: The current VOB files will be overwritten."
		if {$undo_last == ""} {
			set vob_action [::msgbox::MsgBox -title $toolname -message $txt -icon question -default "move vobs back" -type "Move VOBs back|Copy VOBs|Don't restore VOBs|Cancel all" -parent .]
			if {$vob_action=="cancel"} {
				show_msg "$toolname: operation aborted!"
				return 0
			}
		} else {
			set vob_action [::msgbox::MsgBox -title $toolname -message $txt -icon question -default "move vobs back" -type "Move VOBs back|Copy VOBs|Don't restore VOBs" -parent .]
		}
	}

	# Restore IFOs and overwrite BUPs
	show_msg "$toolname: Copying IFOs..." 1 [expr $::pgcs(numvts)+1]
	CopyFile "$::backupdir/VIDEO_TS.IFO" $::dvddir -force $toolname .
	CopyFile "$::backupdir/VIDEO_TS.IFO" "$::dvddir/VIDEO_TS.BUP" -force $toolname .
	for {set vts 1} {$vts <= 99} {incr vts} {
		update_progress [expr $vts+1] [expr $::pgcs(numvts)+1]
		if {[file exists "$::backupdir/VTS_[format %02i $vts]_0.IFO"]} {
			CopyFile "$::backupdir/VTS_[format %02i $vts]_0.IFO" $::dvddir -force $toolname .
			CopyFile "$::backupdir/VTS_[format %02i $vts]_0.IFO" "$::dvddir/VTS_[format %02i $vts]_0.BUP" \
					-force $toolname .
		} else {
			if {$vts > $::pgcs(numvts)} break
		}
	}

	# undo VTS remapping if necessary
	set remappingfile [file join $::backupdir "remap_vts_info.tcl"]
	if {[file exists $remappingfile]} {
		show_msg "$toolname: Remapping VOB files in DVD folder..."
		source $remappingfile
		# rename the VOB files
		remap_vob_files $::dvddir $orig_num_vts undoremapping
		catch {file attributes $remappingfile -permissions +w}
		deletefile $remappingfile
	}

	# restore VOBs from backup folder if needed
	switch $vob_action {
		"move vobs back" {
			set i 0
			foreach f $vobfiles {
				incr i
				show_msg "$toolname: Moving $f" $i [llength $vobfiles]
				movefile [file join $::backupdir $f] $::dvddir -force $toolname .
			}
		}
		"copy vobs" {
			set i 0
			#blocking_msg "Restoring VOB files..."
			foreach f $vobfiles {
				incr i
				set n [file join $::backupdir $f]
				set size [expr round(([file size $n]/(1024*1024*1024.0))*100)/100.0]
				show_msg "$toolname: Copying $f (${size}GB)" $i [llength $vobfiles]
				CopyFile $n $::dvddir -force $toolname .
			}
			#close_blocking_msg
		}
		"don't restore vobs" {
			show_msg "$toolname: VOB files not restored!"
		}
		"cancel all" {
			show_msg "$toolname: operation aborted!"
			return 0
		}
		"no vobs" {}
	}

	# restore PGC comments
	set restore_pgc_comments false
	if {[file exists "$::backupdir/User comments backup.upc"]} {
		set msg "There are user PGC comments and/or GPRM labels in the backup folder."
		set msg "$msg\nThey have been saved when the last backup has been created."
		set msg "$msg\nIf you restore them also, you will lose the commnents and labels defined later."
		set msg "$msg\nBut if you don't restore them, the current comments might be desynchronized,"
		set msg "$msg\nespecially if you have remapped the PGCs or the VTSs since the last backup."
		set msg "$msg\n\nRestore the old Comments and Labels too?"
		set r [tk_messageBox -title $toolname -icon question \
				-message $msg -parent . -type yesno -default yes]
		if {$r == "yes"} {
			set restore_pgc_comments true
			CopyFile "$::backupdir/User comments backup.upc" "$::backupdir/User comments.upc" -force $toolname .
			set ::pgc_comments_modified true
		}
	}

	# reload DVD
	set r [tk_messageBox -title $toolname -icon question \
			-message "Backup restored.\n\nReload DVD in editor?" -parent . \
			-type yesno -default yes -skip "Restore backup: reolad DVD?"]
	show_msg ""
	if {$r == "yes"} {
		if {$restore_pgc_comments} {
			unset -nocomplain ::menus_comments
			unset -nocomplain ::pgcs_comments
		} else {
			::utils::save_pgc_comments_with_confirm no
		}
		set ::pgc_comments_modified false
		ReRead_DVD true true
	}
	return 1
}

proc remap_vob_files {dir numvts upvar_remap_array} {
	upvar $upvar_remap_array remap
	set utilname "Restore backup"

	set vobfiles [lsort [glob -nocomplain -tails -directory $dir "VTS_*.VOB"]]
	foreach f $vobfiles {
		set origvts [string trimleft [string range $f 4 5] 0]
		if {$origvts <= $numvts} {
			movefile [file join $dir $f] [file join $dir tmp_$f] -force $utilname
		}
	}
	foreach f $vobfiles {
		set origvts [string trimleft [string range $f 4 5] 0]
		if {$origvts <= $numvts} {
			set dest VTS_[format %02d $remap($origvts)][string range $f 6 end]
			movefile [file join $dir tmp_$f] [file join $dir $dest] -force $utilname
		}
	}
	return
}

#############################################################################
## Procedure:  Delete_Backup
proc Delete_Backup {} {
	set toolname "Delete Backup"
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {! [file exists $::backupdir]} {
		show_warning -title $toolname -icon error \
			-message "PgcEdit_backup not found!" -parent . -skip "Delete Backup: backup not found"
	} else {
		if {[file exists "$::backupdir/User comments.upc"]} {
			set msg "Warning!"
			set msg "$msg\nAs the PGC comments and GPRM labels are stored in the backup dir,"
			set msg "$msg\nif you continue, you will lose them."
			set msg "$msg\nIf you want to conserve them, export them now to any folder."
			set msg "$msg\n\nExport the current PGC comments and GPRM labels?"
			set r [tk_messageBox -title "Delete backup: User comments" -message $msg \
					-icon question -type yesnocancel -default no]
			if {$r=="cancel"} {
				return 0
			}
			if {$r=="yes"} {
				::utils::export_pgc_comments
			} else {
				set ::pgc_comments_modified true
			}
		}
		file delete -force -- $::backupdir
		show_msg "Backup deleted."
	}

	set backupfiles [glob -directory $::dvddir -nocomplain -tails -- \
			PgcEdit_backup*]
	if {$backupfiles != ""} {
		set txt "There are incremental backups in the VIDEO_TS folder:"
		foreach f $backupfiles {
			set txt "$txt\n    $f"
		}
		set txt "$txt\n\nDo you want to delete them as well?"
		set r [tk_messageBox -message $txt -icon question \
				-title $toolname -type yesno -default yes]
		if {$r == "yes"} {
			foreach f $backupfiles {
				catch [file delete -force -- [file join $::dvddir $f]]
			}
		}
	}

	set backupfiles [glob -directory $::dvddir -nocomplain -tails -- \
			MenuShrinkBackup* VobBlanker_backup* "Copy of *"]
	if {$backupfiles != ""} {
		set txt "There are other backups in the VIDEO_TS folder:"
		foreach f $backupfiles {
			set txt "$txt\n    $f"
		}
		set txt "$txt\n\nDo you want to delete them as well?"
		set r [tk_messageBox -message $txt -icon question \
				-title $toolname -type yesno -default no]
		if {$r == "yes"} {
			foreach f $backupfiles {
				catch [file delete -force -- [file join $::dvddir $f]]
			}
		}
	}

	set dir [file dirname $::dvddir]
	set backupfiles [glob -directory $dir -nocomplain -tails -- \
			PgcEdit_backup* MenuShrinkBackup* VobBlanker_backup* "Copy of *"]
	if {$backupfiles != ""} {
		set txt "There are backups in the DVD root folder:"
		foreach f $backupfiles {
			set txt "$txt\n    $f"
		}
		set txt "$txt\n\nDo you want to delete them as well?"
		set r [tk_messageBox -message $txt -icon question \
				-title $toolname -type yesno -default no]
		if {$r == "yes"} {
			foreach f $backupfiles {
				catch [file delete -force -- [file join $dir $f]]
			}
		}
	}
}

#############################################################################
## replaces the current backup by the last incremental backup, and restore it
## TODO:  check if there are VobBlanker or FixVTS backups newer than the latest incr backup
proc Undo_Incremental_Backup {} {
	set toolname "Undo Incremental Backup"
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}

	set backup_location [file dirname $::backupdir]
	set incrbackups [lsort [glob -nocomplain -directory $backup_location -tails -- PgcEdit_backup_*]]
	# remove special PgcEdit backup made with unreferenced VTSs when loading the DVD for the first time
	set i [lsearch $incrbackups "PgcEdit_backup_unreferenced_VTS"]
	if {$i != -1} {
		set incrbackups [lreplace $incrbackups $i $i]
	}
	if {[llength $incrbackups]==0} {
		show_warning -title $toolname -icon info \
			-message "No incremental backup found!" -parent . -skip "Undo Incr Backup: backup not found"
		return 0
	}
	set lastbackup [lindex $incrbackups end]
	set lastbackupdir [file join $backup_location $lastbackup]

	set txt "Warning: At least one backup created by another program"
	set txt "$txt\nexists and has been created more recently than the last"
	set txt "$txt\nincremental backup folder:"
	if {[other_more_recent_backup_exists $lastbackupdir $txt $toolname]} {
		return 0
	}

	if {[file exists $::backupdir]} {
		if {[file exists [file join $::backupdir "remap_vts_info.tcl"]]} {
			set msg "WARNING!  You must restore the current backup"
			set msg "$msg\nbefore restoring the last incremental backup,"
			set msg "$msg\nas the titlesets have been remapped when the"
			set msg "$msg\ncurrent backup was created.  Restoring the last"
			set msg "$msg\nincremental backup is impossible now, as the VOB"
			set msg "$msg\nfiles will not match the IFOs any more!"
			set msg "$msg\n\nRestore the current backup now?  (Mandatory!)"
			set msg "$msg\n(If you answer Yes, you can call the Undo Last"
			set msg "$msg\nincremental backup function again later.)"
			set r [tk_messageBox -title $toolname -icon error \
				-message $msg -parent . -type okcancel -default ok]
			if {$r=="ok"} {
				show_msg "Undo Incremental Backup cancelled.  Restoring current backup (with VTS remapping) instead..."
				Restore_Backup
			} else {
				show_msg "Undo Incremental Backup cancelled."
			}
			return 0
		}

		set showwarning false
		set vobfilesinbackup [glob -nocomplain -directory $::backupdir -tails -- *.VOB]
		if {[llength $vobfilesinbackup]>0} {
			set showwarning true
		}
		if {$showwarning} {
			set msg "You will REPLACE your current PgcEdit_backup"
			set msg "$msg\nfolder with the last incremental backup"
			set msg "$msg\n\"$lastbackup\""
			if {[llength $vobfilesinbackup]>0} {
				set msg "$msg\n\nWARNING!  It is safer to restore the current"
				set msg "$msg\nbackup folder before removing it, as it contains"
				set msg "$msg\nsome modified VOB files!  THEY WILL BE LOST!"
				set msg "$msg\nYou should be sure that you have already"
				set msg "$msg\nrestored the current backup!"
				set msg "$msg\n\nRestore the current backup now?  (Recommended)"
				set msg "$msg\n(The last incremental backup will NOT be undone.)"
				set r [::msgbox::MsgBox -title $toolname -icon warning \
					-message $msg -parent . -type yesnocancel -default yes -skip "Replace backup folder"]
				if {$r=="yes"} {
					show_msg "Undo Incremental Backup cancelled.  Restoring current backup (with VOBs) instead..."
					Restore_Backup
					return 0
				}
			} else {
				set r [tk_messageBox -title $toolname -icon question \
					-message $msg -parent . -type okcancel -default ok]
			}
			if {$r == "cancel"} {
				show_msg "Undo Incremental Backup cancelled."
				return 0
			}
		}
		# everything is OK?  Continue: delete current backup
		if {[deletefile $::backupdir $toolname .]} {
			return 0
		}
	}

	# move last incr backup
	if {[movefile $lastbackupdir $::backupdir -force $toolname .]} {
		return 0
	}
	if {[file exists "$::backupdir/User comments backup.upc"]} {
		CopyFile "$::backupdir/User comments backup.upc" "$::backupdir/User comments.upc" -force $toolname .
	}
	show_msg "Last backup folder replaced by $lastbackup."
	# and restore it
	Restore_Backup undo_last
	return
}

#############################################################################
## Procedure: Check if there are VobBlanker or FixVTS backups newer than the
## file passed in arguments.  Return a list of those alien backups.
proc other_more_recent_backup_exists {pgceditbackupfolder start_of_message toolname} {
	set abortmsg "$toolname: operation aborted!"
	file stat $pgceditbackupfolder stats
	set pgceditbackupctime $stats(ctime)
	set alienbackups {}
	if {![catch {set m [glob -directory $::dvddir -tails -- *\[Bb\]ackup*]}]} {
		foreach i $m {
			if {[string range $i 0 13]!="PgcEdit_backup" && $i!="BlankedVobBackup"} {
				set f [file join $::dvddir $i]
				if {[file isdirectory $f]} {
					if {[glob -nocomplain -directory $f -tails -- *]==""} {
						deletefile $f
					} else {
						file stat $f stats
						if {$stats(ctime)>$pgceditbackupctime} {
							lappend alienbackups [list $stats(ctime) $f]
						}
					}
				}
			}
		}
	}
	if {[file tail $::dvddir]=="VIDEO_TS"} {
		if {![catch {set m [glob -directory [file dirname $::dvddir] -tails -- *\[Bb\]ackup*]}]} {
			foreach i $m {
				if {[string range $i 0 13]!="PgcEdit_backup" && $i!="BlankedVobBackup"} {
					set f [file join [file dirname $::dvddir] $i]
					if {[file isdirectory $f]} {
						if {[glob -nocomplain -directory $f -tails -- *]==""} {
							deletefile $f
						} else {
							file stat $f stats
							if {$stats(ctime)>$pgceditbackupctime} {
								lappend alienbackups [list $stats(ctime) $f]
							}
						}
					}
				}
			}
		}
	}
	# sort list by creation time (reverse order)
	set all [lsort -dictionary $alienbackups]
	set alienbackups {}
	for {set i [expr [llength $all]-1]} {$i>=0} {incr i -1} {
		lappend alienbackups [lindex [lindex $all $i] 1]
	}

	if {[llength $alienbackups] > 0} {
		set txt $start_of_message
		set txt "$txt\n\n[join $alienbackups \n]"
		set txt "$txt\n\nYou should PROBABLY restore it before restoring the"
		set txt "$txt\ncurrent PgcEdit backup.  Take care anyway!"
		set txt "$txt\n\nPgcEdit cannot restore it automatically."
		set txt "$txt\nYou should move the backup files back in the DVD folder"
		set txt "$txt\nmanually, and delete the backup folder before continuing."
		set txt "$txt\nThen, don't forget to reload the DVD in PgcEdit!"
		set txt "$txt\n\nDo you want to restore this backup manually now?"
		set r [tk_messageBox -message $txt -title $toolname -icon warning \
				-type yesnocancel -default yes]
		if {$r=="yes"} {
			if {$::tcl_platform(platform) == "windows"} {
				set bf [lindex $alienbackups 0]
				show_msg "$toolname: opening backup folder $bf in Explorer, and aborting the automatic restore."
				exec -- explorer.exe [file nativename $bf] &
			} else {
				show_msg $abortmsg
			}
			return 1
		} elseif {$r=="cancel"} {
			show_msg $abortmsg
			return 1
		}
	}
	return 0
}

#############################################################################
## Procedure:  Write_DVD
## write the current DVD to disk.
## Get output filename interactively if saveas is true, else use global $::dvddir
proc Write_DVD {saveas} {
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {[info exists ::save_in_progress]} {
		bell
		return 1
	}

	set ::save_in_progress 1
	set export_pgc_comments_auto_mode true
	if {$saveas} {
		set toolname "Save IFOs as"
		set destdir [tk_chooseNewDirectory -initialdir $::dvddir \
			-title "Save IFOs in folder:"]
		if {$destdir == ""} {
			unset ::save_in_progress
			return -1
		}
		#blocking_msg "Saving IFOs..."
		#set ::config(initialdir) $destdir

		set usevobfromcurdir false
		if {[llength [glob -directory $destdir -tails -nocomplain -- *.VOB]] == 0} {
			if {[llength [glob -directory $::dvddir -tails -nocomplain -- *.VOB]] > 0} {
				set usevobfromcurdir true
			}
		}
		if {$usevobfromcurdir} {
			set txt "There are no VOB files in the target folder, but there"
			set txt "$txt\nare VOB files in the source folder."
			set txt "$txt\n\nPgcEdit can use the VOB files from the source folder"
			set txt "$txt\nto compute the VTS sector pointers."
			set txt "$txt\nOtherwise, the VTS sectors will be wrong!"
			set txt "$txt\n\n(Anyway, it is recommended to load the IFOs from the"
			set txt "$txt\nfinal folder and save them back to recompute the VTS"
			set txt "$txt\nsector pointers just before burning the DVD!)"
			set txt "$txt\n\nUse the VOB files from the source folder to compute"
			set txt "$txt\nthe VTS sector pointers?  (recommended)"
			set r [tk_messageBox -message $txt -title "Save IFOs as" \
				-type yesnocancel -icon question -default yes -parent .]
			if {$r == "cancel"} {
				unset ::save_in_progress
				return -1
			} elseif {$r == "no"} {
				set usevobfromcurdir false
			}
		}
		if {$usevobfromcurdir} {
			set errcode [::writedvd::writedvd $destdir $::dvddir]
		} else {
			set errcode [::writedvd::writedvd $destdir]
		}
		if {$errcode == 0} {
			if {$usevobfromcurdir} {
				set txt "IFOs saved OK!"
			} else {
				set txt "IFOs saved OK!\n\nWARNING:\nVTS sectors calculated with the VOBs of the destination folder!"
			}
			show_warning -title $toolname -icon warning -message $txt -parent . -skip "Save DVD: VTS sectors warning"
		} elseif {$errcode == -1999} {
			show_warning -title $toolname -icon warning \
					-message "IFOs saved OK!\nVTS sectors ERROR!" -parent . -skip "Save DVD: VTS sectors error"
		} else {
			set export_pgc_comments_auto_mode false
		}
	} else {
		set toolname "Save DVD"
		#blocking_msg "Saving DVD..."
		set errcode [::writedvd::writedvd $::dvddir]
		if {$errcode == 0} {
			if {$::config(save_BUP_with_gaps)} {
				set txt "DVD saved OK!\nVTS sectors (with gaps between IFO and BUPs) OK!"
			} else {
				set txt "DVD saved OK!\nVTS sectors (standard method) OK!"
			}
			show_warning -title $toolname -icon info -message $txt -parent . -skip "Save DVD: OK"
			dvd_modified false false
		} elseif {$errcode == -1999} {
			show_warning -title $toolname -icon warning \
					-message "DVD saved OK!\nVTS sectors ERROR!" -parent .  -skip "Save DVD: VTS sectors error"
		} else {
			set export_pgc_comments_auto_mode false
		}

		# save modified menu buttons as well
		if {$errcode == 0 && ([info exists ::buttons::butsinfo(menu,edited)] ||
					[info exists ::buttons::butsinfo(pgc,edited)])} {
			set errcode [::menued::write_all_buttons]
		}
	}
	if {! $errcode && [info exists ::pgc_comments_modified] && \
			! [info exists ::do_not_save_pgc_comments]} {
		if {$::pgc_comments_modified} {
			::utils::export_pgc_comments $export_pgc_comments_auto_mode
		}
	}
	unset ::save_in_progress
	#close_blocking_msg
	return $errcode
}

#############################################################################
## Procedure:  Read_New_DVD
## read a new DVD in memory, and init the interface.
## Get input filename interactively if args is empty.
## reset Trace's breakpoints if args is NOT empty (when reloading the DVD/restoring the backup.)
proc Read_New_DVD {args} {
	set toolname "Open DVD"
	if {$::dvd_modified} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nSave it before?" \
				-icon question -title "PgcEdit" -type yesnocancel -default yes \
				-skip "Open DVD: DVD modified confirmation"]
		if {$r == "cancel"} {return 0}
		if {$r == "yes"} {
			set rc [::main::Write_DVD false]
			if {$rc != 0} {return 0}
		}
	}
	if {![info exists ::do_not_save_pgc_comments] || !$::do_not_save_pgc_comments} {
		::utils::save_pgc_comments_with_confirm yes
		set ::pgc_comments_modified false
	}
	if {$args == ""} {
		set initialdir $::config(initialdir)
		while true {
			set dvddir [tk_chooseOldDirectory -initialdir $initialdir \
				-parent . -title "PgcEdit: Open DVD folder"]
			if {$dvddir == ""} return
			if {![file exists [file join $dvddir "VIDEO_TS.IFO"]] &&
					[file isdirectory [file join $dvddir "VIDEO_TS"]]} {
				set dvddir [file join $dvddir "VIDEO_TS"]
			}
			set initialdir $dvddir
			if {[file exists [file join $dvddir "VIDEO_TS.IFO"]]} break
			bell
		}
		unset -nocomplain ::tr::breakpoints	;# reset breakpoints of Trace mode
		::tr::clear_all_state_banks
		for {set g 0} {$g < 16} {incr g} {
			set ::gprm_labels($g) ""
		}
	} else {
		if {[winfo exists .mf.cf.trace] && [winfo ismapped .mf.cf.trace]} {
			::tr::snapshot_state auto
		}
		set dvddir [file nativename [file normalize [lindex $args 0]]]
	}

	if {$::config(use_preview)} {
		::preview::close_preview
	}

	#blocking_msg "Loading DVD..."
	update idletasks
	set errcode [readdvd $dvddir ::raw_sectors ::pgcs ::menus ::ttns]
	unset -nocomplain ::warning_scan_bov
	if {$errcode == 0} {
		set ::dvddir $dvddir
		if {$::config(backup_in_root) && [file tail $::dvddir]=="VIDEO_TS"} {
			set ::backupdir [file join [file dirname $::dvddir] PgcEdit_backup]
		} else {
			set ::backupdir [file join $::dvddir PgcEdit_backup]
		}
		set ::config(initialdir) $dvddir
		# init GUI
		if {! [info exists ::do_not_save_pgc_comments]} {
			unset -nocomplain ::menus_comments
			unset -nocomplain ::pgcs_comments
		}
		$::widget(main_cmd_lab) configure -text ""
		$::widget(main_cmd_lab2) configure -text ""
		if {! $::scratchbook_active} {
			set ::scratchbook_active true
			scratchbook
		}
		set_window_title
		init_tree_menu

		display_pgc pgc 0 0

		if {$::config(auto_backup_comments)} {
			::utils::import_pgc_comments true
		}

		# close any old information window
		destroy .streamsinfo
		destroy .dump

		if {[info exists ::buttons::dvd_reopened]} {
			if {$::buttons::dvd_reopened == 1} {
				set parsebuttons true
				set parsebutsmsg ""
			} else {
				set parsebuttons false
				set parsebutsmsg "Menu VOBs not scanned for menu buttons."
				for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
					::menued::add_title_buttons_marker $vts
				}
				catch ::main::init_tree_menu_no_move
			}
			unset ::buttons::dvd_reopened
		} else {
			set parsebuttons true
			set parsebutsmsg ""
		}
		if {![info exists ::dvd_reopened] && $::config(auto_fix_streams)} {
			::streams::verify_num_streams -quiet -relax
		}
		if {$parsebuttons && $::config(auto_launch_menubuttons_parser)} {
			show_msg "DVD Opened OK.  $::pgcs(numvts) VTS loaded.  $parsebutsmsg"
			::menued::get_buttons_all_title_domains $::dvddir true "DVD Opened OK.  $::pgcs(numvts) VTS loaded."
		} else {
			show_warning -title $toolname -icon info -parent . \
				-message "DVD Opened OK.\n\n$::pgcs(numvts) VTS loaded.\n$parsebutsmsg\nTitle VOBs not scanned for BOVs."
		}

		# load the menubuttons.but file if the scanning of VOBs failed
		set parsed 0
		for {set i 0} {$i <= $::pgcs(numvts)} {incr i} {
			if {$::buttons::butsinfo(menu,$i,errorcode) == 0 && \
					$::buttons::butsinfo(menu,$i,errorcode) != 2} {
				set parsed true
				break
			}
			if {$i>1 && $parsebuttons && $::config(auto_launch_menubuttons_parser) && \
					[info exists ::buttons::butsinfo(pgc,$i,scanned)] && \
					$::buttons::butsinfo(pgc,$i,errorcode) == 0} {
				set parsed true
				break
			}
		}
		if {!$parsed} {
			set filename ""
			if {[file exists [file join $::backupdir menubuttons.but]]} {
				set filename [file join $::backupdir menubuttons.but]
				set src PgcEdit_backup
			} elseif {[file exists [file join $::dvddir menubuttons.but]]} {
				set filename [file join $::dvddir menubuttons.but]
				set src VIDEO_TS
			} elseif {[file tail $::dvddir]=="VIDEO_TS" && \
					[file exists [file join [file rootname $::dvddir] menubuttons.but]]} {
				set filename [file join [file rootname $::dvddir] menubuttons.but]
				set src "root DVD"
			}
			if {$filename != ""} {
				set msg "PgcEdit cannot parse the VOBs for menu buttons/BOVs."
				set msg "$msg\nHowever, a \"menubuttons.but\" file exists in the $src folder:"
				set msg "$msg\n\"$filename\"."
				set msg "$msg\n\nDo you want to load it?"
				set r [tk_messageBox -message $msg -title "Parse menu buttons" -type yesno \
						-default no -icon question -parent . -skip read_menubuttons.but]
				if {$r == "yes"} {
					::menubuttons::load_buts $filename
				}
			}
		}




		if {$::config(auto_create_backup)} {
			Create_Backup true
		}
	} else {
		unset -nocomplain ::dvddir
		if {$::config(backup_in_root) && [file tail $dvddir]=="VIDEO_TS"} {
			set backupdir [file join [file dirname $dvddir] PgcEdit_backup]
		} else {
			set backupdir [file join $dvddir PgcEdit_backup]
		}
		if {[file exists $backupdir]} {
			set r [tk_messageBox -title $toolname -icon question \
				-message "Error opening DVD.\nRestore backup?" -parent . \
				-type yesno -default yes]
			if {$r == "yes"} {
				CopyFile "$backupdir/VIDEO_TS.IFO" $dvddir -force $toolname .
				CopyFile "$backupdir/VIDEO_TS.IFO" "$dvddir/VIDEO_TS.BUP" -force $toolname .
				for {set vts 1} {$vts <= 1000} {incr vts} {
					if {[file exists "$backupdir/VTS_[format %02i $vts]_0.IFO"]} {
						CopyFile "$backupdir/VTS_[format %02i $vts]_0.IFO" $dvddir -force $toolname .
						CopyFile "$backupdir/VTS_[format %02i $vts]_0.IFO" "$dvddir/VTS_[format %02i $vts]_0.BUP" -force $toolname .
					} else {
						break
					}
				}
				Read_New_DVD $dvddir
			}
		}
	}
	if {$errcode == 0} {
		if {[winfo exists .mf.cf.trace]} {
			if {[info exists ::tr::snapshot_bank(auto)] && $::tr::snapshot_bank(auto) && \
					[winfo ismapped .mf.cf.trace]} {
				::tr::restore_state auto
			} else {
				::tr::reset_trace
			}
		}
		::menus::save_mru_menu
		::menus::build_mru_menu
	}
	#close_blocking_msg
	return $errcode
}

#############################################################################
## Procedure:  ReRead_DVD
## re-read the current DVD in memory, and lose changes.
proc ReRead_DVD {ask_rescan_buttons {ask_save_pgc_comments false}} {
	set ::dvd_reopened true
	if {[info exists ::dvddir]} {
		if {$::dvd_modified} {
			set r [::msgbox::MsgBox -message "The DVD has been modified.\n\nAre you sure you want to reopen it\nand lose all changes?" \
					-icon question -title "PgcEdit" -type yesno -default yes -skip "Reopen DVD w/o saving first"]
			if {$r == "no"} {return 0}
		}
		dvd_modified false false
		if {$ask_rescan_buttons} {
			set ::buttons::dvd_reopened true
		}
		unset -nocomplain ::do_not_save_pgc_comments
		if {! $ask_save_pgc_comments} {
			set ::do_not_save_pgc_comments true
		}
		set errcode [Read_New_DVD $::dvddir]
	} else {
		set errcode -9999
		bell
	}
	unset -nocomplain ::dvd_reopened
	return $errcode
}

proc move_backup {} {
	set toolname "Move backup"
	if {[info exists ::backupdir] && [info exists ::dvddir]} {
		set old_backupdir $::backupdir
		if {$::config(backup_in_root) && [file tail $::dvddir]=="VIDEO_TS"} {
			set ::backupdir [file join [file dirname $::dvddir] PgcEdit_backup]
		} else {
			set ::backupdir [file join $::dvddir PgcEdit_backup]
		}
		if {$::backupdir == $old_backupdir} return
		if {[file exists $old_backupdir]} {
			if {![file exists $::backupdir]} {
				if {![movefile $old_backupdir $::backupdir false $toolname .]} {
					show_msg "Backup moved to \"$::backupdir\"."
				}
			} else {
				set txt "New backup directory already exists:\n\"$::backupdir\"\n"
				set txt "$txt\nOld backup dir not moved:\n\"$old_backupdir\"\nPlease move it yourself now!"
				tk_messageBox -title $toolname -icon warning -message $txt -parent .
				if {$::config(backup_in_root) && [file tail $::dvddir]=="VIDEO_TS"} {
					::macros::explore_dvd
				} else {
					::macros::explore_dvd parent
				}
			}
		}
		set olddir [file dirname $old_backupdir]
		set newdir [file dirname $::backupdir]
		set backupfiles [glob -directory $olddir -nocomplain -tails -- PgcEdit_backup_*]
		foreach f $backupfiles {
			movefile [file join $olddir $f] $newdir false $toolname .
		}
	}
}

proc set_window_title {} {
	if {! [info exists ::dvddir]} return

	set txt ""

	set gn 	[::utils::get_General_Name]
	if {$gn != ""} {
		set txt "$txt - \"$gn\""
	} else {
		set txt "$txt - untitled"
	}

	set id [::utils::get_provider_ID]
	if {$id != ""} {
		set txt "${txt}; provider: $id"
	}

	if {$::dvd_modified} {
		set mod "*"
	} else {
		set mod " "
	}
	if {$::tcl_platform(platform) == "windows"} {
		set sep \\
	} else {
		set sep /
	}

	wm title . "PgcEdit - $mod \"[file nativename $::dvddir]$sep\"$txt"
}

proc dvd_modified {state onlyifnotscratchbook} {
	if {$onlyifnotscratchbook && $::scratchbook_active} return

	if {$state} {
		if {! $::dvd_modified} {
			set ::dvd_modified true
			set_window_title
		}
	} else {
		if {$::dvd_modified} {
			set ::dvd_modified false
			set_window_title
		}
	}
	catch {wm attributes . -modified $state}
}

#############################################################################
## Procedure:  dump_DVD
## Dump all the DVD commands in a text string, and show the text.
proc dump_DVD {} {
	if {[info exists ::dvddir]} {
		set txt [::disasm::dump_all_commands "DVD \"$::dvddir/\""]
		showtext ".dump" "PgcEdit: DVD Dump" $txt
	} else {
		bell
	}
}

#############################################################################
## Display contents of PGC Selector pane as text
proc showlabels_as_text {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set txt "DVD: \"$::dvddir\"\n"
	set txt "${txt}DVD-TEXT General Name: \"[::utils::get_General_Name]\"\n"
	set txt "${txt}Provider ID: \"[::utils::get_provider_ID]\"\n"
	set txt "${txt}Number of VTS: $::pgcs(numvts)\n"

	set str50 [string repeat X 50]
	set txt $txt\n[pgc_label $::config(fixed_width_font) pgc 0 0]
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set lab [pgc_label $::config(fixed_width_font) menu 0 $lu $pgc]
			if {$::menus(0,$lu,$pgc,numcells) > 0} {
				set lab "[padtextat $lab $str50 $::config(fixed_width_font)]  ([string trim [pgc_sublabel menu 0 $lu $pgc] "Chapters: n/a,  "])"
			}
			set txt $txt\n$lab
		}
	}
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set txt $txt\n
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set lab [pgc_label $::config(fixed_width_font) menu $vts $lu $pgc]
				if {$::menus($vts,$lu,$pgc,numcells) > 0} {
					set lab "[padtextat $lab $str50 $::config(fixed_width_font)]  ([string trim [pgc_sublabel menu $vts $lu $pgc] "Chapters: n/a,  "])"
				}
				set txt $txt\n$lab
			}
		}
		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			set lab [pgc_label $::config(fixed_width_font) pgc $vts $pgc]
			if {$::pgcs($vts,$pgc,numcells) > 0} {
				set lab "[padtextat $lab $str50 $::config(fixed_width_font)]  ([string trim [pgc_sublabel pgc $vts $pgc]])"
			}
			set txt $txt\n$lab
		}
	}
	set txt $txt\n
	showtext .pgclist "PGC list" $txt
}


#############################################################################
## Procedure:  init_tree_menu
## Initialize the DVD menubutton with the DVD hierarchy stored in the ::pgcs
## and ::menus global arrays.
proc init_tree_menu {} {

	set old [.mf.f1.mb cget -menu]
	if {$old != ""} {
		destroy $old
	}
	set m .mf.f1.mb.menu
	menu $m
	.mf.f1.mb configure -menu $m

	set ::main::labelslist {}
	set ::main::cmdslist {}


	# VIDEO_TS.IFO

	# FP-PGC
	set lab [pgc_label $::config(global_font) pgc 0 0]
	set labl [split [pgc_label "" pgc 0 0] ,]
	lappend ::main::labelslist $lab
	lappend ::main::cmdslist "::main::display_pgc pgc 0 0"
	$m add cascade -label "VIDEO_TS.IFO: [lindex $labl 0]" -menu $m.vts0
	menu $m.vts0
	$m.vts0 add radiobutton -variable ::main::pgcmenu_selection -value "p,0,0" \
			-label [lindex $labl 1] -command "::main::display_pgc pgc 0 0"

	# VMG Menus
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		set lulabel [lindex [split [pgc_label $::config(global_font) menu 0 $lu 1] ,] 1]
		$m.vts0 add cascade -label $lulabel -menu $m.vts0.lu$lu
		menu $m.vts0.lu$lu
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set lab [pgc_label $::config(global_font) menu 0 $lu $pgc]
			set labl [split $lab ,]
			lappend ::main::labelslist $lab
			lappend ::main::cmdslist "::main::display_pgc_pv menu 0 $lu $pgc"
			if {$pgc > 1 && [expr {$pgc % 20}] == 1} {set break 1} else {set break 0}
			$m.vts0.lu$lu add radiobutton -variable ::main::pgcmenu_selection \
					-value "m,0,$lu,$pgc" -label [lindex $labl 2] -columnbreak $break \
					-command "::main::display_pgc_pv menu 0 $lu $pgc"
		}
	}

	# VTS
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		lappend ::main::labelslist ""
		lappend ::main::cmdslist "bell"

		set vtslabel "VTS_[format %02i $vts]_0.IFO: VTS $vts"
		if {[expr {$vts % 20}] == 0} {set break 1} else {set break 0}
		$m add cascade -label $vtslabel -menu $m.vts$vts -columnbreak $break
		menu $m.vts$vts

		# VTS menus
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			set lulabel [lindex [split [pgc_label $::config(global_font) menu $vts $lu 1] ,] 1]
			$m.vts$vts add cascade -label $lulabel -menu $m.vts$vts.lu$lu
			menu $m.vts$vts.lu$lu
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set lab [pgc_label $::config(global_font) menu $vts $lu $pgc]
				set labl [split $lab ,]
				lappend ::main::labelslist $lab
				lappend ::main::cmdslist "::main::display_pgc_pv menu $vts $lu $pgc"
				if {$pgc > 1 && [expr {$pgc % 20}] == 1} {set break 1} else {set break 0}
				$m.vts$vts.lu$lu add radiobutton -variable ::main::pgcmenu_selection \
					-value "m,$vts,$lu,$pgc" -label [lindex $labl 2] -columnbreak $break \
					-command "::main::display_pgc_pv menu $vts $lu $pgc"
			}
		}

		# VTS Titles
		if {$::pgcs($vts,numpgcs) > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set lab [pgc_label $::config(global_font) pgc $vts $pgc]
				set labl [split $lab ,]
				lappend ::main::labelslist $lab
				lappend ::main::cmdslist "::main::display_pgc_pv pgc $vts $pgc"
				if {$pgc > 1 && [expr {$pgc % 20}] == 1} {set break 1} else {set break 0}
				$m.vts$vts add radiobutton -variable ::main::pgcmenu_selection \
					-value "p,$vts,$pgc" -label [lindex $labl 1] -columnbreak $break \
					-command "::main::display_pgc_pv pgc $vts $pgc"
			}
		}
	}
	lappend ::main::labelslist ""
	lappend ::main::cmdslist "bell"
	highlight_trace_breakpoints
}

proc init_tree_menu_no_move {} {
	if {! [info exists ::dvddir]} {
		return
	}
	if {[winfo exists .mf.f2.selector.a.lb]} {
		set selectoroldpos [lindex [.mf.f2.selector.a.lb yview] 0]
	}
	set listboxoldpos [get_selection_indexes]
	::main::init_tree_menu
	set indexlist [split [lindex $::editor_cmds_from 1] ","]

	set err false
	if {$::pgcs(numvts) >= [lindex $indexlist 0]} {
		if {[lindex $::editor_cmds_from 0] == "pgc"} {
			if {$::pgcs([lindex $indexlist 0],numpgcs) >= [lindex $indexlist 1]} {
				::main::display_pgc pgc [lindex $indexlist 0] [lindex $indexlist 1]
			} else {
				set err true
			}
		} else {
			if {$::menus([lindex $indexlist 0],numlus) >= [lindex $indexlist 1] &&
			$::menus([lindex $indexlist 0],[lindex $indexlist 1],numpgcs) >= [lindex $indexlist 2]} {
				::main::display_pgc menu [lindex $indexlist 0] [lindex $indexlist 1] [lindex $indexlist 2]
			} else {
				set err true
			}
		}
	} else {
		set err true
	}

	if {$err} {
		::main::display_pgc pgc 0 0
	} else {
		if {[winfo exists .mf.f2.selector.a.lb]} {
			.mf.f2.selector.a.lb yview moveto $selectoroldpos
		}
		set_listbox_selection [lindex $listboxoldpos 0] [lindex $listboxoldpos 1]
	}
}

#############################################################################
## Procedure:  highlight Trace breakpoints in PGC selector
proc highlight_trace_breakpoints {} {
	if {[winfo exists .mf.f2.selector.a.lb] && [info exists ::dvddir]} {
		set redlist {}
		set bluelist {}
		set purplelist {}
		if {[info exists ::tr::breakpoints]} {
			if {[info exists ::tr::breakpoints(pgc,0,0,prepost)] && \
					[info exists ::tr::breakpoints(pgc,0,0,playback)]} {
				lappend purplelist 0
			} elseif {[info exists ::tr::breakpoints(pgc,0,0,prepost)]} {
				lappend redlist 0
			} elseif {[info exists ::tr::breakpoints(pgc,0,0,playback)]} {
				lappend bluelist 0
			}
			for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
					for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
						set lab [pgc_label $::config(global_font) menu $vts $lu $pgc]
						if {[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,prepost)] && \
								[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,playback)]} {
							lappend purplelist [lsearch $::main::labelslist $lab]
						} elseif {[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,prepost)]} {
							lappend redlist [lsearch $::main::labelslist $lab]
						} elseif {[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,playback)]} {
							lappend bluelist [lsearch $::main::labelslist $lab]
						}
					}
				}
			}
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
					set lab [pgc_label $::config(global_font) pgc $vts $pgc]
					if {[info exists ::tr::breakpoints(pgc,$vts,$pgc,prepost)] && \
							[info exists ::tr::breakpoints(pgc,$vts,$pgc,playback)]} {
						lappend purplelist [lsearch $::main::labelslist $lab]
					} elseif {[info exists ::tr::breakpoints(pgc,$vts,$pgc,prepost)]} {
						lappend redlist [lsearch $::main::labelslist $lab]
					} elseif {[info exists ::tr::breakpoints(pgc,$vts,$pgc,playback)]} {
						lappend bluelist [lsearch $::main::labelslist $lab]
					}
				}
			}
		}
		# set colors
		for {set i 0} {$i < [llength $::main::labelslist]} {incr i} {
			set curcolor [.mf.f2.selector.a.lb itemcget $i -foreground]
			if {$curcolor != "" && $curcolor != "black"} {
				.mf.f2.selector.a.lb itemconfigure $i -foreground black -selectforeground black
			}
		}
		foreach i $redlist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground red -selectforeground red
		}
		foreach i $bluelist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground blue -selectforeground blue
		}
		foreach i $purplelist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground purple -selectforeground purple
		}
	}
}

#############################################################################
## Procedure:  selector
## mode arg: "show", "hide" or "toggle"
proc selector {mode labellistvar cmdslistvar} {
	set w .mf.f2.selector

	if {! [info exists $labellistvar]} {
		set $labellistvar {}
		set $cmdslistvar {}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if {[winfo exists $w.a]} {
				set show 0
			} else {
				set show 1
			}
		}
	}
	if {$show} {
		pack  configure .mf.f2.selector -side $::config(selector_side)

		set ::main::selector_opened 1

		frame $w.a
		listbox   $w.a.lb -exportselection 0 -selectmode browse \
				-listvariable $labellistvar -width 0 -height $::config(selector_height) \
				-activestyle dotbox -selectbackground gray80 \
				-selectforeground black -yscrollcommand "$w.a.ys set"
		scrollbar $w.a.ys -command "$w.a.lb yview"
		grid  $w.a.lb -column 0 -row 0 -sticky nesw
		grid  $w.a.ys -column 1 -row 0 -sticky ns
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		bind $w.a.lb <1> "focus $w.a.lb; ::main::selector_launch_cmd_click $w $cmdslistvar"
		bind $w.a.lb <Double-1> "focus $w.a.lb; if \[::main::selector_launch_cmd_click $w $cmdslistvar\] ::main::action_Edit_PGC"
		bind $w.a.lb <Return> "::main::action_Edit_PGC"
		bind $w.a.lb <space>  "::main::selector_launch_cmd_key $w $cmdslistvar"
		bind $w.a.lb <Up>     "::main::selector_launch_cmd_up $w $cmdslistvar"
		bind $w.a.lb <Down>   "::main::selector_launch_cmd_down $w $cmdslistvar"
		bind $w.a.lb <Home>   "::main::selector_set_selection $w 0; ::main::selector_launch_cmd_index $w $cmdslistvar 0"
		bind $w.a.lb <End>    "::main::selector_set_selection $w end; ::main::selector_launch_cmd_index $w $cmdslistvar end"
		bind $w.a.lb <MouseWheel> "if {%D>0} {$w.a.lb yview scroll 1 units} else {$w.a.lb yview scroll -1 units}"
		bind $w.a.lb <Right>   "::main::activate_listbox"
		bind $w.a.lb <Left>    "::main::activate_listbox"
		bind .mf.f1.mb <Right> "::main::activate_listbox"
		bind .mf.f1.mb <Left>  "::main::activate_listbox"

		::menus::build_popup_menus

		bind $w.a.lb <2> "::main::selector_middle_click $w $cmdslistvar"
		bind $w.a.lb <Control-1> "::main::selector_middle_click $w $cmdslistvar"
		bind $w.a.lb <Shift-1> "::main::selector_shift_click $w $cmdslistvar"
		bind $w.a.lb <Control-Double-1> {}
		bind $w.a.lb $::config(popups_button) "::main::selector_context_menu $w $cmdslistvar"
		pack  $w.a -side top -fill both -expand 1

		init_tree_menu_no_move
		highlight_trace_breakpoints
		focus $w.a.lb
	} else {
		set ::main::selector_opened 0

		#bind . <Destroy> {}
		destroy $w.a
		#bind . <Destroy> {cleanexit 0}
		$w configure -width 1	;# don't work with 0

		activate_listbox
	}
}
proc selector_middle_click {w cmdslist} {
	focus $w.a.lb
	if {[selector_launch_cmd_click $w $cmdslist]} {
		set idx [$w.a.lb index @[expr {[winfo pointerx .]-[winfo rootx $w.a.lb]}],[expr {[winfo pointery .]-[winfo rooty $w.a.lb]}]]
		set domain [lindex [$w.a.lb get $idx] 0]
		#if {$domain == "VMG" } return
		#if {$domain == "VTST"} return
		::menubuttons::display_current_menu_buttons info
	}
}
proc selector_shift_click {w cmdslist} {
	focus $w.a.lb

	set idx [$w.a.lb index @[expr {[winfo pointerx .]-[winfo rootx $w.a.lb]}],[expr {[winfo pointery .]-[winfo rooty $w.a.lb]}]]
	set cmd [lindex [subst $$cmdslist] [$w.a.lb index $idx]]
	if {$cmd == "bell"} {bell; return}
	catch {
		set cur_type [lindex $cmd 1]
		if {$cur_type == "menu"} {
			set index "[lindex $cmd 2],[lindex $cmd 3],[lindex $cmd 4]"
		} else {
			set index "[lindex $cmd 2],[lindex $cmd 3]"
		}
		if {[info exists ::tr::breakpoints($cur_type,$index,prepost)] && \
				[info exists ::tr::breakpoints($cur_type,$index,playback)]} {
			unset ::tr::breakpoints($cur_type,$index,prepost)
			unset ::tr::breakpoints($cur_type,$index,playback)
		} elseif {[info exists ::tr::breakpoints($cur_type,$index,prepost)]} {
			unset ::tr::breakpoints($cur_type,$index,prepost)
			set ::tr::breakpoints($cur_type,$index,playback) 1
		} elseif {[info exists ::tr::breakpoints($cur_type,$index,playback)]} {
			set ::tr::breakpoints($cur_type,$index,prepost) 1
			set ::tr::breakpoints($cur_type,$index,playback) 1
		} else {
			set ::tr::breakpoints($cur_type,$index,prepost) 1
		}
		highlight_trace_breakpoints
	}
}
proc selector_context_menu {w cmdslist} {
	focus $w.a.lb
	selector_launch_cmd_click $w $cmdslist
	set idx [$w.a.lb index @[expr {[winfo pointerx .]-[winfo rootx $w.a.lb]}],[expr {[winfo pointery .]-[winfo rooty $w.a.lb]}]]
	set label [$w.a.lb get $idx]
	switch "[lindex $label 0]" {
		"VMG" {
			tk_popup .selector_popup_menu_fp     [winfo pointerx .] [winfo pointery .] \
					$::main::last_fp_menu_selected
		}
		"VMGM" -
		"VTSM" {
			tk_popup .selector_popup_menu_menus  [winfo pointerx .] [winfo pointery .] \
					$::main::last_menus_menu_selected
		}
		"VTST" {
			tk_popup .selector_popup_menu_titles [winfo pointerx .] [winfo pointery .] \
					$::main::last_titles_menu_selected
		}
		default {
			tk_popup .selector_popup_menu_notpgc [winfo pointerx .] [winfo pointery .] \
					$::main::last_notpgc_menu_selected
		}
	}
}
proc selector_set_selection {w idx} {
	$w.a.lb activate $idx
	$w.a.lb selection clear 0 end
	$w.a.lb selection set $idx
	$w.a.lb see $idx
}
proc selector_launch_cmd_index {w cmdslist index} {
	selector_set_selection $w $index
	set list [subst $$cmdslist]
	if {$index == "end"} {set index [expr {[llength $list]-1}]}
	eval [lindex $list $index]
}
proc selector_launch_cmd_click {w cmdslist} {
	set idx [$w.a.lb index @[expr {[winfo pointerx .]-[winfo rootx $w.a.lb]}],[expr {[winfo pointery .]-[winfo rooty $w.a.lb]}]]
	if {$idx == [$w.a.lb curselection] && $::scratchbook_active} {
		set oldselection $::scratchbook_old_selection
		if {[lindex $oldselection 0] == "-"} {set oldselection {0 0}}
		set fromscratchbook true
	} else {
		set fromscratchbook false
	}
	if {$idx != [$w.a.lb curselection] || $::scratchbook_active} {
		selector_set_selection $w $idx
		set cmd [lindex [subst $$cmdslist] [$w.a.lb index $idx]]
		if {$cmd == "bell"} {return false}
		eval $cmd
		if {$fromscratchbook} {
			set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
		}
	}
	return true
}
proc selector_launch_cmd_key {w cmdslist} {
	eval [lindex [subst $$cmdslist] [$w.a.lb index active]]
}
proc selector_launch_cmd_up {w cmdslist} {
	set idx [expr {[$w.a.lb index active] -1}]
	if {$idx >= 0} {
		if {[lindex [subst $$cmdslist] $idx] != "bell"} {
			eval [lindex [subst $$cmdslist] $idx]
		}
		incr idx
		selector_set_selection $w $idx
	}
}
proc selector_launch_cmd_down {w cmdslist} {
	set idx [expr {[$w.a.lb index active] +1}]
	if {$idx < [llength [subst $$cmdslist]]} {
		if {[lindex [subst $$cmdslist] $idx] != "bell"} {
			eval [lindex [subst $$cmdslist] $idx]
		}
		incr idx -1
		selector_set_selection $w $idx
	}
}
proc activate_listbox {} {
	focus $::widget(main_listbox)
	set sel [$::widget(main_listbox) curselection]
	if {[llength $sel]<=1 && [lindex $sel 0]<=0} {
		if {[llength $::editor_cmds_list]>1} {
			set_listbox_selection 1
		} else {
			set_listbox_selection 0
		}
	}
	set sel [$::widget(main_listbox) curselection]
	set_listbox_selection [lindex $sel 0] [lindex $sel end]
	if {[winfo exists .mf.f2.selector.a.lb]} {
		set idx [.mf.f2.selector.a.lb index active]
		if {$idx > -1} {
			.mf.f2.selector.a.lb see $idx
		}
	}
}

#############################################################################
## Procedure:  refresh_listbox
## refresh the listbox with the current listbox contents: re-disassemble all
## commands, regenerate line numbers...
proc init_refresh_listbox {} {
	if {$::config(highlight_gotos)} {
		proc refresh_listbox {} {
			set ::editor_cmds_list_no_hex $::editor_cmds_list
			set lbitem 0
			set cmdnum 0
			set gotolines(1) ""
			set gotolines(2) ""
			set gotolines(3) ""
			set sectionnum 0
			set sectionstart(1) 0
			set sectionstart(2) 1
			set sectionstart(3) 2
			set sectionstart(4) 3
			set errorlines ""
			set jumplines ""
			foreach cmd $::editor_cmds_list {
				if {[string range $cmd 0 0] == {*}} {
					set cmdnum 0
					incr sectionnum
					set sectionstart($sectionnum) $lbitem
					set sectionlbnum $lbitem
					set newcmd [formatCommand "* [lindex $cmd 1]" 0]
					$::widget(main_listbox) itemconfigure $lbitem -background $::config(systembuttonface)
					$::widget(main_listbox) itemconfigure $lbitem -selectbackground gray65
					lset ::editor_cmds_list_no_hex $lbitem $newcmd
				} else {
					incr cmdnum
					if {[string range $cmd 0 0] == {[}} {
						set cmd [string range $cmd 1 23]
					}
					set newcmd [formatCommand $cmd $cmdnum]
					if {[string first "ILLEGAL" $newcmd] >= 0} {
						lappend errorlines $lbitem
					} elseif {!$::scratchbook_active} {
						set target [highlight_jump_cmd [lindex $::editor_cmds_from 0] [lindex $::editor_cmds_from 1] \[$cmd\] -1 false false]
						if {$target!="" && ![lindex $target 4]} {
							lappend jumplines $lbitem
						}
					}
					if {[string first "Goto" $newcmd] >= 0} {
						lappend gotolines($sectionnum) [expr {$sectionlbnum + [x2d [string range $newcmd 22 23]]}]
						lappend gotolines($sectionnum) $lbitem
					}
					$::widget(main_listbox) itemconfigure $lbitem -background $::config(system_window_color)
					$::widget(main_listbox) itemconfigure $lbitem -selectbackground $::config(systemhighlight)
					lset ::editor_cmds_list_no_hex $lbitem [string range $newcmd 25 end]
				}
				lset ::editor_cmds_list $lbitem $newcmd
				incr lbitem
			}
			set sectionstart([expr {$sectionnum+1}]) $lbitem
			for {set s 1} {$s <= $sectionnum} {incr s} {
				for {set g 0} {$g < [llength $gotolines($s)]} {incr g 2} {
					set l  [lindex $gotolines($s) $g]
					set l2 [lindex $gotolines($s) [expr {$g+1}]]
					if {$l > $sectionstart($s) && $l < $sectionstart([expr {$s+1}]) && $l != $l2} {
						catch {
							$::widget(main_listbox) itemconfigure $l -background #FFFFAF -selectbackground #BFBF84
						}
					} else {
						$::widget(main_listbox) itemconfigure $l2 -background orange -selectbackground #AA7000
					}
				}
			}
			foreach l $jumplines {
				$::widget(main_listbox) itemconfigure $l -background orange -selectbackground #AA7000
			}
			foreach l $errorlines {
				$::widget(main_listbox) itemconfigure $l -background pink -selectbackground #C09090
			}
		}

	} else {

		if {[info exists ::widget(main_listbox)] && \
				[winfo exists $::widget(main_listbox)]} {
			for {set i 0} {$i < [$::widget(main_listbox) index end]} {incr i} {
				$::widget(main_listbox) itemconfigure $i -selectbackground $::config(systemhighlight)
			}
		}

		proc refresh_listbox {} {
			set ::editor_cmds_list_no_hex $::editor_cmds_list
			set cmdnum 0
			set linenum 1
			foreach cmd $::editor_cmds_list {
				switch [string range $cmd 0 0] {
					{*} {
						set linenum 1
						set newcmd [formatCommand "* [lindex $cmd 1]" 0]
						$::widget(main_listbox) itemconfigure $cmdnum -background $::config(systembuttonface)
						lset ::editor_cmds_list_no_hex $cmdnum $newcmd
					}
					{[} {
						set newcmd [formatCommand [string range $cmd 1 23] $linenum]
						incr linenum
						$::widget(main_listbox) itemconfigure $cmdnum -background $::config(system_window_color)
						lset ::editor_cmds_list_no_hex $cmdnum [string range $newcmd 25 end]
					}
					default {
						set newcmd [formatCommand $cmd $linenum]
						incr linenum
						$::widget(main_listbox) itemconfigure $cmdnum -background $::config(system_window_color)
						lset ::editor_cmds_list_no_hex $cmdnum [string range $newcmd 25 end]
					}
				}
				lset ::editor_cmds_list $cmdnum $newcmd
				incr cmdnum
			}
		}
	}
}
init_refresh_listbox

proc highlight_illegal_cmds {} {
 	if {$::scratchbook_active || !$::config(highlight_gotos)} {
 		return
 	}
	set type [lindex $::editor_cmds_from 0]
	set vts  [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	set cmds [subst \$::${type}s([lindex $::editor_cmds_from 1])]
	set area all
	for {set i 0} {$i<[llength $cmds]} {incr i} {
		set hex [lindex $cmds $i]
		if {[lindex $hex 0]=="*"} {
			set area [lindex $hex 1]
		} else {
			set cmdname [::disasm::cmdname $hex]
			if {$cmdname == "???"} {
				set valid 0
			} else {
				set valid [::ced::valid_command $cmdname $type $vts $area]
			}
			if {!$valid} {
				$::widget(main_listbox) itemconfigure $i -background pink
				$::widget(main_listbox) itemconfigure $i -selectbackground #C09090
			} else {
				set cmdname [::disasm::cmdname2 $hex]
				if {$cmdname!=""} {
					set valid [::ced::valid_command $cmdname $type $vts $area]
					if {!$valid} {
						$::widget(main_listbox) itemconfigure $i -background pink
						$::widget(main_listbox) itemconfigure $i -selectbackground #C09090
					}
				}
			}
		}
	}
	return
}

# Return the PGC label as it is displayed in the PGC Selector.
# - font: used to padd the text so that all fields are aligned.
#   Use "" to disable this padding, and return the label unformatted.
# - args: must be the PGC number for a title PGC, and the LU and PGC
#   numbers for a menu PGC.
# - -nocomment: After the LU and PGC arguments, a "-nocomment" keyword
#   can be added to remove the user's comment at the end of the label.
proc pgc_label {font type vts args} {
	if {$vts == 0} {
		set label "VMG"
		if {$type == "menu"} {
			set label "${label}M"
		}
	} else {
		set label "VTS"
		if {$type == "menu"} {
			set label "${label}M"
		} else {
			set label "${label}T"
		}
		set label "$label $vts"
	}
	if {$type == "pgc"} {
		set pgc [lindex $args 0]
		if {$vts == 0} {
			set label "$label , First-Play PGC "
			if {[info exists ::pgcs_comments(0,0)]} {
				set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy)  99b. ChapterM" $font]
				set label "$label '$::pgcs_comments(0,0)'"
			}
			return $label
		} else {
			set label "$label , $pgc"
			set label [padtextat $label "VTST 9 , 9999" $font]
			if {[info exists ::pgcs($vts,$pgc,TTN)] && $::pgcs($vts,$pgc,TTN) != 0} {
				set label "$label  TTN $::pgcs($vts,$pgc,TTN)"
			}
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999" $font]

			if {[info exists ::pgcs($vts,$pgc,numprograms)] \
					&& $::pgcs($vts,$pgc,numprograms) > 0} {
				set label "$label ([short_pbtime [string range $::pgcs($vts,$pgc,header) 4 7]])"
				if {[info exists ::pgcs($vts,$pgc,numbuttons)] && \
						$::pgcs($vts,$pgc,numbuttons)>0} {
					set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy) " $font]
					set label "$label $::pgcs($vts,$pgc,numbuttons)b."
				}
			} else {
				set label "$label (dummy)"
			}
			#set label [padtextat $label "VTST 9 , 999 (TTN 99) (9:99:99)" $font]
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy)  99b." $font]
			if {[info exists ::pgcs($vts,$pgc,Title)]} {
				set label "$label Title $::pgcs($vts,$pgc,Title)"
			}
			if {[info exists ::pgcs_comments($vts,$pgc)] && [lsearch $args "-nocomment"]==-1} {
				set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy)  99b. ChapterM" $font]
				set label "$label '$::pgcs_comments($vts,$pgc)'"
			}
		}
		#if [info exists ::tr::usedpgcs(pgc,$vts,$pgc)] {
		#	set label "$label *"
		#}
	} else {
		set lu  [lindex $args 0]
		set label "$label , LU $lu ($::menus($vts,$lu,lang))"
		set pgc [lindex $args 1]
		set label "$label , $pgc"
		set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999" $font]
		if {[info exists ::menus($vts,$lu,$pgc,numprograms)] \
				&& $::menus($vts,$lu,$pgc,numprograms) > 0} {
			set label "$label ([short_pbtime [string range $::menus($vts,$lu,$pgc,header) 4 7]])"
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy) " $font]
			if {[info exists ::menus($vts,$lu,$pgc,numbuttons)] && \
					$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
				if {$::menus($vts,$lu,$pgc,numbuttons) > 0} {
					set label "$label $::menus($vts,$lu,$pgc,numbuttons)b."
				}
			} else {
				set label "$label ?b."
			}
		} else {
			set label "$label (dummy)"
		}
		set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy)  99b." $font]
		if {[info exists ::menus($vts,$lu,$pgc,masktext)]} {
			set label "$label $::menus($vts,$lu,$pgc,masktext)"
		}
		if {[info exists ::menus_comments($vts,$lu,$pgc)] && [lsearch $args "-nocomment"]==-1} {
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (dummy)  99b. ChapterM" $font]
			set label "$label '$::menus_comments($vts,$lu,$pgc)'"
		}
	}
	return [string trimright $label " "]
}
proc padtextat {txt template font} {
	if {$font != ""} {
		set minpos [font measure $font $template]
		incr minpos -5
		while {[font measure $font $txt] < $minpos} {
			set txt "$txt "
		}
	} else {
		set txt "$txt "
	}
	return $txt
}
proc short_pbtime {bcd} {
	set t1 [format %i [expr { [get_int bcd 0 1] / 16}]][format  %i [expr { [get_int bcd 0 1] & 15}]]:
	set t2 [format %i [expr { [get_int bcd 1 1] / 16}]][format  %i [expr { [get_int bcd 1 1] & 15}]]:
	set t3 [format %i [expr { [get_int bcd 2 1] / 16}]][format  %i [expr { [get_int bcd 2 1] & 15}]]

	if {$t1 == "00:"} {
		set t1 ""
		if {[string range $t2 0 0] == "0"} {
			set t2 [string range $t2 1 end]
		}
	} else {
		set t1 [string trimleft $t1 "0"]
	}

	return $t1$t2$t3
}
proc pgc_sublabel {type vts args} {
	if {$type == "pgc"} {
		set pgcnum [lindex $args 0]
		if {$vts == 0} {
			set numchapters "n/a"
			set numprograms "n/a"
			set numcells    "n/a"
		} else {
			if {[info exists ::pgcs($vts,$pgcnum,pttlist)]} {
				set numchapters [llength $::pgcs($vts,$pgcnum,pttlist)]
			} else {
				set numchapters 0
			}
			set numprograms $::pgcs($vts,$pgcnum,numprograms)
			set numcells    $::pgcs($vts,$pgcnum,numcells)
		}
	} else {
		set lunum  [lindex $args 0]
		set pgcnum [lindex $args 1]
		set numchapters "n/a"
		set numprograms $::menus($vts,$lunum,$pgcnum,numprograms)
		set numcells    $::menus($vts,$lunum,$pgcnum,numcells)
	}
	return " Chapters: $numchapters,  Programs: $numprograms,  Cells: $numcells "
}

#############################################################################
## Procedure:  display_pgc
## Display a PGC in the listbox, and update the PGC label beside the tree menubutton
## usage:
##  display_pgc "pgc" vtsnum pgcnum
##  display_pgc "menu" vtsnum lunum pgcnum
## example:
##  ::main::display_pgc_pv menu 9 1 1
proc display_pgc {type vts args} {
	if {$type == "pgc"} {
		set pgcnum [lindex $args 0]
		set label [pgc_label "" $type $vts $pgcnum]
		set label_selector [pgc_label $::config(global_font) $type $vts $pgcnum]
		cmds_to_listbox pgc $vts,$pgcnum $label
		set ::main::pgcmenu_selection "p,$vts,$pgcnum"
		$::widget(main_cmd_lab2) configure -text [pgc_sublabel $type $vts $pgcnum]
	} else {
		set lunum  [lindex $args 0]
		set pgcnum [lindex $args 1]
		set label [string trim [pgc_label "" $type $vts $lunum $pgcnum] "+"]
		set label_selector [string trim [pgc_label $::config(global_font) $type $vts $lunum $pgcnum] "+"]
		cmds_to_listbox menu $vts,$lunum,$pgcnum $label
		set ::main::pgcmenu_selection "m,$vts,$lunum,$pgcnum"
		$::widget(main_cmd_lab2) configure -text [pgc_sublabel $type $vts $lunum $pgcnum]
	}
	# needed for the Trace operation
	namespace eval ::tr {}
	if {$vts > 0} {
		set ::tr::last_visited_vts $vts
	} elseif {! [info exists ::tr::last_visited_vts]} {
		set ::tr::last_visited_vts 1
	}

	# refresh the PGC Selector
	set w .mf.f2.selector.a.lb
	if {[winfo exists $w]} {
		set idx [lsearch [$w get 0 end] $label_selector]
		$w selection clear 0 end
		if {$idx != -1} {
			$w selection set $idx
			$w activate $idx
			$w see $idx
		}
	}
}

# same as Display PGC, but launch the preview if needed
proc display_pgc_pv {type vts args} {
	set pv false
	if {$type == "pgc"} {
		display_pgc $type $vts [lindex $args 0]
		if {[info exists ::pgcs($vts,[lindex $args 0],numcells)] && \
				$::pgcs($vts,[lindex $args 0],numcells) > 0} {
			set pv $::config(autopreview)
		}
	} else {
		display_pgc $type $vts [lindex $args 0] [lindex $args 1]
		if {[info exists ::menus($vts,[lindex $args 0],[lindex $args 1],numcells)] && \
				$::menus($vts,[lindex $args 0],[lindex $args 1],numcells) > 0} {
			set pv $::config(autopreview)
		}
	}
	if {$pv && $::config(use_preview)} {
		::preview::preview_current_pgc_2
		#after 500 {if {[focus]!="."} {focus -force .}}
	}
}

#############################################################################
## Procedure:  cmds_to_listbox
## init the global variable ::editor_cmds_list with the current commands
proc cmds_to_listbox {type index label} {
	$::widget(main_cmd_lab) configure -text $label

	if {$::scratchbook_active} {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::scratchbook_active false
	}

	if {[info exists ::editor_cmds_from]} {
		set oldtype $::editor_cmds_from
	} else {
		set oldtype ""
	}

	# remember from where the commands are coming from
	set ::editor_cmds_from [list $type $index]

	# init editor list
	switch $type {
		pgc {
			set ::editor_cmds_list $::pgcs($index)
			if {$oldtype != $::editor_cmds_from} {
				set ::editor_undo_buffer $::pgcs($index)
			}
		}
		menu {
			set ::editor_cmds_list $::menus($index)
			if {$oldtype != $::editor_cmds_from} {
				set ::editor_undo_buffer $::menus($index)
			}
		}
	}
	refresh_listbox
	highlight_illegal_cmds
	set_listbox_selection 0
}

#############################################################################
## Procedure:  listbox_to_cmds
## copies the listbox contents from $::editor_cmds_list to the original part
## of the DVD $::pgcs($index) or $::menus($index)
proc listbox_to_cmds {{highlight_illegal_cmds true}} {
	set rawcmds {}
	foreach line $::editor_cmds_list {
		if {[string range $line 0 0] == "*"} {
			lappend rawcmds "* [lindex $line 1]"
		} else {
			lappend rawcmds "[string range $line 1 23]"
		}
	}
	if {$::scratchbook_active} {
		set ::scratchbook_buffer $rawcmds
	} else {
		set type  [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]
		set [subst ::${type}s($index)] $rawcmds
	}
	if {$highlight_illegal_cmds} {
		highlight_illegal_cmds
	}
}

#############################################################################
## Procedure:  scratchbook
## Configure the listbox for the scratchbook or the normal DVD commands upon
## state of the bool $::scratchbook_active
proc scratchbook {} {
	set oldselection $::scratchbook_old_selection
	if {[lindex $oldselection 0] == "-"} {set oldselection {0 0}}
	if {$::scratchbook_active} {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::editor_cmds_list $::scratchbook_buffer
		set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
	} else {
		if {[info exists ::editor_cmds_from]} {
			set ::scratchbook_old_selection [get_selection_indexes]
			set type  [lindex $::editor_cmds_from 0]
			set index [lindex $::editor_cmds_from 1]
			set ::editor_cmds_list [subst \$::${type}s($index)]
			set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
		} else {
			set ::scratchbook_active true
			bell
		}
	}
	refresh_listbox
}
proc save_scratchbook {} {
	if {[llength $::scratchbook_buffer] < 2} {
		bell
		return
	}
	set filename [tk_getSaveFile -filetypes {{"scratchbook files" {*.sbk}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Save scratchbook" -parent .]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".SBK"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	saveFile $filename binary $::scratchbook_buffer
}
proc open_scratchbook {} {
	set filename [tk_getOpenFile \
		-filetypes {{"scratchbook files" {*.sbk}} {"text files" {*.txt}} {"all files" {*.*}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Open scratchbook" -parent .]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".SBK"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	set buf [readFile $filename binary]
	if {[catch {set line0 [lindex $buf 0]}]} {
		bell
		return
	}
	if {[lindex $line0 1] != "scratchbook"} {
		bell
		return
	}
	set ::scratchbook_buffer $buf
	set ::scratchbook_active true
	scratchbook
}

#############################################################################
## Procedures:  Import/Export current/all PGC commands

## TODO: Check all commands for the correct format
proc import_one_pgc {filename varnamevar} {
	set buf [readFile $filename binary]
	set pre  [lsearch $buf "* pre"]
	set post [lsearch $buf "* post"]
	if {$pre != 0} {bell; return 0}
	if {$post < $pre} {bell; return 0}
	if {[lsearch $buf "* cell"] < $post} {bell; return 0}

	set $varnamevar $buf
	return true
}
proc import_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	set filename [tk_getOpenFile \
		-filetypes {{"PGC commands files" {*.pgc}} {"text files" {*.txt}} {"all files" {*.*}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile [get_default_filename $::editor_cmds_from] -title "PgcEdit: Import PGC commands"]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".PGC"} {
		set filename "${filename}.pgc"
	}
	set ::config(import_initialdir) [file dirname $filename]

	import_one_pgc $filename ::${type}s($index)
	dvd_modified true false

	set indexlst [split $index ","]
	display_pgc ${type} [lindex $indexlst 0] [lindex $indexlst 1] [lindex $indexlst 2]
}
proc import_all_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set txt "PgcEdit will try to (re)import all the PGCs of the current DVD."
	set txt "$txt\nThe filenames must be in the standard Import/Export PGC format."
	set txt "$txt\nIf a file does't exist, the PGC will not be imported."
	set txt "$txt\n\nWarning: Only some basic checks are done to ensure that the datas"
	set txt "$txt\nare in the correct format."
	set txt "$txt\n\nContinue?"
	set r [::msgbox::MsgBox -title "PgcEdit: Import All PGC commands" -icon question \
			-message $txt -parent . -type yesno -default yes -skip "Import All PGC commands"]
	if {$r == "no"} {
		return
	}
	set dirname [tk_chooseOldDirectory -parent . \
		-initialdir $::config(import_initialdir) \
		-title "PgcEdit: Import All PGC commands from folder"]
	if {$dirname == ""} {
		return
	}
	set ::config(import_initialdir) $dirname

	dvd_modified true false
	set count   0
	set skipped 0
	set errors  0

	set progress_max  [expr ($::pgcs(numvts)+1)*2]

	set name [get_default_filename {::pgc 0,0}]
	set filename [file join $dirname $name]
	if {[file exists $filename]} {
		show_msg "Importing $name" 1 $progress_max
		if {[import_one_pgc $filename ::pgcs(0,0)]} {
			incr count
		} else {
			incr errors
		}
	} else {
		incr skipped
	}
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set name [get_default_filename [list "menu" "$vts,$lu,$pgc"]]
				set filename [file join $dirname $name]
				if {[file exists $filename]} {
					show_msg "Importing $name" [expr ($vts*2)+1] $progress_max
					if {[import_one_pgc $filename ::menus($vts,$lu,$pgc)]} {
						incr count
					} else {
						incr errors
					}
				} else {
					incr skipped
				}
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set name [get_default_filename [list "pgc" "$vts,$pgc"]]
				set filename [file join $dirname $name]
				if {[file exists $filename]} {
					show_msg "Importing $name" [expr ($vts*2)+2] $progress_max
					if {[import_one_pgc $filename ::pgcs($vts,$pgc)]} {
						incr count
					} else {
						incr errors
					}
				} else {
					incr skipped
				}
			}
		}
	}
	display_pgc "pgc" 0 0
	show_msg "$count PGCs imported OK.  $skipped PGCs skipped.  $errors errors."
}
proc export_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	set cmds  [subst \$::${type}s($index)]
	set filename [tk_getSaveFile -filetypes {{"PGC commands files" {*.pgc}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile [get_default_filename $::editor_cmds_from] -title "PgcEdit: Export PGC commands"]
	if {$filename == ""} {
		return
	}
	if {[string toupper [file extension $filename]] != ".PGC"} {
		set filename "${filename}.pgc"
	}
	set ::config(import_initialdir) [file dirname $filename]
	saveFile $filename binary $cmds
}
proc export_all_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set dirname [tk_chooseNewDirectory -parent . \
			-initialdir $::config(import_initialdir) \
			-title "Export All PGC commands to folder:"]
	if {$dirname == ""} {
		return
	}
	set ::config(import_initialdir) $dirname

	set count 0
	set progress_max  [expr ($::pgcs(numvts)+1)*2]

	set cmds $::pgcs(0,0)
	set name [get_default_filename {::pgc 0,0}]
	show_msg "Exporting $name" 1 $progress_max
	set filename [file join $dirname $name]
	if {[saveFile $filename binary $cmds] != 0} {
		show_msg "$count PGCs exported OK.  Error exporting $name!"
		bell
		return
	}
	incr count
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		update_progress $vts $::pgcs(numvts)
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set cmds $::menus($vts,$lu,$pgc)
				set name [get_default_filename [list "menu" "$vts,$lu,$pgc"]]
				show_msg "Exporting $name" [expr ($vts*2)+1] $progress_max
				set filename [file join $dirname $name]
				if {[saveFile $filename binary $cmds] != 0} {
					show_msg "$count PGCs exported OK.  Error exporting $name!"
					bell
					return
				}
				incr count
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set cmds $::pgcs($vts,$pgc)
				set name [get_default_filename [list "pgc" "$vts,$pgc"]]
				show_msg "Exporting $name" [expr ($vts*2)+2] $progress_max
				set filename [file join $dirname $name]
				if {[saveFile $filename binary $cmds] != 0} {
					show_msg "$count PGCs exported OK.  Error exporting $name!"
					bell
					return
				}
				incr count
			}
		}
	}
	show_msg "$count PGCs exported OK."
}
proc get_default_filename {from} {
	set type  [lindex $from 0]
	set index [lindex $from 1]
	set indexlst [split $index ","]
	if {[lindex $indexlst 0] == 0} {
		set vts "vmg"
	} else {
		set vts "vts[format "%02d" [lindex $indexlst 0]]"
	}
	if {$type == "menu"} {
		set defaultfilename "${vts}_menu[lindex $indexlst 1]_[format "%02d" [lindex $indexlst 2]].pgc"
	} else {
		if {[lindex $indexlst 0] == 0} {
			set ttn "fp"
		} else {
			set ttn "ttn[format "%02d" [lindex $indexlst 1]]"
		}
		set defaultfilename "${vts}_$ttn.pgc"
	}
}

#############################################################################
## Procedure:  set_listbox_selection
## set the selection of the listbox to the provided range
## Args: index1 ?index2? (if index2 is not provided, select only one line)
proc set_listbox_selection {idx1 args} {
	if {$args == ""} {set idx2 $idx1} else {set idx2 $args}
	if {[catch {$::widget(main_listbox) activate $idx1}]} {
		return 0
	}
	$::widget(main_listbox) selection clear 0 end
	$::widget(main_listbox) selection set $idx1 $idx1
	catch {
		$::widget(main_listbox) selection set $idx1 $idx2
		$::widget(main_listbox) see $idx2
		$::widget(main_listbox) see $idx1
	}
	return 1
}

#############################################################################
## Procedure:  get_selection_indexes
## retreave the first and last indexes of the selection in the listbox.
## return a list with the two indexes, or {- -} if nothing is selected.
## Note: the two indexes are identical if only one command is selected.
## Do NOT check for selection of pre, post or cell line.
proc get_selection_indexes {} {
	set selection [$::widget(main_listbox) curselection]
	if {$selection == ""} {
		return [list - -]
	}
	set idx1 [lindex $selection 0]
	set idx2 [lindex $selection end]
	return [list $idx1 $idx2]
}

#############################################################################
## Procedure:  get_type
## return a list with the type (pre/post/cell/scratchbook) and line number
## of the provided index.
## Return "" if the index is out of range.
proc get_type {idx} {
	if {$idx >= [$::widget(main_listbox) index end]} {return ""}
	for {set c 0} {$c <= $idx} {incr c} {
		set cmdtext [lindex $::editor_cmds_list $c]
		if {[string range $cmdtext 0 0] == "*"} {
			set type [lindex $cmdtext 1]
			set line 0
		} else {
			incr line
		}
	}
	return [list $type $line]
}

#############################################################################
## Procedure:  get_type_cmds
## return a list with all commands of the provided type (pre/post/cell/scratchbook).
proc get_type_cmds {type} {
	set ok false
	set r ""
	for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
		set cmdtext [lindex $::editor_cmds_list $c]
		if {[string range $cmdtext 0 0] == "*"} {
			if {[lindex $cmdtext 1] == $type} {
				set ok true
				continue
			} else {
				if {$ok} {return $r}
				set ok false
				continue
			}
		}
		if {$ok} {
			lappend r [string range $cmdtext 1 23]
		}
	}
	return $r
}

#############################################################################
## Procedure:  set_type_cmds
## replace all commands of the provided type (pre/post/cell/scratchbook).
proc set_type_cmds {type cmds} {
	set ok true
	set r ""
	for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
		set cmdtext [lindex $::editor_cmds_list $c]
		if {[string range $cmdtext 0 0] == "*"} {
			if {[lindex $cmdtext 1] == $type} {
				set ok false
				lappend r "* $type"
				set r [concat $r $cmds]
			} else {
				set ok true
			}
		}
		if {$ok} {
			lappend r $cmdtext
		}
	}
	set ::editor_cmds_list $r
	refresh_listbox
	return
}

#############################################################################
## Procedure:  renumber_gotos_del
## renumber the GoTos in the commands as if the commands from idx1 to idx2
## (inclusive) would be deleted.
## DO NOT actually delete the commands
proc renumber_gotos_del {cmds idx1 idx2} {
	set diff [expr {$idx1 - $idx2 - 1}]
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		if {$idx < [expr {$idx1 - 1}] || $idx >= $idx2} {
			set cmd [lindex $cmds $idx]
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [x2d [lindex $cmd end]]
				if {$ln > $idx2} {
					incr ln $diff
					lset cmd end [format "%02X" $ln]
					lset cmds $idx $cmd
				}
			}
		}
	}
	return $cmds
}

#############################################################################
## Procedure:  renumber_gotos_add
## renumber the GoTos in the commands as if the commands from idx1 to idx2
## (inclusive) would be added.
## The new commands must be added before calling this function
proc renumber_gotos_add {cmds idx1 idx2} {
	set diff [expr {$idx2 - $idx1 + 1}]
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		if {$idx < [expr {$idx1 - 1}] || $idx >= $idx2} {
			set cmd [lindex $cmds $idx]
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [x2d [lindex $cmd end]]
				if {$ln >= $idx1} {
					incr ln $diff
					lset cmd end [format "%02X" $ln]
					lset cmds $idx $cmd
				}
			}
		}
	}
	return $cmds
}

#############################################################################
## Procedure:  get_selected_cmds
## get the selected commands from the listbox
## return a list with the commands in hex form, or bell if none is selected,
## or if a pre, post or cell label is in the selection.
proc get_selected_cmds {{renumber true}} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return ""
	}
	set cmds {}
	if {$renumber && $::config(renumber_gotos)} {
		set numdeleted [expr [get_cmd_linenum ::editor_cmds_list $idx1]-1]
		for {set i 0} {$i < $numdeleted} {incr i} {
			lappend cmds {00 00 00 00 00 00 00 00}
		}
	}
	for {set cidx $idx1} {$cidx <= $idx2} {incr cidx} {
		set cmdtext [lindex $::editor_cmds_list $cidx]
		if {[string range $cmdtext 0 0] != {[} } {
			bell
			return ""
		}
		lappend cmds [string range $cmdtext 1 23]
	}
	if {$renumber && $::config(renumber_gotos) && $numdeleted > 0} {
		set cmds [lrange [renumber_gotos_del $cmds 0 [expr $numdeleted-1]] $numdeleted end]
	}
	return $cmds
}
#############################################################################
## Procedure:  delete_selected_cmds
## Return the original starting index of the selection, or "-" for error.
proc delete_selected_cmds {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return "-"
	}

	if {$::config(renumber_gotos)} {
		set type1list [get_type $idx1]
		set type [lindex $type1list 0]
		set cmds [get_type_cmds $type]
		set renumbered [renumber_gotos_del $cmds \
				[lindex $type1list 1] [lindex [get_type $idx2] 1]]

		set_type_cmds $type $renumbered
	}
	set ::editor_cmds_list [lreplace $::editor_cmds_list $idx1 $idx2]
	return $idx1
}

#############################################################################
## Procedure:  insert_cmds
## insert commands in the listbox, just before the provided index.
proc insert_cmds {index commands} {
	if {[expr {[llength $::editor_cmds_list]+[llength $commands]}] > 131} {
		bell
		return -1
	}

	if {$::config(renumber_gotos)} {
		set cmdtype [get_cmd_type ::editor_cmds_list [expr $index-1]]
		set cmdnum  [get_cmd_linenum ::editor_cmds_list [expr $index-1]]
		if {$cmdnum > 0} {
			set padcmds ""
			for {set i 0} {$i < $cmdnum} {incr i} {
				lappend padcmds {00 00 00 00 00 00 00 00}
			}
			set commands [lrange [renumber_gotos_add [concat $padcmds $commands] 1 $cmdnum] $cmdnum end]
		}
	}

	set formatted_commands ""
	foreach i $commands {
		lappend formatted_commands "\[$i\]"
	}
	set ::editor_cmds_list [concat \
			[lrange $::editor_cmds_list 0 [expr {$index -1}]] \
			$formatted_commands \
			[lrange $::editor_cmds_list $index end]]
	set ::editor_cmds_list_no_hex [concat \
			[lrange $::editor_cmds_list_no_hex 0 [expr {$index -1}]] \
			$formatted_commands \
			[lrange $::editor_cmds_list_no_hex $index end]]

	if {$::config(renumber_gotos)} {
		set type1list [get_type $index]
		set type [lindex $type1list 0]
		set typecmds [get_type_cmds $type]

		set renumbered [renumber_gotos_add $typecmds \
				[lindex $type1list 1] \
				[lindex [get_type [expr {$index + [llength $commands] - 1}]] 1]]

		set_type_cmds $type $renumbered
	}
	return [llength $::editor_cmds_list]
}

#############################################################################
## highlight goto line
proc highlight_goto {linenum jump} {
	set line [lindex $::editor_cmds_list $linenum]
	set cmd [string range $line 0 23]
	set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
	if {$cmdchk=="001" || $cmdchk=="003"} {
		set err 0
		set ln [x2d [string range $cmd 22 23]]
		if {$ln>0 && $ln<129} {
			set highln [expr {$linenum + $ln - [string range $line 26 28]}]
			if {![catch {set origcolor [$::widget(main_listbox) itemcget $highln -background]}]} {
				set destline [lindex $::editor_cmds_list $highln]
				if {[string range $destline 0 0]=="\["} {
					if {[expr {[string range $destline 26 28]}]==$ln} {
						$::widget(main_listbox) itemconfigure $highln -background lightblue
						after 333 "catch {$::widget(main_listbox) itemconfigure $highln \
								-background $origcolor}"
						if {$jump} {
							set_listbox_selection $linenum
							after 333 "catch {::main::set_listbox_selection $highln}"
						}
					} else {
						set err 1
					}
				} else {
					set err 1
				}
			} else {
				set err 1
			}
		} else {
			set err 1
		}
		if {$err} {
			$::widget(main_listbox) itemconfigure $linenum -selectbackground red
			if {$jump} bell
			if {$::config(highlight_gotos)} {
				set color #C09090
			} else {
				set color $::config(systemhighlight)
			}
			after 333 "$::widget(main_listbox) itemconfigure $linenum \
					-selectbackground $color"
		}
	}
}
#############################################################################
## highlight target PGC (if relevant)
proc highlight_jump {linenum jump} {
	if {$::scratchbook_active} {
		if {$jump} bell
		return 0
	}

	set cmd [string range [lindex $::editor_cmds_list $linenum] 0 23]
	set currenttype  [lindex $::editor_cmds_from 0]
	set currentindex [lindex $::editor_cmds_from 1]

	return [highlight_jump_cmd $currenttype $currentindex $cmd $linenum $jump true]
}
proc highlight_jump_cmd {currenttype currentindex cmd linenum jump highlight {setnvtmrtargettype 0}} {
	set cmd [string toupper $cmd]
	set currentvts  [lindex [split $currentindex ","] 0]
	set jumpfound false
	set targettype ""
	set targetvts -1
	set targetpgc -1
	set targetlu  "all"
	set targetcell 1	;#!!!
	if {$currenttype == "menu"} {
		set jumptolu [lindex [split $currentindex ","] 1]
	} else {
		set jumptolu 1
	}
	set targetsection "na"
	set targetline -1

	set linkpgc false

	set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
	switch $cmdchk {
		"201" {
			# Link*PGC
			set linkpgc true
		}
		"204" {
			# LinkPGCN
			set jumpfound true
			set targettype $currenttype
			set targetvts $currentvts
			if {$targettype == "menu"} {
				set targetlu $jumptolu
			}
			set targetpgc [x2d [string range $cmd 19 20][string range $cmd 22 23]]
		}
		"205" {
			# Link_PTT
			set jumpfound true
			set targettype "vts"
			if {$currentvts == 0} {
				set targetvts -1
			} else {
				set targetvts $currentvts
				set currentpgc [lindex [split $currentindex ","] 1]
				if {[info exists ::pgcs($targetvts,$currentpgc,TTN)]} {
					set targetttn $::pgcs($targetvts,$currentpgc,TTN)
				} else {
					set targetttn 0
				}
				set targetptt [expr {[x2d [string range $cmd 20 20][string range $cmd 22 23]] & 0x03FF}]
				for {set pgc 1} {$pgc <= $::pgcs($targetvts,numpgcs)} {incr pgc} {
					if {[info exists ::pgcs($targetvts,$pgc,TTN)] && \
							$::pgcs($targetvts,$pgc,TTN) == $targetttn} {
						foreach j $::pgcs($targetvts,$pgc,pttlist) {
							if {[lindex $j 1] == $targetptt} {
								set targetpgc $pgc
								break
							}
						}
					}
					if {$targetpgc != -1} break
				}
			}
		}
		"302" {
			# JumpTT
			set jumpfound true
			set title [x2d [string range $cmd 16 17]]
			if {$title>0 && $title<100} {
				if {[info exists ::pgcs(Title,$title)]} {
					set vts [lindex $::pgcs(Title,$title) 0]
					set pgc [lindex $::pgcs(Title,$title) 1]
					if {[info exists ::pgcs($vts,$pgc)]} {
						set targettype "vts"
						set targetvts $vts
						set targetpgc $pgc
					}
				}
			}
		}
		"303" -
		"305" {
			# JumpVTS_TT and JumpVTS_PTT
			set jumpfound true
			set targettype "vts"
			if {$currentvts > 0} {
				set targetvts $currentvts
				set targetttn [x2d [string range $cmd 16 17]]
				if {[info exists ::ttns($targetvts,$targetttn)]} {
					if {$cmdchk == "305"} {
						# JumpVTS_PTT must check if the PTT
						# is in the first PGC of a multi-PGC title
						set targetptt [x2d [string range $cmd 8 8][string range $cmd 10 11]]
						for {set pgc 1} {$pgc <= $::pgcs($targetvts,numpgcs)} {incr pgc} {
							if {[info exists ::pgcs($targetvts,$pgc,TTN)] && \
									$::pgcs($targetvts,$pgc,TTN) == $targetttn} {
								foreach j $::pgcs($targetvts,$pgc,pttlist) {
									if {[lindex $j 1] == $targetptt} {
										set targetpgc $pgc
										break
									}
								}
							}
							if {$targetpgc != -1} break
						}
					} else {
						set targetpgc $::ttns($targetvts,$targetttn)
					}
				}
			}
		}
		"306" -
		"308" {
			# JumpSS and CallSS
			set jumpfound true
			set targettype "menu"
			set jumptype [string range $cmd 16 16]
			switch $jumptype {
				"0" { # FP
					set targettype "pgc"
					set targetvts 0
					set targetpgc 0
				}
				"4" { # VMGM-menu
					if {[info exists ::menus(0,$jumptolu,numpgcs)]} {
						for {set pgc 1} {$pgc <= $::menus(0,$jumptolu,numpgcs)} {incr pgc} {
							if {[expr {$::menus(0,$jumptolu,$pgc,mask) & 0x7F}] == 2} {
								set targetvts 0
								set targetpgc $pgc
							}
						}
					} else {
						set targetvts 0
						set targetpgc 0
					}
				}
				"8" { # VTSM-menu
					if {$cmdchk == "306"} {
						set vts [x2d [string range $cmd 13 14]]
					} else {
						set vts $currentvts
					}
					if {$vts == 0} {	;# VTS=0 means Last visited VTS, or current VTS
						if {$currentvts > 0} {
							set vts $currentvts
						} else {
							set jumpfound false
						}
					}
					if {$jumpfound} {
						set menutype [x2d [string range $cmd 17 17]]
						set targetlu [target_lu $currenttype $currentindex $vts]
						if {$targetlu != -1} {
							if {[info exists ::menus($vts,$targetlu,numpgcs)]} {
								for {set pgc 1} {$pgc <= $::menus($vts,$targetlu,numpgcs)} {incr pgc} {
									if {[expr {$::menus($vts,$targetlu,$pgc,mask) & 0x7F}] == $menutype} {
										set targetvts $vts
										set targetpgc $pgc
									}
								}
							} else {
								set targetvts 0
								set targetpgc 0
							}
						}
					}
				}
				"C" { # VMGM-pgc
					set targetvts 0
					set targetpgc [x2d [string range $cmd 10 11]]
				}
			}
		}
		"411" - "421" - "431" - "441" - "461" - "511" - "521" - "531" - "541" - "561" -
		"611" - "621" - "631" - "641" - "651" - "661" - "671" - "681" - "691" - "6A1" - "6B1" -
		"711" - "721" - "731" - "741" - "751" - "761" - "771" - "781" - "791" - "7A1" - "7B1" {
			# Set SPRM and Link*PGC
			set linkpgc true
		}
		"414" - "424" - "434" - "444" - "464" - "514" - "524" - "534" - "544" - "564" -
		"614" - "624" - "634" - "644" - "654" - "664" - "674" - "684" - "694" - "6A4" - "6B4" -
		"714" - "724" - "734" - "744" - "754" - "764" - "774" - "784" - "794" - "7A4" - "7B4" {
			# Set SPRM and LinkPGCN
			set jumpfound true
			set targettype $currenttype
			set targetvts $currentvts
			if {$targettype == "menu"} {
				set targetlu $jumptolu
			}
			set targetpgc [x2d [string range $cmd 19 20][string range $cmd 22 23]]
		}
	}
	switch "[string range $cmd 1 2]" {
		"42" - "52" {
			# SetNVTMR
			# BUG: WARNING: ONLY ONE TARGET IS RETURNED!
			# THE TARGET NAV TIMER PGC IS IGNORED IF A LINK IS PRESENT!
			if {$setnvtmrtargettype==0 || [string range $cmd 5 5]==0} {
				set jumpfound true
				set targettype $currenttype
				set targetvts $currentvts
				set targetpgc [x2d [string range $cmd 13 14][string range $cmd 16 17]]
			}
		}
	}
	switch "[string range $cmd 1 1]" {
		"8" - "9" - "A" - "B" - "C" - "D" {
			# CSetCLnk...
			set linkpgc true
		}
	}

	if {$linkpgc} {
		set targettype $currenttype
		set targetvts  $currentvts
		if {$targettype == "menu"} {
			set currentlu  $jumptolu
			set currentpgc [lindex [split $currentindex ","] 2]
			if {[catch {set pgcheader $::menus($currentvts,$currentlu,$currentpgc,header)}]} {
				set pgcheader [string repeat \000 162]
			}
		} else {
			set currentlu  "all"
			set currentpgc [lindex [split $currentindex ","] 1]
			if {[catch {set pgcheader $::pgcs($currentvts,$currentpgc,header)}]} {
				set pgcheader [string repeat \000 162]
			}
		}
		switch "[string range $cmd 22 23]" {
			"09" {
				# LinkTopPGC
				set jumpfound true
				set targetpgc $currentpgc
				set targetsection "pre"
			}
			"0D" {
				# LinkTailPGC
				set jumpfound true
				set targetpgc $currentpgc
				set targetsection "post"
			}
			"0A" {
				# LinkNextPGC
				set jumpfound true
				set tmp_targetpgc [get_int pgcheader 0X9C 2]
				if {$currenttype == "pgc"} {
					if {[info exists ::pgcs($currentvts,$tmp_targetpgc,TTN)] && \
							$::pgcs($currentvts,[lindex [split $currentindex ,] 1],TTN) \
							== $::pgcs($currentvts,$tmp_targetpgc,TTN)} {
						set targetpgc $tmp_targetpgc
					} else {
						set targetpgc -1
					}
				} else {
					set targetpgc $tmp_targetpgc
					return 	[list $targettype $targetvts $targetlu $targetpgc true]
				}
			}
			"0B" {
				# LinkPrevPGC
				set jumpfound true
				set tmp_targetpgc [get_int pgcheader 0X9E 2]
				if {$currenttype == "pgc"} {
					if {[info exists ::pgcs($currentvts,$tmp_targetpgc,TTN)] && \
							$::pgcs($currentvts,[lindex [split $currentindex ,] 1],TTN) \
							== $::pgcs($currentvts,$tmp_targetpgc,TTN)} {
						set targetpgc $tmp_targetpgc
					} else {
						set targetpgc -1
					}
				} else {
					set targetpgc $tmp_targetpgc
					return 	[list $targettype $targetvts $targetlu $targetpgc true]
				}
			}
			"0C" {
				# LinkGoUpPGC
				set jumpfound true
				set tmp_targetpgc [get_int pgcheader 0XA0 2]
				if {$currenttype == "pgc"} {
					if {[info exists ::pgcs($currentvts,$tmp_targetpgc,TTN)] && \
							$::pgcs($currentvts,[lindex [split $currentindex ,] 1],TTN) \
							== $::pgcs($currentvts,$tmp_targetpgc,TTN)} {
						set targetpgc $tmp_targetpgc
					} else {
						set targetpgc -1
					}
				} else {
					set targetpgc $tmp_targetpgc
					return 	[list $targettype $targetvts $targetlu $targetpgc true]
				}
			}
		}
	}

	set err false
	if {$jumpfound} {
		if {$targetpgc == -1} {
			set err true
		} else {
			if {$targetsection != "na"} {
				if {$targetsection == "pre"} {
					set targetline 0
				} else {
					set targetline -1
					for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
						set cmd [lindex $::editor_cmds_list $c]
						if {[lindex $cmd 1] == "post"} {
							set targetline $c
							break
						}
					}
				}
				if {$targetline != -1 && $linenum != 0} {
					set line1 [expr {$targetline+1}]
					set cmd1 [lindex $::editor_cmds_list $line1]
					if {[string range $cmd1 0 0] == "\["} {
						set targetline $line1
					}
					if {$highlight} {
						set origcolor [$::widget(main_listbox) itemcget $targetline -background]
						$::widget(main_listbox) itemconfigure $targetline -background lightblue
						after 333 "catch {$::widget(main_listbox) itemconfigure $targetline \
								-background $origcolor}"
					}
					if {$jump} {
						if {$linenum != -1} {set_listbox_selection $linenum}
						after 400 "catch {::main::set_listbox_selection $targetline}"
					}
				}
			} else {
				if {$targetvts > $::pgcs(numvts)} {
					set label "ERROR"
				} else {
					if {[catch {
						if {$targettype == "menu"} {
							set label [pgc_label $::config(global_font) "menu" $targetvts [target_lu $currenttype $currentindex $targetvts] $targetpgc]
						} else {
							set label [pgc_label $::config(global_font) "pgc" $targetvts $targetpgc]
						}
					}]} {
						set label "ERROR"
					}
				}
				set highln [lsearch $::main::labelslist $label]
				if {$highln != -1} {
					if {$targettype == "menu"} {
						if {$targetlu == "all"} {
							set start 1
							if {[catch {set end $::menus($targetvts,numlus)}]} {
								set end 1
							}
						} else {
							set start $targetlu
							set end   $targetlu
						}
						for {set lu $start} {$lu <= $end} {incr lu} {
							if {$highlight && [winfo exists .mf.f2.selector.a.lb]} {
								if {! [catch {set label [pgc_label $::config(global_font) "menu" $targetvts $lu $targetpgc]}]} {
									set cmd2 ""
									if {$::main::selector_opened} {
										set origcolor [.mf.f2.selector.a.lb itemcget $highln -background]
										if {$origcolor != "lightblue"} {
											if {$origcolor == ""} {set origcolor $::config(system_window_color)}
											if {! [info exists ::main::selector_shifted]} {
												set oldpos [lindex [.mf.f2.selector.a.lb yview] 0]
												.mf.f2.selector.a.lb see $highln
												if {!$jump && [lindex [.mf.f2.selector.a.lb yview] 0] != $oldpos} {
													set ::main::selector_shifted true
													set cmd2 ".mf.f2.selector.a.lb yview moveto $oldpos; \
															unset ::main::selector_shifted"
												}
											}
											.mf.f2.selector.a.lb itemconfigure $highln -background lightblue
											set cmd1 ".mf.f2.selector.a.lb itemconfigure $highln -background $origcolor"
											after 1000 "$cmd1; $cmd2"
										}
									}
								}
							}
						}
						if {$jump} {
							if {$targetlu == "all"} {set targetlu 1}
							display_pgc "menu" $targetvts $targetlu $targetpgc
						}
					} else {
						if {$highlight && [winfo exists .mf.f2.selector.a.lb]} {
							set cmd2 ""
							if {$::main::selector_opened} {
								set origcolor [.mf.f2.selector.a.lb itemcget $highln -background]
								if {$origcolor != "lightblue"} {
									if {$origcolor == ""} {set origcolor $::config(system_window_color)}
									if {! [info exists ::main::selector_shifted]} {
										set oldpos [lindex [.mf.f2.selector.a.lb yview] 0]
										.mf.f2.selector.a.lb see $highln
										if {!$jump && [lindex [.mf.f2.selector.a.lb yview] 0] != $oldpos} {
											set ::main::selector_shifted true
											set cmd2 ".mf.f2.selector.a.lb yview moveto $oldpos; \
													unset ::main::selector_shifted"
										}
									}
									.mf.f2.selector.a.lb itemconfigure $highln -background lightblue
									set cmd1 ".mf.f2.selector.a.lb itemconfigure $highln -background $origcolor"
									after 1000 "$cmd1; $cmd2"
								}
							}
						}
						if {$jump} {
							display_pgc "pgc" $targetvts $targetpgc
						}
					}
				} else {
					set err true
				}
			}
		}
	}
	if {$err} {
		if {$highlight} {
			if {$linenum < 1} {
				bell
			} else {
				$::widget(main_listbox) itemconfigure $linenum -background orange -selectbackground red
				if {$jump} bell
				if {$::config(highlight_gotos)} {
					set color #AA7000
				} else {
					set color $::config(systemhighlight)
				}
				after 333 "$::widget(main_listbox) itemconfigure $linenum -selectbackground $color"
			}
		}
		return 	[list $targettype $targetvts $targetlu $targetpgc false]
	} else {
		if {$jumpfound} {
			if {$targettype == "menus"} {set targettype "menu"}
			return 	[list $targettype $targetvts $targetlu $targetpgc true]
		} else {
			return ""
		}
	}
}
#############################################################################
## return the LU number in the target domain matching the LU language of the
## source lu, or 1 if there is no matching language.
proc target_lu {currenttype currentindex targetvts} {
	if {$currenttype == "pgc" || $currenttype == "vts"} {
		return 1
	}
	set idx [split $currentindex ,]
	if {[lindex $idx 1]=="all"} {
		return 1
	}
	if {$targetvts > $::pgcs(numvts)} {
		return -1
	}
	set lu 1
	if {[catch {set langcode $::menus([lindex $idx 0],[lindex $idx 1],lang)}]} {
		return 1
	}
	for {set l 1} {$l <= $::menus($targetvts,numlus)} {incr l} {
		if {$::menus($targetvts,$l,lang) == $langcode} {
			set lu $l
			break
		}
	}
	return $lu
}


#############################################################################
## highlight goto/jump/call/link line, and eventually jump to it
## command bound to mouse button 1 on a line of the main listbox
proc highlight_goto_bound {jump} {
	set linenum [$::widget(main_listbox) index \
			@[expr {[winfo pointerx .]-[winfo rootx $::widget(main_listbox)]}],[expr {[winfo pointery .]-[winfo rooty $::widget(main_listbox)]}]]
	highlight_goto $linenum $jump
	highlight_jump $linenum $jump
}

#############################################################################
## Find procedures
proc find_in_string {expr str case regexpr} {
	if {$regexpr} {
		if {$case} {
			set found [regexp -- $expr $str]
		} else {
			set found [regexp -nocase -- $expr $str]
		}
		return $found
	} else {
		if {!$case} {
			set expr [string toupper $expr]
			set str  [string toupper $str]
		}
		if {[string first $expr $str] >= 0} {
			return 1
		} else {
			return 0
		}
	}
}
proc find_in_pgc {direction} {
	# check expression sybtax
	if {$::main::searchstring == ""} {
		show_msg "Type a search string first!"
		return -1
	}
	if {$::main::searchregexp} {
		set err [catch {set found [regexp -- $::main::searchstring "blah blah"]} errstr]
		if {$err != 0} {
			bell
			show_msg [string map {": " ":\n"} $errstr]
			return -1
		}
	}

	if {$::scratchbook_active} {
		set cmds  $::scratchbook_buffer
	} else {
		set type  [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]
		set cmds  [subst \$::${type}s($index)]
	}

	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		set idx1 0
		set idx2 0
	}
	set found 0
	if {$direction == "-forwards"} {
		set start [expr {$idx2 + 1}]
		set end [expr {[llength $cmds] - 1}]
		for {set i $start} {$i <= $end} {incr i} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found != 0} break
		}
	} else {
		set start [expr {$idx1 - 1}]
		set end 1
		for {set i $start} {$i >= $end} {incr i -1} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found != 0} break
		}
	}
	if {$found == 1} {
		set_listbox_selection $i
		highlight_goto $i false
		highlight_jump $i false
	}
	return $found
}
proc find_in_cmds {expr cmds direction case regexpr {start 0}} {
	if {$direction == "-forwards"} {
		for {set i $start} {$i < [llength $cmds]} {incr i} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $expr [::disasm::formatCommand $cmd $i] $case $regexpr]
			if {$found == 1}  {return $i}
		}
	} else {
		for {set i [expr {[llength $cmds] - 1}]} {$i >= 0} {incr i -1} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $expr [::disasm::formatCommand $cmd $i] $case $regexpr]
			if {$found == 1} {return $i}
		}
	}
	return -1
}
proc find_in_dvd {direction} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		show_msg "Find in DVD doesn't work with the Scratchbook!"
		return -1
	}

	# check expression sybtax
	if {$::main::searchstring == ""} {
		show_msg "Type a search string first!"
		return -1
	}
	if {$::main::searchregexp} {
		set err [catch {set found [regexp -- $::main::searchstring "blah blah"]} errstr]
		if {$err != 0} {
			bell
			show_msg [string map {": " ":\n"} $errstr]
			return -1
		}
	}

	show_msg "Searching in DVD..."

	# try to find in the current PGC first
	set found [find_in_pgc $direction]
	if {$found != 0} {
		show_msg ""
		return 1
	}

	# build forward or backward links for each pgc
	if {$direction == "-forwards"} {
		set last ""
		for {set vts $::pgcs(numvts)} {$vts >= 0} {incr vts -1} {
			if {$vts > 0} {
				for {set pgc $::pgcs($vts,numpgcs)} {$pgc >= 1} {incr pgc -1} {
					set next(pgc,$vts,$pgc) $last
					set last "pgc $vts $pgc"
				}
			}
			for {set lu $::menus($vts,numlus)} {$lu >= 1} {incr lu -1} {
				for {set pgc $::menus($vts,$lu,numpgcs)} {$pgc >= 1} {incr pgc -1} {
					set next(menu,$vts,$lu,$pgc) $last
					set last "menu $vts $lu $pgc"
				}
			}
		}
		set next(pgc,0,0) $last
	} else {
		set next(pgc,0,0) ""
		set last "pgc 0 0"
		for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set next(menu,$vts,$lu,$pgc) $last
					set last "menu $vts $lu $pgc"
				}
			}
			if {$vts > 0} {
				for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
					set next(pgc,$vts,$pgc) $last
					set last "pgc $vts $pgc"
				}
			}
		}
	}

	# find in previous or next PGCs
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	while {$next($type,$index) != ""} {
		set lnext $next($type,$index)
		set type  [lindex $lnext 0]
		set vts   [lindex $lnext 1]
		if {$type == "menu"} {
			set index "$vts,[lindex $lnext 2],[lindex $lnext 3]"
			set cmds $::menus($index)
		} else {
			set index "$vts,[lindex $lnext 2]"
			set cmds $::pgcs($index)
		}
		set line [find_in_cmds $::main::searchstring $cmds $direction \
				$::main::searchcase $::main::searchregexp]
		if {$line >= 0} {
			if {$type == "menu"} {
				display_pgc $type $vts [lindex $lnext 2] [lindex $lnext 3]
			} else {
				display_pgc $type $vts [lindex $lnext 2]
			}
			set_listbox_selection $line
			show_msg ""
			return 1
		}
	}
	bell
	show_msg "Search string not found."
	return 0
}

#############################################################################
## Action Procedures
## procs called by the action buttons (below the listbox)
## There are no args to theses commands

proc action_Edit_Cmd {} {
	set idxs [get_selection_indexes]
	set idx [lindex $idxs 0]
	if {$idx!=[lindex $idxs 1]} {
		bell
		return
	}
	set cmd [get_selected_cmds false]
	if {$cmd == ""} {
		bell
		return
	}
	set ::hexcommand [lindex $cmd 0]

	if {$::scratchbook_active} {
		set cmdstype all
		set vts -1
		set area scratchbook
	} else {
		set cmdstype [lindex $::editor_cmds_from 0]
		set cmdsidx  [lindex $::editor_cmds_from 1]
		set vts [lindex [split $cmdsidx ,] 0]
		set area [get_cmd_type ::${cmdstype}s($cmdsidx) $idx]
	}
	if {[::ced::cmdeditor $cmdstype $vts $area] == 1} {
		lset ::editor_cmds_list $idx $::hexcommand
		refresh_listbox
		listbox_to_cmds
		highlight_goto $idx false
		highlight_jump $idx false
		dvd_modified true true
	}
}

proc action_Edit_PGC {{default_field ""}} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set vtsnum [lindex [split $index ","] 0]
	set pgcheader    [subst \$::${type}s($index,header)]
	set numprograms  [subst \$::${type}s($index,numprograms)]
	set programmap   [subst \$::${type}s($index,programmap)]
	set numcells     [subst \$::${type}s($index,numcells)]
	set cellpbtable  [subst \$::${type}s($index,cellpbtable)]
	set cellpostable [subst \$::${type}s($index,cellpostable)]
	set pttlist ""
	set usercomment ""
	if {$type == "pgc"} {
		if {[info exists ::${type}s($index,pttlist)]} {
			set pttlist [subst \$::${type}s($index,pttlist)]
		}
		if {[info exists ::pgcs_comments($index)]} {
			set usercomment $::pgcs_comments($index)
		}
	} else {
		if {[info exists ::menus_comments($index)]} {
			set usercomment $::menus_comments($index)
		}
	}
	set r [::pgced::edit_pgc $type $index $pgcheader $numprograms $programmap \
			$numcells $cellpbtable $cellpostable $pttlist $usercomment $default_field]
	if {$r != ""} {
		set ::${type}s($index,header)       [lindex $r 0]
		set ::${type}s($index,cellpbtable)  [lindex $r 1]
		set ::${type}s($index,cellpostable) [lindex $r 2]
		set ::${type}s($index,programmap)	[lindex $r 3]
		set ::${type}s($index,numprograms)	[string length [lindex $r 3]]
		if {[string range [lindex $r 3] end end] == "\000"} {
			incr ::${type}s($index,numprograms)	-1
		}
		set ::${type}s($index,pttlist)		[lindex $r 4]
		set ::${type}s($index,numcells)     [lindex $r 5]
		set newusercomment					[lindex $r 6]
		if {$newusercomment!=$usercomment} {
			set ::pgc_comments_modified true
			if {$newusercomment == ""} {
				if {$type == "pgc"} {
					unset -nocomplain ::pgcs_comments($index)
				} else {
					unset -nocomplain ::menus_comments($index)
				}
			} else {
				if {$type == "pgc"} {
					set ::pgcs_comments($index) $newusercomment
				} else {
					set ::menus_comments($index) $newusercomment
				}
			}
		}

		# Fixes the number of chapters in the VMG table of titles
		if {$type == "pgc" && [info exists ::pgcs($vtsnum,[lindex [split $index ","] 1],TTN)]} {
			set thisttn $::pgcs($vtsnum,[lindex [split $index ","] 1],TTN)
			if {$thisttn > 0} {
				set numchapters 0
				set thistitle 0
				for {set i 1} {$i <= $::pgcs($vtsnum,numpgcs)} {incr i} {
					if {[info exists ::pgcs($vtsnum,$i,TTN)] && $::pgcs($vtsnum,$i,TTN) == $thisttn} {
						incr numchapters [llength $::pgcs($vtsnum,$i,pttlist)]
						if {[info exists ::pgcs($vtsnum,$i,Title)] && $::pgcs($vtsnum,$i,Title)!= 0 \
								&& $::pgcs($vtsnum,$i,Title)!=""} {
							set thistitle $::pgcs($vtsnum,$i,Title)
						}
					}
				}
				if {$thistitle != 0} {
					set data $::raw_sectors(0,VMG_TT_SRPT)
					set srptnumtitles [get_int data 0 2]
					for {set i 1} {$i <= $srptnumtitles} {incr i} {
						set entryvts [get_int data [expr {($i*12)+2}] 1]
						set entryttn [get_int data [expr {($i*12)+3}] 1]
						if {$entryvts == $vtsnum && $entryttn == $thisttn} {
							set_int data [expr {($i*12)-2}] $numchapters 2
							set ::raw_sectors(0,VMG_TT_SRPT) $data
							break
						}
					}
				}
			}
		}

		set numbuts 0
		if {$type == "menu"} {
			for {set c 1} {$c <= $::menus($index,numcells)} {incr c} {
				set idx [expr {($c - 1) * 4}]
				set vobid  [get_int ::menus($index,cellpostable) $idx 2]
				set cellid [get_int ::menus($index,cellpostable) [expr {$idx + 3}] 1]
				if {[info exists ::buttons::butsinfo(menu,$vtsnum,$vobid,$cellid,numbuts)]} {
					incr numbuts $::buttons::butsinfo(menu,$vtsnum,$vobid,$cellid,numbuts)
				}
			}
		} else {
			for {set c 1} {$c <= $::pgcs($index,numcells)} {incr c} {
				set idx [expr {($c - 1) * 4}]
				set vobid  [get_int ::pgcs($index,cellpostable) $idx 2]
				set cellid [get_int ::pgcs($index,cellpostable) [expr {$idx + 3}] 1]
				if {[info exists ::buttons::butsinfo(pgc,$vtsnum,$vobid,$cellid,numbuts)]} {
					incr numbuts $::buttons::butsinfo(pgc,$vtsnum,$vobid,$cellid,numbuts)
				}
			}
		}
		set ::pgcs($index,numbuttons) $numbuts

		# verify if we must change the not-one_sequential flag of the title
		if {$type=="pgc" && $vtsnum!=0} {
			set pgc [lindex [split $index ,] end]
			set seqstatus [::srptinfo::get_real_not_sequential_title_status $vtsnum $pgc]
			if {$seqstatus==-1} {
				set pgc $::ttns($vtsnum,$::pgcs($index,TTN))
				set seqstatus [::srptinfo::get_real_not_sequential_title_status $vtsnum $pgc]
			}
			if {$seqstatus!=-1} {
				set idx [expr ($::pgcs($vtsnum,$pgc,Title)*12)-4]
				set titletype [get_int ::raw_sectors(0,VMG_TT_SRPT) $idx 1]
				if {[expr $titletype & 0x40]!=0 && $seqstatus==0} {
					set txt "Title $::pgcs($vtsnum,$pgc,Title) is flagged as NOT-one_sequential_title in the Title Play Map table."
					set txt "$txt\n(The user cannot normally seek to a specific time in not-one_sequential Titles.)"
					set txt "$txt\n\nHowever, as this Title is not made of several PGCs, and is not played at random"
					set txt "$txt\nor shuffle mode, there is probably no need to declare it as not-one_sequential."
					set txt "$txt\n\nDo you want to clear the not-one_sequential_title flag for Title $::pgcs($vtsnum,$pgc,Title)?  (Recommended)"
					set r [tk_messageBox -title "Edit PGC" -message $txt \
							-type yesno -default yes -icon question]
					if {$r=="yes"} {
						# then clear the not-one_sequential flag in Title Play Map table
						set_int ::raw_sectors(0,VMG_TT_SRPT) $idx [expr $titletype & 0xBF] 1
					}
				}
				if {[expr $titletype & 0x40]==0 && $seqstatus==1} {
					set txt "Title $::pgcs($vtsnum,$pgc,Title) is flagged as one_sequential_title in the Title Play Map table."
					set txt "$txt\nHowever, due to the Title characteristics or to its PG playback mode, it is probably not\nsafe to declare it so."
					set txt "$txt\n\nSetting the not-one_sequential_title flag has one drawback: normally, the user cannot"
					set txt "$txt\nseek to a specific time in non-sequential Titles.  However, some non-compliant players"
					set txt "$txt\nlet it use that function anyway."
					set txt "$txt\n\nDo you want to set the not-one_sequential_title flag for Title $::pgcs($vtsnum,$pgc,Title)?"
					set r [tk_messageBox -title "Edit PGC" -message $txt \
							-type yesno -default yes -icon question]
					if {$r=="yes"} {
						# set the not-one_sequential flag in Title Play Map table
						set_int ::raw_sectors(0,VMG_TT_SRPT) $idx [expr $titletype | 0x40] 1
					}
				}
			}

		}

		init_tree_menu_no_move
		dvd_modified true true
	}
}

proc action_Undo_pgc {} {
	if {$::scratchbook_active} {
		if {[file exists $::config(default_scratchbook_filename)]} {
			set ::scratchbook_buffer [readFile $::config(default_scratchbook_filename) binary]
			set ::editor_cmds_list $::scratchbook_buffer
			refresh_listbox
			set_listbox_selection 0
		} else bell
		return
	}
	if {[info exists ::editor_undo_buffer]} {
		set ::editor_cmds_list $::editor_undo_buffer
		refresh_listbox
		set_listbox_selection 0
		listbox_to_cmds
		dvd_modified true true
	} else bell
}

proc action_Delete {} {
	if {[get_selected_cmds] != ""} {
		set newidx [delete_selected_cmds]
		if {$newidx != ""} {
			refresh_listbox
			set_listbox_selection $newidx
			if {[lindex [get_selection_indexes] 0] == "-"} {
				set_listbox_selection end
			}
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_delete_all_commands {} {
	if {$::scratchbook_active} {
		set ::editor_cmds_list {{* scratchbook}}
	} else {
		set ::editor_cmds_list {{* pre} {* post} {* cell}}
		dvd_modified true true
	}
	refresh_listbox
	set_listbox_selection 0
	listbox_to_cmds false
}

proc action_Duplicate {} {
	set newcmds [get_selected_cmds]
	if {$newcmds != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {[insert_cmds [expr {$idx2+1}] $newcmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx1 $idx2
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Add_before {} {
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Add_after {} {
	set idx [lindex [get_selection_indexes] 1]
	if {$idx != "-"} {
		incr idx 1
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Jump_to_target {} {
	set idxs [get_selection_indexes]
	set idx [lindex $idxs 0]
	if {$idx!=[lindex $idxs 1]} {
		bell
		return
	}
	set cmd [get_selected_cmds]
	if {$cmd == ""} {
		bell
		return
	}
	highlight_goto $idx true
	highlight_jump $idx true
}

# Move selected commands up in the list
# (in fact, take the command that is above the selection ($cutcmd) and move it below the selection)
proc action_Move_up {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		set cutidx [expr {$idx1 - 1}]
		if {$idx1 < 2 || [string range [lindex $::editor_cmds_list $cutidx] 1 1] == "*"} {
			bell
			return
		}

		set typelist  [get_type [expr {$idx1-1}]]
		set type      [lindex $typelist 0]
		set srccmdnum [lindex $typelist 1]
		set dstcmdnum [expr {$srccmdnum + $idx2 - $idx1 + 1}]
		set typecmds  [get_type_cmds $type]

		if {$::config(renumber_gotos)} {
			for {set c 1} {$c <= [llength $typecmds]} {incr c} {
				set cmd [lindex $typecmds [expr {$c-1}]]
				set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
				if {$cmdchk=="001" || $cmdchk=="003"} {
					set gotoln [x2d [string range $cmd 21 22]]
					if {$gotoln == $srccmdnum} {
						set gotoln $dstcmdnum
					} elseif {$gotoln > $srccmdnum && $gotoln <= $dstcmdnum} {
						incr gotoln -1
					}
					lset typecmds [expr {$c-1}] \
						"[string range $cmd 0 19] [format %02X $gotoln]"
				}
			}
		}

		# move cutcmd
		set idx [expr {$srccmdnum-1}]
		set cutcmd [lindex $typecmds $idx]
		set typecmds [lreplace $typecmds $idx $idx]
		set typecmds [linsert $typecmds [expr {$dstcmdnum-1}] $cutcmd]

		set_type_cmds $type $typecmds

		refresh_listbox
		set_listbox_selection [expr {$idx1-1}] [expr {$idx2-1}]
		listbox_to_cmds
		dvd_modified true true
	}
}

# Move selected commands down in the list
# (in fact, take the command that is below the selection ($cutcmd) and move it above the selection)
proc action_Move_down {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		set cutidx [expr {$idx2 + 1}]
		if {$cutidx >= [expr {[llength $::editor_cmds_list]}] \
				|| [string range [lindex $::editor_cmds_list $cutidx] 1 1] == "*"} {
			bell
			return
		}

		set typelist  [get_type [expr {$idx2+1}]]
		set type      [lindex $typelist 0]
		set srccmdnum [lindex $typelist 1]
		set dstcmdnum [expr {$srccmdnum - ($idx2 - $idx1) - 1}]
		set typecmds  [get_type_cmds $type]

		if {$::config(renumber_gotos)} {
			for {set c 1} {$c <= [llength $typecmds]} {incr c} {
				set cmd [lindex $typecmds [expr {$c-1}]]
				set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
				if {$cmdchk=="001" || $cmdchk=="003"} {
					set gotoln [x2d [string range $cmd 21 22]]
					if {$gotoln == $srccmdnum} {
						set gotoln $dstcmdnum
					} elseif {$gotoln < $srccmdnum && $gotoln >= $dstcmdnum} {
						incr gotoln
					}
					lset typecmds [expr {$c-1}] \
						"[string range $cmd 0 19] [format %02X $gotoln]"
				}
			}
		}

		# move cutcmd
		set idx [expr {$srccmdnum-1}]
		set cutcmd [lindex $typecmds $idx]
		set typecmds [lreplace $typecmds $idx $idx]
		set typecmds [linsert $typecmds [expr {$dstcmdnum-1}] $cutcmd]

		set_type_cmds $type $typecmds

		refresh_listbox
		set_listbox_selection [expr {$idx1+1}] [expr {$idx2+1}]
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Copy {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		if {[llength $cmds] == 1} {
			set cmds [string trim $cmds "{}"]
		}
		clipboard clear
		clipboard append -- $cmds
	}
}

proc send_to_clipboard {} {
	if {$::scratchbook_active} {
		bell
		return
	}
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		foreach c $cmds {
			lappend ::scratchbook_buffer $c
		}
	}
}

proc copy_as_text {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == ""} {
		bell
	} else {
		clipboard clear
		for {set line $idx1} {$line <= $idx2} {incr line} {
			clipboard append -- "[lindex $::editor_cmds_list $line]"
			if {$line < $idx2} {
				clipboard append -- "\n"
			}
		}
	}
}

proc action_Append {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		set oldcmds [action_Paste_get_clipboard]
		if {$oldcmds != ""} {
			set cmds [concat $oldcmds $cmds]
			if {$::config(renumber_gotos)} {
				set cmds [renumber_gotos_add $cmds 0 [expr [llength $oldcmds]-1]]
			}
			clipboard clear
			clipboard append -- $cmds
		}
	}
}

proc action_Cut {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		if {[llength $cmds] == 1} {
			set cmds [string trim $cmds "{}"]
		}
		clipboard clear
		clipboard append -- $cmds
		action_Delete
		dvd_modified true true
	}
}

proc action_Paste_get_clipboard {} {
	if {[catch {set cmds [clipboard get]}]} {
		bell
		return ""
	}
	set cmds [string trim $cmds " \t\n"]
	if {[string range $cmds 0 0] != "\{" && [string length $cmds] == 23} {
		set cmds "\{$cmds\}"
	}

	# convert text commands to internal format
	set cmds2 ""
	set idx1 [string first "\[" $cmds]
	while {$idx1 > -1} {
		set idx2 [expr {$idx1 + 24}]
		set idx3 [expr {$idx1 + 3}]
		if {[string range $cmds $idx2 $idx2] == "\]" && [string range $cmds $idx3 $idx3] == " "} {
			incr idx1
			incr idx2 -1
			lappend cmds2 [string range $cmds $idx1 $idx2]
			set cmds [string range $cmds $idx2 end]
		} else {
			incr idx1
			set cmds [string range $cmds $idx1 end]
		}
		set idx1 [string first "\[" $cmds]
	}
	if {$cmds2 != ""} {
		set cmds $cmds2
	}

	if {[string range $cmds 0 0] != "\{"} {
		bell
		return ""
	}
	return $cmds
}

proc action_Paste_before {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr {[llength $::editor_cmds_list]+[llength $cmds]}] > 131} {
		bell
		return
	}
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		if {[insert_cmds $idx $cmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx [expr {$idx + [llength $cmds] -1}]
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Paste_after {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr {[llength $::editor_cmds_list]+[llength $cmds]}] > 131} {
		bell
		return
	}
	set idx [expr {[lindex [get_selection_indexes] 1] + 1}]
	if {$idx != "-"} {
		if {[insert_cmds $idx $cmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx [expr {$idx + [llength $cmds] -1}]
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

# TODO: Rewrite this shit
proc action_Paste_replace {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	set renumber false

	if {[get_selected_cmds] != ""} {
		if {$::config(renumber_gotos)} {
			set renumber true
			set ::config(renumber_gotos) false

			set sidxs [get_selection_indexes]
			set sidx1 [lindex $sidxs 0]
			set sidx2 [lindex $sidxs 1]

			# renumber pasted commands too
			set typeline [expr [get_cmd_linenum ::editor_cmds_list $sidx1]-1]
			if {$typeline > 0} {
				set padcmds {}
				for {set i 0} {$i < $typeline} {incr i} {
					lappend padcmds {00 00 00 00 00 00 00 00}
				}
				set cmds [lrange [renumber_gotos_add [concat $padcmds $cmds] 1 $typeline] $typeline end]
			}

			set oldnumcmds [expr {$sidx2 - $sidx1 + 1}]
			set numnewcmds [llength $cmds]
			set diff [expr {$numnewcmds - $oldnumcmds}]
			if {$diff < 0} {
				set typelist [get_type $sidx2]
				set type [lindex $typelist 0]
				set lastselectedline [lindex $typelist 1]
				set newcmds [get_type_cmds $type]
				set renumbered [renumber_gotos_del $newcmds \
						[expr {$lastselectedline + $diff + 1}] $lastselectedline]
				set_type_cmds $type $renumbered
			}
		}
		set idx [delete_selected_cmds]
		refresh_listbox

		if {[insert_cmds $idx $cmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx [expr {$idx + [llength $cmds] -1}]
			listbox_to_cmds
		} else {
			bell
			return
		}

		if {$renumber} {
			if {$diff > 0} {
				set typelist [get_type $idx]
				set type [lindex $typelist 0]
				set firstline [expr {[lindex $typelist 1] + [llength $cmds] - $diff}]
				set lastline  [expr {$firstline + $diff -1}]
				set newcmds [get_type_cmds $type]
				set renumbered [renumber_gotos_add $newcmds $firstline $lastline]
				set_type_cmds $type $renumbered

				# replace the pasted cmds again, to reset the GOTOs to the original ones.
				if {[lindex [get_selection_indexes] 1] == [expr {[llength $::editor_cmds_list]-1}]} {
					action_Delete
					set idx [expr [lindex [get_selection_indexes] 1]+1]
				} else {
					action_Delete
					set idx [lindex [get_selection_indexes] 1]
				}
				insert_cmds $idx $cmds
				refresh_listbox
				set_listbox_selection $idx [expr {$idx + [llength $cmds] -1}]
				listbox_to_cmds
			}
			set ::config(renumber_gotos) true
		}
		dvd_modified true true
	}
}

proc action_Mod_GOTOs {increment} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return
	}
	set modified false
	for {set idx $idx1} {$idx <= $idx2} {incr idx} {
		set cmd [string range [lindex $::editor_cmds_list $idx] 0 23]
		set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [x2d [string range $cmd 22 23]]
			if {$cmdchk == "003" && $ln == 0} continue
			incr ln $increment
			if {$cmdchk == "003" && $ln == 0} continue
			if {$ln <= 0 || $ln > 128} {
				bell
			} else {
				lset ::editor_cmds_list $idx \
						"\[[string range $cmd 1 20] [format %02X $ln]\]"
				set modified true
			}
		}
	}
	if {$modified} {
		refresh_listbox
		listbox_to_cmds false
		dvd_modified true true
	}
}

proc action_Incr_GOTOs {} {
	action_Mod_GOTOs 1
}

proc action_Decr_GOTOs {} {
	action_Mod_GOTOs -1
}

## action buttons and/or popup menu
## mode arg: "show", "hide", "toggle" or "init"
proc action_buttons {mode} {
	set cmdlist {
		{{Edit_Cmd}		{Space}			{Edit Cmd}}
		{{Edit_PGC}		{Return}		{Edit PGC}}
		{{Add_before}	{Control-<}		{Add before}}
		{{Add_after}	{Control->}		{Add after}}
		{{Delete}		{Delete}		{Delete}}
		{{Duplicate}	{Control-d}		{Duplicate}}
		{{Copy}			{Control-c}		{Copy}}
		{{Append}		{Insert}		{Append}}
		{{Cut}			{Control-x}		{Cut}}
		{{Paste_replace} {Control-v}	{Paste replace}}
		{{Paste_before}	{Shift-V}		{Paste before}}
		{{Paste_after}	{Control-Shift-V} {Paste after}}
		{{Decr_GOTOs}	{Control--}		{Decr GOTOs}}
		{{Incr_GOTOs}	{Control-+}		{Incr GOTOs}}
		{{Move_up}		{Control-[}		{Move up}}
		{{Move_down}	{Control-]}		{Move down}}
	}
	if {! [winfo exists .actionmenu]} {
		menu .actionmenu
		if {$::config(tearoff_menus)} {
			set menunum 1
		} else {
			set menunum 0
		}
		set ::main::last_menu_selected $menunum
		foreach item $cmdlist {
			set cmd [lindex $item 0]
			if {$cmd != "-"} {
				set cmdname "::main::action_$cmd"
				.actionmenu add command \
						-label [lindex $item 2] -command "set ::main::last_menu_selected $menunum; $cmdname" \
						-accelerator [string map {"\\" "" "Control-" "Ctrl+" "Shift-" "Shift+"} [lindex $item 1]]
				incr menunum
			}
		}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if {[winfo exists .mf.cf.f]} {
				set show 0
			} else {
				set show 1
			}
		}
		"init" {
			if {$::config(buttons)} {
				set show 1
			} else {
				set show 0
			}
		}
	}
	if {$show} {
		set ::main::buttons_opened 1

		if {! [winfo exists .mf.cf.f]} {
			frame .mf.cf.f
			set col 0
			set row 0
			foreach item $cmdlist {
				set cmd [lindex $item 0]
				if {$cmd != "-"} {
					set cmdname "::main::action_$cmd"
					button .mf.cf.f.b${row}_$col -text [lindex $item 2] -command $cmdname -width 10 -pady 0
					::balloon::set_balloon .mf.cf.f.b${row}_$col "[lindex $item 2]\nShortcut: [lindex $item 1]"
					if {$::config(tk) == "x11"} {
						grid .mf.cf.f.b${row}_$col -column $col -row $row -sticky ew -padx 0
					} else {
						grid .mf.cf.f.b${row}_$col -column $col -row $row -sticky ew -padx 1
					}
				}
				if {$row == 0} {
					incr row
					grid columnconf .mf.cf.f $col -weight 1
				} else {
					set row 0
					incr col
				}
				if {$::config(tk) == "x11"} {
					grid rowconf .mf.cf.f 0 -pad 0
					grid rowconf .mf.cf.f 1 -pad 0
				} else {
					grid rowconf .mf.cf.f 0 -pad 2
					grid rowconf .mf.cf.f 1 -pad 2
				}
			}
			pack .mf.cf.f -fill x -pady 1
		}
	} else {
		set ::main::buttons_opened 0

		if {[winfo exists .mf.cf.f]} {
			destroy .mf.cf.f
			.mf.cf configure -height 1
		}
	}
}

proc display_doc {} {
	set rootpath [file dirname $::config(bindir)]
	set filename ""
	if {$::tcl_platform(platform) == "windows"} {
		if {[file exists [file join $rootpath "PgcEdit_Manual.chm"]]} {
			set filename [file join $rootpath "PgcEdit_Manual.chm"]
		}
		if {$filename == "" && [file exists [file join [file join $rootpath doc] "PgcEdit_Manual.chm"]]} {
			set filename [file join [file join $rootpath doc] "PgcEdit_Manual.chm"]
		}
	}
	if {$filename == "" && [file exists [file join [file join $rootpath doc] "PgcEdit_Manual.htm"]]} {
		set filename [file join [file join $rootpath doc] "PgcEdit_Manual.htm"]
	}
	if {$filename == "" && [file exists [file join $rootpath "PgcEdit_Manual.htm"]]} {
		set filename [file join $rootpath "PgcEdit_Manual.htm"]
	}
	if {$filename == "" && [file exists [file join [file join $rootpath doc] "PgcEdit_Manual.txt"]]} {
		set filename [file join [file join $rootpath doc] "PgcEdit_Manual.txt"]
	}
	if {$filename == "" && [file exists [file join $rootpath "PgcEdit_Manual.txt"]]} {
		set filename [file join $rootpath "PgcEdit_Manual.txt"]
	}
	if {$filename == ""} {
		set filename "http://www.videohelp.com/~r0lZ/pgcedit/doc/PgcEdit_Manual.htm"
	}
	switch [string toupper [file extension $filename]] {
		.CHM {
			if {[info exists ::freewrap::progname]} {
				if {[catch {set viewer \
						[lindex [string map {\\ /} [::freewrap::shell_getCmd_imp .chm open]] 0] }]} {
					set viewer "C:/WINDOWS/hh.exe"
				}
			} else {
				set viewer "C:/WINDOWS/hh.exe"
			}
			exec -- $viewer $filename &
		}
		.HTM {
			display_html $filename
		}
		.TXT {
			showtextfile .doc "PgcEdit Documentation" $filename
		}
	}
}

proc display_html {page} {
	get_web_browser
	if {[catch {exec -- $::config(web_browser) $page &}]} {
		set ::config(web_browser) ""
	}
}

proc get_web_browser {} {
	if {$::config(web_browser) == ""} {
		if {$::tcl_platform(platform) == "windows"} {
			set tmp [tk_getOpenFile -filetypes {{{Program files} {*.EXE}}} -initialdir "/" -parent . -title {PgcEdit first time setup: Please select your web browser}]
		} else {
			set tmp [tk_getOpenFile -initialdir "/" -parent . -title {PgcEdit first time setup: Please select your web browser}]
		}
		if {$tmp == ""} return
		set ::config(web_browser) $tmp
	}
	if {! [file exists $::config(web_browser)] && $::config(platform) != "mac"} {
		if {$::tcl_platform(platform) == "windows"} {
			set tmp [tk_getOpenFile -filetypes {{{Program files} {*.EXE}}} -initialdir "/" -parent . -title {PgcEdit first time setup: Please select your web browser}]
		} else {
			set tmp [tk_getOpenFile -initialdir "/" -parent . -title {PgcEdit first time setup: Please select your web browser}]
		}
		if {$tmp == ""} return
		set ::config(web_browser) $tmp
	}
}

#############################################################################
## Main program entry
#############################################################################
wm withdraw .
update idletasks

wm protocol . WM_DELETE_WINDOW {cleanexit 0}

namespace import ::readdvd::*
namespace import ::disasm::*

set ::scratchbook_old_selection {0 0}
set ::scratchbook_active true
set ::scratchbook_buffer [list {* scratchbook}]
if {[file exists $::config(default_scratchbook_filename)]} {
	set ::scratchbook_buffer [readFile $::config(default_scratchbook_filename) binary]
}
set ::dvd_modified false

#################
#
# Build GUI.
#

# the DVD commands as they appear in the listbox window
set ::editor_cmds_list {}
set ::editor_cmds_list_no_hex {}

frame .mf
pack  .mf -fill both -expand 1 -padx 3 -pady 2
frame .mf.tb ;# toolbar container
pack  .mf.tb -fill x -anchor w
::toolbar::init_toolbar $::config(toolbar)
set ::toolbar::toolbar_opened $::config(toolbar)
frame .mf.f1
pack  .mf.f1 -fill x -pady 1 -padx 1
menubutton .mf.f1.mb -text "PGC" -pady 0 -width 7

pack  .mf.f1.mb -side left -fill y
bind  .mf.f1.mb <3> "::main::selector toggle ::main::labelslist ::main::cmdslist"
bind  .mf.f1.mb <Control-Double-2> {console show}
if {$::config(use_preview)} {
	set play {
		#define play_bitmap_width 8
		#define play_bitmap_height 13
		static unsigned char play_bitmap_bits[] = {
		   0x01,
		   0x03,
		   0x07,
		   0x0f,
		   0x1f,
		   0x3f,
		   0x7f,
		   0x3f,
		   0x1f,
		   0x0f,
		   0x07,
		   0x03,
		   0x01};
	}
	image create bitmap ::main::bitmapplay -background {} -data $play
	unset play
	checkbutton .mf.f1.cp -image ::main::bitmapplay -variable ::config(autopreview) \
			-onvalue true -offvalue false -padx 0 -bd 2
	pack  .mf.f1.cp -side left -padx 0
	::balloon::set_balloon .mf.f1.cp "Automatically start Preview\nwhen a PGC is selected."
}
frame .mf.f1.lf -bd 1 -relief sunken
pack  .mf.f1.lf -side left -padx 6 -pady 1 -fill both -expand 1
label .mf.f1.lf.lab -text "" -anchor center -bd 0
pack  .mf.f1.lf.lab -side left -padx 2 -fill both -expand 1
set    ::widget(main_cmd_lab) .mf.f1.lf.lab
bind  $::widget(main_cmd_lab) <3> {::toolbar::init_toolbar toggle}
label .mf.f1.lf.lab2 -text "" -anchor center -bd 0
pack  .mf.f1.lf.lab2 -side left -padx 2 -fill y
set ::widget(main_cmd_lab2) .mf.f1.lf.lab2
checkbutton .mf.f1.sb -text "Scratchbook" -command ::main::scratchbook \
	-variable ::scratchbook_active -onvalue true -offvalue false
if {$::config(tk) != "aqua"} {
	.mf.f1.sb configure -width 12 -indicatoron false -relief sunken \
			-offrelief raised -selectcolor pink -pady 0
}
pack .mf.f1.sb -side right -fill y

# main listbox
frame .mf.f2
pack  .mf.f2 -fill both -expand 1 -pady 1 -padx 1
frame .mf.f2.selector	;# placeholder for Selector
pack  .mf.f2.selector -fill both -expand 0 -side left
frame .mf.f2.main
pack  .mf.f2.main -fill both -expand 1 -side left
set ::widget(main_listbox) .mf.f2.main.l
listbox $::widget(main_listbox) -exportselection 0 -selectmode extended \
		-xscrollcommand ".mf.f2.main.xscr set" -yscrollcommand ".mf.f2.main.yscr set" \
		-listvariable ::editor_cmds_list -font $::config(fixed_width_font) \
		-activestyle none -height 3
if {!$::config(show_hex_codes)} {
	$::widget(main_listbox) configure -listvariable ::editor_cmds_list_no_hex
}
scrollbar .mf.f2.main.yscr -command "$::widget(main_listbox) yview"
scrollbar .mf.f2.main.xscr -command "$::widget(main_listbox) xview" -orient horizontal
grid  $::widget(main_listbox)    -column 0 -row 0 -sticky nesw
grid  .mf.f2.main.yscr -column 1 -row 0 -sticky ns
grid  .mf.f2.main.xscr -column 0 -row 1 -sticky ew
grid columnconf .mf.f2.main 0 -weight 1
grid rowconf    .mf.f2.main 0 -weight 1

# action buttons
frame .mf.cf
pack  .mf.cf -fill x
action_buttons init

# status bar and Find GUI
frame  .mf.st
pack   .mf.st -fill x -side bottom

frame  .mf.st.f -bd 2 -relief sunken
pack   .mf.st.f -side left -fill both -expand 1
canvas .mf.st.f.status -height 8 -width 100
pack   .mf.st.f.status -side left -fill both -expand 1
set ::progressbar_text [.mf.st.f.status create text 3 1 -anchor nw -text "" -font $::config(global_font)]
bind   .mf.st.f.status <1> {.mf.st.f.status configure -width 50}
bind   .mf.st.f.status <Double-1> {openlog}
bind   .mf.st.f.status <3> {::main::action_buttons toggle}

entry  .mf.st.e -textvariable ::main::searchstring -width 26 -font $::config(fixed_width_font)
pack   .mf.st.e -side left -padx 2
bind   .mf.st.e <1> {.mf.st.f.status configure -width 50}

update
.mf.st.f.status configure -height [expr [winfo height .mf.st.e] - 6]

bind   .mf.st.e <Return> "::main::find_in_pgc -forwards"
bind   .mf.st.e <Control-z> {set ::main::searchstring $::config(searchstring); .mf.st.e icursor end}
bind   .mf.st.e $::config(popups_button) "tk_popup .mf.st.e.searcheditmenu \[winfo pointerx .\] \[winfo pointery .\] 5"
menu .mf.st.e.searcheditmenu -tearoff 0
.mf.st.e.searcheditmenu add command -label "Restore default" -accelerator "Ctrl+Z" -command {set ::main::searchstring $::config(searchstring); .mf.st.e icursor end}
.mf.st.e.searcheditmenu add separator
.mf.st.e.searcheditmenu add command -label "Copy"  -accelerator "Ctrl+C" -command {clipboard clear; clipboard append $::main::searchstring}
.mf.st.e.searcheditmenu add command -label "Paste" -accelerator "Ctrl+V" -command {if [catch {set ::main::searchstring [string trim [clipboard get] "\012 "]}] {bell} else {.mf.st.e icursor end}}
.mf.st.e.searcheditmenu add separator
.mf.st.e.searcheditmenu add command -label "Clear" -accelerator "Ctrl+K" -command {set ::main::searchstring ""}
set ::main::searchstring $::config(searchstring)

if {$::tcl_platform(platform) == "windows"} {set minwidth 1} else {set minwidth 0}
button .mf.st.b1 -text "<<" -width 2 -pady 0 -command "::main::find_in_dvd -backwards"
pack   .mf.st.b1 -side left -padx $minwidth -fill y
button .mf.st.b2 -text "<"  -width 2 -pady 0 -command "::main::find_in_pgc -backwards"
pack   .mf.st.b2 -side left -padx $minwidth -fill y
button .mf.st.b3 -text ">"  -width 2 -pady 0 -command "::main::find_in_pgc -forwards"
pack   .mf.st.b3 -side left -padx $minwidth -fill y
button .mf.st.b4 -text ">>" -width 2 -pady 0 -command "::main::find_in_dvd -forwards"
pack   .mf.st.b4 -side left -padx $minwidth -fill y
button .mf.st.ba -text "all" -width 2 -pady 0 -command "::utils::find_results \$::main::searchstring \$::main::searchcase \$::main::searchregexp"
pack   .mf.st.ba -side left -padx $minwidth -fill y
checkbutton .mf.st.c1 -bd $::config(min_bd) -text "case" -variable ::main::searchcase
pack   .mf.st.c1 -side left
checkbutton .mf.st.c2 -bd $::config(min_bd) -text "reg expr" -variable ::main::searchregexp
pack   .mf.st.c2 -side left
if {$::config(platform) == "mac"} {
	pack [frame .mf.st.macpad] -side left -padx 5
}

# bindings
bind .mf.st.ba <3> "::utils::find_results \$::main::searchstring \$::main::searchcase \$::main::searchregexp true"
::balloon::set_balloon .mf.st.ba "Display all matches.\nRight-click to search only\nin menu buttons and BOVs."
bind $::widget(main_listbox) <1> {focus $::widget(main_listbox); ::main::highlight_goto_bound false}
#bind $::widget(main_listbox) <Control-1>     {$::widget(main_listbox) selection clear 0 end}
bind $::widget(main_listbox) <Double-1>      ::main::action_Edit_Cmd
bind $::widget(main_listbox) <Control-1> {focus $::widget(main_listbox); ::main::set_listbox_selection [$::widget(main_listbox) index \
			@[expr {[winfo pointerx .]-[winfo rootx $::widget(main_listbox)]}],[expr {[winfo pointery .]-[winfo rooty $::widget(main_listbox)]}]]; \
			::main::highlight_goto_bound true}
bind $::widget(main_listbox) <2> {focus $::widget(main_listbox); ::main::set_listbox_selection [$::widget(main_listbox) index \
			@[expr {[winfo pointerx .]-[winfo rootx $::widget(main_listbox)]}],[expr {[winfo pointery .]-[winfo rooty $::widget(main_listbox)]}]]; \
			::main::highlight_goto_bound true}
bind $::widget(main_listbox) $::config(popups_button) {tk_popup .actionmenu [winfo pointerx .] [winfo pointery .] $::main::last_menu_selected}
bind $::widget(main_listbox) <space>			::main::action_Edit_Cmd
bind $::widget(main_listbox) <Return>			::main::action_Edit_PGC
bind $::widget(main_listbox) <Control-z>		::main::action_Undo_pgc
bind $::widget(main_listbox) <Delete>			::main::action_Delete
bind $::widget(main_listbox) <BackSpace>		::main::action_Delete
bind $::widget(main_listbox) <Control-d>		::main::action_Duplicate
bind $::widget(main_listbox) <Control-less>	::main::action_Add_before
bind $::widget(main_listbox) <Control-greater> ::main::action_Add_after
bind $::widget(main_listbox) <Insert>			::main::action_Append
bind $::widget(main_listbox) <Control-bracketleft> ::main::action_Move_up
bind $::widget(main_listbox) <Control-bracketright> ::main::action_Move_down
bind $::widget(main_listbox) <Control-c>		::main::action_Copy
bind $::widget(main_listbox) <Control-x>		::main::action_Cut
bind $::widget(main_listbox) <Control-v>		::main::action_Paste_replace
bind $::widget(main_listbox) <V> 				::main::action_Paste_before
bind $::widget(main_listbox) <Control-V>		::main::action_Paste_after
bind $::widget(main_listbox) <Control-plus>	::main::action_Incr_GOTOs
bind $::widget(main_listbox) <Control-minus>	::main::action_Decr_GOTOs
bind $::widget(main_listbox) <Control-j>		::main::action_Jump_to_target
bind $::widget(main_listbox) <Control-Insert>	::main::send_to_clipboard
#bind $::widget(main_listbox) <Control-a> {::main::set_listbox_selection 0 end}
bind $::widget(main_listbox) <Escape> {$::widget(main_listbox) see 0; $::widget(main_listbox) activate 0; $::widget(main_listbox) selection set 0}
bind $::widget(main_listbox) <Home>   {$::widget(main_listbox) see 0; $::widget(main_listbox) activate 0; $::widget(main_listbox) selection clear 0 end; $::widget(main_listbox) selection set 0}
bind $::widget(main_listbox) <End>    {$::widget(main_listbox) see end; $::widget(main_listbox) activate end; $::widget(main_listbox) selection clear 0 end; $::widget(main_listbox) selection set end}
bind $::widget(main_listbox) <MouseWheel> {
	if {%D > 0} {
		$::widget(main_listbox) yview scroll  1 units
	} else {
		$::widget(main_listbox) yview scroll -1 units
	}
}
bind $::widget(main_listbox) <Right> ::main::select_selector
bind $::widget(main_listbox) <Left>  ::main::select_selector
proc ::main::select_selector {} {
	if {[winfo exists .mf.f2.selector.a.lb]} {
		focus .mf.f2.selector.a.lb
		set idx [.mf.f2.selector.a.lb index active]
		if {$idx > -1} {
			.mf.f2.selector.a.lb see $idx
		}
	} else {
		focus .mf.f1.mb
	}
}

::menus::build_main_menu

proc renumber_gotos_warning {} {
	if {$::config(renumber_gotos)} {
		set txt "Warning!\n\nWhen this option is turned ON, the GOTO lines are"
		set txt "$txt\nrenumbered when commands are inserted or deleted."
		set txt "$txt\nHowever, the commands that are copied to the clipboard"
		set txt "$txt\nand the new commands that are inserted in the listbox"
		set txt "$txt\nare NOT renumbered, because it is impossible to know"
		set txt "$txt\nwhere the GOTOs were originally branching to."
		set txt "$txt\n\nRule:\nOnly the commands that are present in the listbox"
		set txt "$txt\nBEFORE AND AFTER the operation are renumbered."
		set txt "$txt\n\nKeep this in mind, and carefully verify all GOTOs after"
		set txt "$txt\neach operation!"
		::msgbox::MsgBox -title "PgcEdit: Renumber GOTOs" -icon warning \
				-message $txt -parent . -skip "Renumber GOTOs warning"
	}
}

# start is Scratchbook mode
scratchbook

# open selector if requested
if {$::config(selector)} {
	selector show ::main::labelslist ::main::cmdslist
}

wm title . "PgcEdit v $::program_version"
show_msg "Welcome to PgcEdit version $::program_version ($::release_date)"
update
if {[catch {wm geometry . $::config(main_win_geometry)}]} {
	wm geometry . 820x500
}
wm deiconify .
focus .
update idletasks

} ;# end namespace ::main
